#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <math.h>

#include "phoebe_build_config.h"

#include "phoebe_accessories.h"
#include "phoebe_allocations.h"
#include "phoebe_base.h"
#include "phoebe_calculations.h"
#include "phoebe_error_handling.h"
#include "phoebe_fortran_interface.h"
#include "phoebe_global.h"
#include "phoebe_keywords.h"
#include "phoebe_type_handling.h"

int read_in_synthetic_data (PHOEBE_data *data, int curve, PHOEBE_output_indep indep, PHOEBE_output_dep dep, PHOEBE_output_weight weight)
	{
	/* This function creates the WD input file 'lcin.active' and calls LC to    */
	/* calculate the data. The points in which values should be calculated are  */
	/* passed by the 'data.indep' argument. In general, these should correspond */
	/* to experimental data, but may be also some equidistant set of times or   */
	/* phases if only synthetic curve is to be created without any experimental */
	/* counterpart.                                                             */
	/*                                                                          */
	/* Return values:                                                           */
	/*                                                                          */
	/*   -8  ..  dependent variable enumeration readout failed.                 */
	/*   -2  ..  the requested curve isn't initialized                          */
	/*    0  ..  everything ok, read-in successful                              */

	int i;
	int mpage, jdphs;
	int errno;
	double L1, L2;

	char *filename = resolve_relative_filename ("lcin.active");
	WD_LCI_parameters params;

	switch (dep)
		{
		case OUTPUT_TOTAL_FLUX:
			mpage = 1;
		break;
		case OUTPUT_MAGNITUDE:
			mpage = 1;
		break;
		case OUTPUT_PRIMARY_RV:
			mpage = 2;
		break;
		case OUTPUT_SECONDARY_RV:
			mpage = 2;
		break;
		default:
			phoebe_warning ("function read_in_synthetic_data () called with invalid dep type, aborting.\n");
			return -8;
		break;
		}

	errno = read_in_wd_lci_parameters (&params, mpage, curve);
	if (errno == -2) return -2;

	create_lci_file (filename, params);

	switch (dep)
		{
		case OUTPUT_MAGNITUDE:
			call_wd_to_get_fluxes (data);
		break;
		case OUTPUT_PRIMARY_FLUX:
			call_wd_to_get_fluxes (data);
		break;
		case OUTPUT_SECONDARY_FLUX:
			call_wd_to_get_fluxes (data);
		break;
		case OUTPUT_TOTAL_FLUX:
			call_wd_to_get_fluxes (data);
		break;
		case OUTPUT_PRIMARY_RV:
			call_wd_to_get_rv1    (data);
		break;
		case OUTPUT_SECONDARY_RV:
			call_wd_to_get_rv2    (data);
		break;
		}

	remove (filename);
	free (filename);

	if ( dep == OUTPUT_MAGNITUDE )
		{
		double mnorm;
		phoebe_get_parameter_value ("phoebe_mnorm", &mnorm);
		transform_flux_to_magnitude (data, mnorm);
		}

	if ( dep == OUTPUT_PRIMARY_RV || dep == OUTPUT_SECONDARY_RV )
		{
		for (i = 0; i < data->ptsno; i++)
			data->dep[i] *= 100.0;
		}

	return data->ptsno;
	}

char *parse_data_line (char *in)
	{
	/* This function takes a string, parses it (cleans it of all comment deli-  */
	/* meters, spaces, tabs and newlines), copies the contents to the newly al- */
	/* located string. The original string isn't modified (passed by value!).   */

	char *value = NULL;

	/* If the line contains the comment delimeter '#', discard everything that  */
	/* follows it (strip the input line):                                       */

	if (strchr (in, '#') != NULL)
		in[strlen(in)-strlen(strchr(in,'#'))] = '\0';
	if (strlen(in) == 0) return NULL;

	/* If we have spaces, tabs or newlines in front of the first character in   */
	/* line, remove them by incrementing the pointer by 1:                      */
	while ( (*in ==  ' ') || (*in == '\t') || (*in == '\n') )
		{
		if (strlen (in) > 1) in++;
		else return NULL;
		}

	/* Let's do the same for the tail of the string:                            */
	while ( (in[strlen(in)-1] ==  ' ') || (in[strlen(in)-1] == '\t') || (in[strlen(in)-1] == '\n') )
		in[strlen(in)-1] = '\0';

	value = strdup (in);
	return value;
	}

int read_in_experimental_data (char *filename, PHOEBE_data *data, int indep, int outindep, int dep, int outdep, int weight, int outweight)
	{
	/* This function should be called *EXCLUSIVELY* by PHOEBE_INPUT_* and       */
	/* PHOEBE_OUTPUT_* enumerate types. "int" in the prototype is used to allow */
	/* certain calls to pass -1 rather than the enumerate type, which means     */
	/* that no potential conversion should be made to that variable. A typical  */
	/* example is updating the statistics window.                               */

	FILE *file;
	int i = 0;

	char line[255];

	char *in;

	int no_of_columns;
	int line_number = 0;

	double hjd0, period, dpdt, pshift;

	/* First step: check whether the user wants to plot something that doesn't  */
	/* make sense:                                                              */
	if ( (dep == INPUT_MAGNITUDE) && (outdep == OUTPUT_PRIMARY_FLUX) )
		{
		phoebe_warning ("read_in_experimental_data () called with OUTPUT_PRIMARY_FLUX.\n\tIt is not possible to extract primary flux contributions from data!\n");
		return 1;
		}

	if ( (dep == INPUT_MAGNITUDE) && (outdep == OUTPUT_SECONDARY_FLUX) )
		{
		phoebe_warning ("read_in_experimental_data () called with OUTPUT_SECONDARY_FLUX.\n\tIt is not possible to extract secondary flux contributions from data!\n");
		return 1;
		}

	if ( (indep == INPUT_PHASE) && (outindep == OUTPUT_HJD) )
		phoebe_warning ("read_in_experimental_data () called with PHASE -> HJD transformation\n\trequest. This isn't possible, leaving data intact.");

	/* Check for the filename validity:                                         */

	if ( (!filename_exists (filename)) || 
			 (!filename_is_regular_file (filename)) )
		{
		phoebe_warning ("read_in_experimental_data () notice: filename not valid!\n");
		return -2;
		}

	/* Just to be sure, we check the fopen statement against NULL:              */
	if ( (file = fopen (filename, "r")) == NULL )
		{
		phoebe_warning ("read_in_experimental_data () notice: file opening failed!");
		return -2;
		}

	while (!feof (file))
		{
		fgets (line, 255, file);
		line_number++;

		/* If the line is commented or empty, skip it:                            */
		if ( (in = parse_data_line (line)) == NULL ) continue;

		/* Check how many columns are expected in the input file:                 */
		if (weight == INPUT_UNAVAILABLE) no_of_columns = 2; else no_of_columns = 3;

		phoebe_data_realloc (data, i+1);

		if (no_of_columns == 2)
			{
			if (sscanf (in, "%lf %lf", &data->indep[i], &data->dep[i]) != 2)
				{
				phoebe_warning ("read_in_experimental_data () notice: line %d discarded!\n", line_number);
				continue;
				}
			data->weight[i] = 1.0;
			}

		if (no_of_columns == 3)
			if (sscanf (in, "%lf %lf %lf", &data->indep[i], &data->dep[i], &data->weight[i]) != 3)
				{
				phoebe_warning ("read_in_experimental_data () notice: line %d discarded!\n", line_number);
				continue;
				}
		
		i++;
		}
	fclose (file);
	data->ptsno = i-1;

	/* Now we do all the necessary transformations:                             */

	if ( (indep == INPUT_HJD) && (outindep == OUTPUT_PHASE) )
		{
		read_in_ephemeris_parameters (&hjd0, &period, &dpdt, &pshift);
		transform_hjd_to_phase (data, hjd0, period, dpdt, pshift);
		}

	if ( (dep == INPUT_MAGNITUDE) && (outdep == OUTPUT_TOTAL_FLUX) )
		{
 		double mnorm = phoebe_get_value_double ("phoebe_mnorm");
 		transform_magnitude_to_flux (data, mnorm);
		if (weight == INPUT_STANDARD_DEVIATION)
			/* Be sure *ALWAYS* to call this *AFTER* magnitude->flux conversion!!!  */
			transform_magnitude_sigma_to_flux_sigma (data, mnorm);
		}

	if ( (dep == INPUT_FLUX) && (outdep == OUTPUT_MAGNITUDE) )
		transform_flux_to_magnitude (data, phoebe_get_value_double ("phoebe_mnorm"));

	if ( weight == INPUT_STANDARD_DEVIATION && outweight == OUTPUT_STANDARD_WEIGHT )
		transform_sigma_to_weight (data);

	if ( weight == INPUT_STANDARD_WEIGHT && outweight == OUTPUT_STANDARD_DEVIATION )
		transform_weight_to_sigma (data);

	return data->ptsno;
	}

int read_in_wd_lci_parameters (WD_LCI_parameters *params, int MPAGE, int curve)
	{
	/* This function reads out all variables that build up the LCI file.        */
	/*                                                                          */
	/* Return values:                                                           */
	/*                                                                          */
	/*   -2  ..  the passed curve number refers to non-existent curve           */
	/*    0  ..  everything ok                                                  */
	/*    1  ..  the passed mpage isn't (yet) supported                         */

	int lcno = phoebe_get_value_int ("phoebe_lcno");
	int rvno = phoebe_get_value_int ("phoebe_rvno");

	char *filter = NULL;
	const char *readout_str;

	/* This switch determines which operation should the program do and it is   */
	/* set from the passed value:                                               */
	params->MPAGE = MPAGE;

	/* First let's check whether we have out-of-bounds situation:               */
	if (MPAGE == 1)
		if (curve < 0 || curve > lcno-1) return -2;
	if (MPAGE == 2)
		if (curve < 0 || curve > rvno-1) return -2;
	if (MPAGE != 1 && MPAGE != 2) return 1;

	if (phoebe_get_value_bool ("phoebe_reffect_switch") == YES) params->MREF = 2; else params->MREF = 1;
	params->NREF = phoebe_get_value_int ("phoebe_reffect_reflections");

	if (phoebe_get_value_bool ("phoebe_spots_move1") == YES) params->IFSMV1 = 1; else params->IFSMV1 = 0;
	if (phoebe_get_value_bool ("phoebe_spots_move2") == YES) params->IFSMV2 = 1; else params->IFSMV2 = 0;
	if (phoebe_get_value_bool ("phoebe_proximity_rv1_switch") == YES) params->ICOR1  = 1; else params->ICOR1  = 0;
	if (phoebe_get_value_bool ("phoebe_proximity_rv2_switch") == YES) params->ICOR2  = 1; else params->ICOR2  = 0;

	readout_str = phoebe_get_value_string ("phoebe_ld_model");
	if (strcmp (readout_str, "Linear cosine law") == 0) params->LD = 1;
	if (strcmp (readout_str, "Logarithmic law")   == 0) params->LD = 2;
	if (strcmp (readout_str, "Square root law")   == 0) params->LD = 3;

	readout_str = phoebe_get_value_string ("phoebe_indep");
	if (strcmp (readout_str, "Time (HJD)")  == 0) params->JDPHS = 1;
	if (strcmp (readout_str, "Phase") == 0) params->JDPHS = 2;

	params->HJD0   = phoebe_get_value_double ("phoebe_hjd0");
	params->PERIOD = phoebe_get_value_double ("phoebe_period");
	params->DPDT   = phoebe_get_value_double ("phoebe_dpdt");
	params->PSHIFT = phoebe_get_value_double ("phoebe_pshift");

	if (phoebe_get_value_bool ("phoebe_synscatter_switch") == YES)
		{
		params->SIGMA  = phoebe_get_value_double ("phoebe_synscatter_sigma");
		readout_str = phoebe_get_value_string ("phoebe_synscatter_levweight");
		if (strcmp (readout_str, "No level-dependent weighting") == 0) params->WEIGHTING = 0;
		if (strcmp (readout_str, "Poissonian scatter")           == 0) params->WEIGHTING = 1;
		if (strcmp (readout_str, "Low light scatter")            == 0) params->WEIGHTING = 2;
		params->SEED   = phoebe_get_value_double ("phoebe_synscatter_seed");
		}
	else
		{
		params->SIGMA = 0.0; params->WEIGHTING = 0; params->SEED = 0.0;
		}

	/* The following parameters are dummies, since PHOEBE calls WD algorithm    */
	/* for each phase individually:                                             */
	params->HJDST  = 0.0;
	params->HJDSP  = 1.0;
	params->HJDIN  = 0.1;
	params->PHSTRT = 0.0;
	params->PHSTOP = 1.0;
	params->PHIN   = 0.1;

	/* Normalizing magnitude (PHNORM) tells WD at what phase the normalized     */
	/* flux should be 1. Usually this is 0.25, but could in general be at some  */
	/* other phase. Since there is no support for that in PHOEBE yet, it is     */
	/* hardcoded to 0.25, but it should be changed.                             */
	params->PHNORM = 0.25;

	readout_str = phoebe_get_value_string ("phoebe_model");
	if (strcmp (readout_str, "X-ray binary"                                         ) == 0) params->MODE = -1;
	if (strcmp (readout_str, "Unconstrained binary system"                          ) == 0) params->MODE =  0;
	if (strcmp (readout_str, "Overcontact binary of the W UMa type"                 ) == 0) params->MODE =  1;
	if (strcmp (readout_str, "Detached binary"                                      ) == 0) params->MODE =  2;
	if (strcmp (readout_str, "Overcontact binary not in thermal contact"            ) == 0) params->MODE =  3;
	if (strcmp (readout_str, "Semi-detached binary, primary star fills Roche lobe"  ) == 0) params->MODE =  4;
	if (strcmp (readout_str, "Semi-detached binary, secondary star fills Roche lobe") == 0) params->MODE =  5;
	if (strcmp (readout_str, "Double contact binary"                                ) == 0) params->MODE =  6;

	readout_str = phoebe_get_value_string ("phoebe_cc");
	if (strcmp (readout_str,    "None"                    ) == 0) params->CC = 0;
	if (strcmp (readout_str,    "Main-sequence constraint") == 0) params->CC = 1;

	if (phoebe_get_value_bool ("phoebe_calculate_hla")    == YES) params->CALCHLA = 1; else params->CALCHLA = 0;
	if (phoebe_get_value_bool ("phoebe_calculate_vga")    == YES) params->CALCVGA = 1; else params->CALCVGA = 0;
	params->ASINI = phoebe_get_value_bool ("phoebe_asini_switch");

	if (phoebe_get_value_bool ("phoebe_usecla_switch") == YES) params->IPB  = 1; else params->IPB  = 0;
	if (phoebe_get_value_bool ("phoebe_atm1_switch")  == YES) params->IFAT1 = 1; else params->IFAT1 = 0;
	if (phoebe_get_value_bool ("phoebe_atm2_switch")  == YES) params->IFAT2 = 1; else params->IFAT2 = 0;

	params->N1 = phoebe_get_value_int ("phoebe_grid_finesize1");
	params->N2 = phoebe_get_value_int ("phoebe_grid_finesize2");

	params->PERR0  = phoebe_get_value_double ("phoebe_perr0");
	params->DPERDT = phoebe_get_value_double ("phoebe_dperdt");

	/* THE applies only to X-ray binaries, but it isn't supported yet.          */
	params->THE    = 0.0;

	/* VUNIT is the internal radial velocity unit. This should be calculated    */
	/* rather than determined by the user, however it isn't written yet. It is  */
	/* thus hardcoded to the most obvious number, that is 100 km/s.             */
	params->VUNIT  = 100.0;

	params->E      = phoebe_get_value_double ("phoebe_ecc");
	params->SMA    = phoebe_get_value_double ("phoebe_sma");
	params->F1     = phoebe_get_value_double ("phoebe_f1");
	params->F2     = phoebe_get_value_double ("phoebe_f2");
	params->VGA    = phoebe_get_value_double ("phoebe_vga");
	params->INCL   = phoebe_get_value_double ("phoebe_incl");
	params->GR1    = phoebe_get_value_double ("phoebe_grb1");
	params->GR2    = phoebe_get_value_double ("phoebe_grb2");
	params->LOGG1  = phoebe_get_value_double ("phoebe_logg1");
	params->LOGG2  = phoebe_get_value_double ("phoebe_logg2");
	params->MET1   = phoebe_get_value_double ("phoebe_met1");
	params->MET2   = phoebe_get_value_double ("phoebe_met2");
	params->TAVH   = phoebe_get_value_double ("phoebe_teff1");
	params->TAVC   = phoebe_get_value_double ("phoebe_teff2");
	params->ALB1   = phoebe_get_value_double ("phoebe_alb1");
	params->ALB2   = phoebe_get_value_double ("phoebe_alb2");
	params->PHSV   = phoebe_get_value_double ("phoebe_pot1");
	params->PCSV   = phoebe_get_value_double ("phoebe_pot2");
	params->RM     = phoebe_get_value_double ("phoebe_rm");
	params->XBOL1  = phoebe_get_value_double ("phoebe_ld_xbol1");
	params->XBOL2  = phoebe_get_value_double ("phoebe_ld_xbol2");
	params->YBOL1  = phoebe_get_value_double ("phoebe_ld_ybol1");
	params->YBOL2  = phoebe_get_value_double ("phoebe_ld_ybol2");

	if (MPAGE == 2)
		filter = strdup (phoebe_get_value_list_string ("phoebe_rv_filter", curve));
	else
		filter = strdup (phoebe_get_value_list_string ("phoebe_lc_filter", curve));

	params->IBAND = get_passband_id (filter);

	if (MPAGE == 1)
		{
		params->HLA = phoebe_get_value_list_double ("phoebe_hla", curve);
		params->CLA = phoebe_get_value_list_double ("phoebe_cla", curve);
		}
	else
		{
		/* HLAs and CLAs don't make any sense for RV curves, so we initialize     */
		/* them to some canonic dummy values.                                     */
		params->HLA = 1.0;
		params->CLA = 1.0;
		}

	if (MPAGE == 2)
		{
		params->X1A = phoebe_get_value_list_double ("phoebe_ld_rvx1", curve);
		params->X2A = phoebe_get_value_list_double ("phoebe_ld_rvx2", curve);
		params->Y1A = phoebe_get_value_list_double ("phoebe_ld_rvy1", curve);
		params->Y2A = phoebe_get_value_list_double ("phoebe_ld_rvy2", curve);
		}
	else
		{
		params->X1A = phoebe_get_value_list_double ("phoebe_ld_lcx1", curve);
		params->X2A = phoebe_get_value_list_double ("phoebe_ld_lcx2", curve);
		params->Y1A = phoebe_get_value_list_double ("phoebe_ld_lcy1", curve);
		params->Y2A = phoebe_get_value_list_double ("phoebe_ld_lcy2", curve);
		}

	if (MPAGE == 1)
		{
		params->EL3   = phoebe_get_value_list_double ("phoebe_el3", curve);
		params->OPSF  = phoebe_get_value_list_double ("phoebe_opsf", curve);
		}
	else
		{
		/* Third light and opacity function don't make sense for RVs.             */
		params->EL3 = 1.0;
		params->OPSF = 1.0;
		}

	/* MZERO and FACTOR variables set offsets in synthetic light curves. PHOEBE */
	/* controls this by its own variables, so we hardcode these to 0 and 1.     */
	params->MZERO  = 0.0;
	params->FACTOR = 1.0;
	params->WLA    = get_passband_wavelength (filter);

	/* Spots are not supported yet:                                             */
	params->SPRIM = 0;
	params->SSEC  = 0;
	params->XLAT1  = &(params->THE);
	params->XLONG1 = &(params->THE);
	params->RADSP1 = &(params->THE);
	params->TEMSP1 = &(params->THE);
	params->XLAT2  = &(params->THE);
	params->XLONG2 = &(params->THE);
	params->RADSP2 = &(params->THE);
	params->TEMSP2 = &(params->THE);

	return 0;
	}

int read_in_wd_dci_parameters (WD_DCI_parameters *params)
	{
	/****************************************************************************/
	/* This function reads in DCI parameters. Be sure to allocate (and free)    */
	/* the passed WD_DCI_parameters structure (and its substructures).          */
	/*                                                                          */
	/* Return values:                                                           */
	/*                                                                          */
	/*  -9 .. independent variable enumeration readout failed.                  */
	/*  -8 .. dependent variable enumeration readout failed.                    */
	/*  -7 .. weighting variable enumeration readout failed.                    */
	/*  -6 .. no curves selected for adjustment, nothing to be done.            */
	/*  -5 .. no parameters selected for adjustment, nothing to be done.        */
	/*  -2 .. experimental data cannot be read in, abort the execution.         */
	/*  -1 .. passed values are invalid, abort the execution.                   */
	/*   0 .. everything ok, simplex converged.                                 */
	/****************************************************************************/

	char *pars[] =
		{
		"phoebe_spots_lat1",
		"phoebe_spots_long1",
		"phoebe_spots_rad1",
		"phoebe_spots_temp1",
		"phoebe_spots_lat2",
		"phoebe_spots_long2",
		"phoebe_spots_rad2",
		"phoebe_spots_temp2",
		"phoebe_sma",
		"phoebe_ecc",
		"phoebe_perr0",
		"phoebe_f1",
		"phoebe_f2",
		"phoebe_pshift",
		"phoebe_vga",
		"phoebe_incl",
		"phoebe_grb1",
		"phoebe_grb2",
		"phoebe_teff1",
		"phoebe_teff2",
		"phoebe_alb1",
		"phoebe_alb2",
		"phoebe_pot1",
		"phoebe_pot2",
		"phoebe_rm",
		"phoebe_hjd0",
		"phoebe_period",
		"phoebe_dpdt",
		"phoebe_dperdt",
		"",
		"phoebe_hla",
		"phoebe_cla",
		"phoebe_ld_lcx1",
		"phoebe_ld_lcx2",
		"phoebe_el3"
		};

	int i;

	int lcno = phoebe_get_value_int ("phoebe_lcno");
	int rvno = phoebe_get_value_int ("phoebe_rvno");
	int cno = lcno + rvno;

	/* DC features 35 adjustable parameters and we initialize such arrays:      */
	bool    *tba;
	double *step;

	int marked_tba = 0;

	/* Check whether there are any experimental data present:                   */
	if (cno == 0) return -6;

	/* Allocate memory:                                                         */
	 tba = phoebe_malloc (35 * sizeof ( *tba));
	step = phoebe_malloc (35 * sizeof (*step));

	/* Read in TBA states and step-sizes; note the '!' in front of the tba      */
	/* readout function; that's because WD uses 1 for false and 0 for true.     */
	for (i = 0; i < 35; i++)
		{
		if (i == 29) { tba[i] = !FALSE; continue; } /* reserved WD channel */
		 tba[i] = !phoebe_get_parameter_tba  (pars[i]);
		step[i] =  phoebe_get_parameter_step (pars[i]);
		if (i > 29)
			marked_tba += lcno * phoebe_get_parameter_tba (pars[i]);
		else
			marked_tba += phoebe_get_parameter_tba (pars[i]);
		}

	/* Check if there are no parameters set for adjustment:                     */
	if (marked_tba == 0)
		{
		free (tba);
		free (step);
		return -5;
		}

	params->tba  = tba;
	params->step = step;

	/* Check the presence of RV and LC data:                                    */
	{
	PHOEBE_input_dep dep;
	params->rv1data = FALSE; params->rv2data = FALSE;
	for (i = 0; i < rvno; i++)
		{
		dep = get_input_dependent_variable (phoebe_get_value_list_string ("phoebe_rv_dep", i));
		if (dep == INPUT_PRIMARY_RV)   params->rv1data = TRUE;
		if (dep == INPUT_SECONDARY_RV) params->rv2data = TRUE;
		if (dep == -1) phoebe_warning ("error in dependent variable readout!\n");
		}
	params->nlc = lcno;
	}

	/* DC-related parameters:                                                   */
	params->dclambda = phoebe_get_value_double ("phoebe_dc_lambda");
	params->symder = phoebe_get_value_bool ("phoebe_dc_symder_switch");
	params->n1c = phoebe_get_value_int ("phoebe_grid_coarsesize1");
	params->n2c = phoebe_get_value_int ("phoebe_grid_coarsesize2");
	params->n1f = phoebe_get_value_int ("phoebe_grid_finesize1");
	params->n2f = phoebe_get_value_int ("phoebe_grid_finesize2");

	/* Reflection effect:                                                       */
	params->refswitch = (int) phoebe_get_value_bool ("phoebe_reffect_switch") + 1;
	params->refno     = phoebe_get_value_int ("phoebe_reffect_reflections");

	/* Eclipse/proximity effect:                                                */
	params->rv1proximity = phoebe_get_value_bool ("phoebe_proximity_rv1_switch");
	params->rv2proximity = phoebe_get_value_bool ("phoebe_proximity_rv2_switch");

	/* Limb darkening effect:                                                   */
	{
	char *ld = (char *) phoebe_get_value_string ("phoebe_ld_model");
	params->ldmodel = 0;
	if (strcmp (ld, "Linear cosine law") == 0) params->ldmodel = 1;
	if (strcmp (ld, "Logarithmic law")   == 0) params->ldmodel = 2;
	if (strcmp (ld, "Square root law")   == 0) params->ldmodel = 3;
	if (params->ldmodel == 0) phoebe_warning ("error in LD model readout!\n");
	}

	/* Morphological constraint:                                                */
	{
	char *mode = (char *) phoebe_get_value_string ("phoebe_model");
	params->morph = -2;
	if (strcmp (mode, "X-ray binary"                                         ) == 0) params->morph = -1;
	if (strcmp (mode, "Unconstrained binary system"                          ) == 0) params->morph =  0;
	if (strcmp (mode, "Overcontact binary of the W UMa type"                 ) == 0) params->morph =  1;
	if (strcmp (mode, "Detached binary"                                      ) == 0) params->morph =  2;
	if (strcmp (mode, "Overcontact binary not in thermal contact"            ) == 0) params->morph =  3;
	if (strcmp (mode, "Semi-detached binary, primary star fills Roche lobe"  ) == 0) params->morph =  4;
	if (strcmp (mode, "Semi-detached binary, secondary star fills Roche lobe") == 0) params->morph =  5;
	if (strcmp (mode, "Double contact binary"                                ) == 0) params->morph =  6;
	if (params->morph == -2) phoebe_warning ("error in morphological constraint readout!\n");
	}

	/* Independent fit variable:                                                */
	{
	char *indep = (char *) phoebe_get_value_string ("phoebe_indep");
	params->indep = 0;
	if (strcmp (indep, "Time (HJD)") == 0) params->indep = 1;
	if (strcmp (indep, "Phase") == 0) params->indep = 2;
	if (params->indep == 0) phoebe_warning ("error in independent fit variable readout!\n");
	}

	/* Luminosity decoupling:                                                   */
	params->cladec = phoebe_get_value_bool ("phoebe_usecla_switch");

	/* Model atmosphere switches:                                               */
	params->ifat1 = phoebe_get_value_bool ("phoebe_atm1_switch");
	params->ifat2 = phoebe_get_value_bool ("phoebe_atm2_switch");

	/* Model parameters:                                                        */
	params->hjd0   = phoebe_get_value_double ("phoebe_hjd0");
	params->period = phoebe_get_value_double ("phoebe_period");
	params->dpdt   = phoebe_get_value_double ("phoebe_dpdt");
	params->pshift = phoebe_get_value_double ("phoebe_pshift");
	params->perr0  = phoebe_get_value_double ("phoebe_perr0");
	params->dperdt = phoebe_get_value_double ("phoebe_dperdt");
	params->ecc    = phoebe_get_value_double ("phoebe_ecc");
	params->sma    = phoebe_get_value_double ("phoebe_sma");
	params->f1     = phoebe_get_value_double ("phoebe_f1");
	params->f2     = phoebe_get_value_double ("phoebe_f2");
	params->vga    = phoebe_get_value_double ("phoebe_vga");
	params->incl   = phoebe_get_value_double ("phoebe_incl");
	params->grb1   = phoebe_get_value_double ("phoebe_grb1");
	params->grb2   = phoebe_get_value_double ("phoebe_grb2");
	params->met1   = phoebe_get_value_double ("phoebe_met1");
	params->teff1  = phoebe_get_value_double ("phoebe_teff1");
	params->teff2  = phoebe_get_value_double ("phoebe_teff2");
	params->alb1   = phoebe_get_value_double ("phoebe_alb1");
	params->alb2   = phoebe_get_value_double ("phoebe_alb2");
	params->pot1   = phoebe_get_value_double ("phoebe_pot1");
	params->pot2   = phoebe_get_value_double ("phoebe_pot2");
	params->rm     = phoebe_get_value_double ("phoebe_rm");
	params->xbol1  = phoebe_get_value_double ("phoebe_ld_xbol1");
	params->xbol2  = phoebe_get_value_double ("phoebe_ld_xbol2");
	params->ybol1  = phoebe_get_value_double ("phoebe_ld_ybol1");
	params->ybol2  = phoebe_get_value_double ("phoebe_ld_ybol2");

	/* Passband-dependent parameters:                                           */
	{
	int index;

	params->passband   = phoebe_malloc (cno * sizeof (*(params->passband)));
	params->wavelength = phoebe_malloc (cno * sizeof (*(params->wavelength)));
	params->sigma      = phoebe_malloc (cno * sizeof (*(params->sigma)));
	params->hla        = phoebe_malloc (cno * sizeof (*(params->hla)));
	params->cla        = phoebe_malloc (cno * sizeof (*(params->cla)));
	params->x1a        = phoebe_malloc (cno * sizeof (*(params->x1a)));
	params->y1a        = phoebe_malloc (cno * sizeof (*(params->y1a)));
	params->x2a        = phoebe_malloc (cno * sizeof (*(params->x2a)));
	params->y2a        = phoebe_malloc (cno * sizeof (*(params->y2a)));
	params->el3        = phoebe_malloc (cno * sizeof (*(params->el3)));
	params->opsf       = phoebe_malloc (cno * sizeof (*(params->opsf)));
	params->levweight  = phoebe_malloc (cno * sizeof (*(params->levweight)));

	for (i = 0; i < rvno; i++)
		{
		PHOEBE_input_dep dep = get_input_dependent_variable (phoebe_get_value_list_string ("phoebe_rv_dep", i));
		if (dep == INPUT_SECONDARY_RV && rvno == 2) index = 1; else index = 0;

		params->passband[index]   = get_passband_id         (phoebe_get_value_list_string ("phoebe_rv_filter", i));
		params->wavelength[index] = get_passband_wavelength (phoebe_get_value_list_string ("phoebe_rv_filter", i));
		params->sigma[index]      = phoebe_get_value_list_double ("phoebe_rv_sigma", i);
		params->hla[index]        = 10.0;
		params->cla[index]        = 10.0;
		params->x1a[index]        = phoebe_get_value_list_double ("phoebe_ld_rvx1", i);
		params->y1a[index]        = phoebe_get_value_list_double ("phoebe_ld_rvy1", i);
		params->x2a[index]        = phoebe_get_value_list_double ("phoebe_ld_rvx2", i);
		params->y2a[index]        = phoebe_get_value_list_double ("phoebe_ld_rvy2", i);
		params->el3[index]        = 0.0;
		params->opsf[index]       = 0.0;
		params->levweight[index]  = 0;
		}

	for (i = rvno; i < cno; i++)
		{
		params->passband[i]   = get_passband_id         (phoebe_get_value_list_string ("phoebe_lc_filter", i-rvno));
		params->wavelength[i] = get_passband_wavelength (phoebe_get_value_list_string ("phoebe_lc_filter", i-rvno));
		params->sigma[i]      = phoebe_get_value_list_double ("phoebe_lc_sigma", i-rvno);
		params->hla[i]        = phoebe_get_value_list_double ("phoebe_hla", i-rvno);
		params->cla[i]        = phoebe_get_value_list_double ("phoebe_cla", i-rvno);
		params->x1a[i]        = phoebe_get_value_list_double ("phoebe_ld_lcx1", i-rvno);
		params->y1a[i]        = phoebe_get_value_list_double ("phoebe_ld_lcy1", i-rvno);
		params->x2a[i]        = phoebe_get_value_list_double ("phoebe_ld_lcx2", i-rvno);
		params->y2a[i]        = phoebe_get_value_list_double ("phoebe_ld_lcy2", i-rvno);
		params->el3[i]        = phoebe_get_value_list_double ("phoebe_el3", i-rvno);
		params->opsf[i]       = phoebe_get_value_list_double ("phoebe_opsf", i-rvno);
		params->levweight[i]  = get_level_weighting_id (phoebe_get_value_list_string ("phoebe_lc_levweight", i-rvno));
		}
	}

	/* Spot parameters:                                                         */
	params->spot1no  = phoebe_get_value_int ("phoebe_spots_no1");
	params->spot2no  = phoebe_get_value_int ("phoebe_spots_no2");

	params->spot1src   = phoebe_get_value_int  ("phoebe_dc_spot1src");
	params->spot2src   = phoebe_get_value_int  ("phoebe_dc_spot2src");
	params->spot1id    = phoebe_get_value_int  ("phoebe_dc_spot1id");
	params->spot2id    = phoebe_get_value_int  ("phoebe_dc_spot2id");
	params->spots1move = phoebe_get_value_bool ("phoebe_spots_move1");
	params->spots2move = phoebe_get_value_bool ("phoebe_spots_move2");

	params->spot1lat  = phoebe_malloc (params->spot1no * sizeof (*(params->spot1lat)));
	params->spot1long = phoebe_malloc (params->spot1no * sizeof (*(params->spot1long)));
	params->spot1rad  = phoebe_malloc (params->spot1no * sizeof (*(params->spot1rad)));
	params->spot1temp = phoebe_malloc (params->spot1no * sizeof (*(params->spot1temp)));
	params->spot2lat  = phoebe_malloc (params->spot2no * sizeof (*(params->spot2lat)));
	params->spot2long = phoebe_malloc (params->spot2no * sizeof (*(params->spot2long)));
	params->spot2rad  = phoebe_malloc (params->spot2no * sizeof (*(params->spot2rad)));
	params->spot2temp = phoebe_malloc (params->spot2no * sizeof (*(params->spot2temp)));

	for (i = 0; i < params->spot1no; i++)
		{
		params->spot1lat[i]  = phoebe_get_value_list_double ("phoebe_spots_lat1", i);
		params->spot1long[i] = phoebe_get_value_list_double ("phoebe_spots_long1", i);
		params->spot1rad[i]  = phoebe_get_value_list_double ("phoebe_spots_rad1", i);
		params->spot1temp[i] = phoebe_get_value_list_double ("phoebe_spots_temp1", i);
		}
	for (i = 0; i < params->spot2no; i++)
		{
		params->spot2lat[i]  = phoebe_get_value_list_double ("phoebe_spots_lat2", i);
		params->spot2long[i] = phoebe_get_value_list_double ("phoebe_spots_long2", i);
		params->spot2rad[i]  = phoebe_get_value_list_double ("phoebe_spots_rad2", i);
		params->spot2temp[i] = phoebe_get_value_list_double ("phoebe_spots_temp2", i);
		}

	/* Observational data:                                                      */
	{
	int indep;

	params->observations = phoebe_malloc (cno * sizeof (*(params->observations)));
	for (i = 0; i < cno; i++)
		params->observations[i] = phoebe_data_new ();

	if (params->indep == 1) indep = OUTPUT_HJD; else indep = OUTPUT_PHASE;

	if (params->rv1data)
		{
		int status;

		if (get_input_dependent_variable (phoebe_get_value_list_string ("phoebe_rv_dep", 0)) == INPUT_PRIMARY_RV) i = 0; else i = 1;

		status = read_in_experimental_data
			(
			phoebe_get_value_list_string ("phoebe_rv_filename", i),
			params->observations[i],
			get_input_independent_variable (phoebe_get_value_list_string ("phoebe_rv_indep", i)),
			indep,
			get_input_dependent_variable   (phoebe_get_value_list_string ("phoebe_rv_dep", i)),
			OUTPUT_PRIMARY_RV,
			get_input_weight               (phoebe_get_value_list_string ("phoebe_rv_indweight", i)),
			OUTPUT_STANDARD_WEIGHT
			);

		if (status < 0) return status;
		}
	if (params->rv2data)
		{
		int status;

		int index;

		if (get_input_dependent_variable (phoebe_get_value_list_string ("phoebe_rv_dep", 0)) == INPUT_SECONDARY_RV) i = 0; else i = 1;
		if (params->rv1data) index = 1; else index = 0;

		status = read_in_experimental_data
			(
			phoebe_get_value_list_string ("phoebe_rv_filename", i),
			params->observations[index],
			get_input_independent_variable (phoebe_get_value_list_string ("phoebe_rv_indep", i)),
			indep,
			get_input_dependent_variable   (phoebe_get_value_list_string ("phoebe_rv_dep", i)),
			OUTPUT_SECONDARY_RV,
			get_input_weight               (phoebe_get_value_list_string ("phoebe_rv_indweight", i)),
			OUTPUT_STANDARD_WEIGHT
			);

		if (status < 0) return status;
		}
	for (i = rvno; i < cno; i++)
		{
		int status;

		status = read_in_experimental_data
			(
			phoebe_get_value_list_string ("phoebe_lc_filename", i - rvno),
			params->observations[i],
			get_input_independent_variable (phoebe_get_value_list_string ("phoebe_lc_indep", i - rvno)),
			indep,
			get_input_dependent_variable   (phoebe_get_value_list_string ("phoebe_lc_dep", i - rvno)),
			OUTPUT_TOTAL_FLUX,
			get_input_weight               (phoebe_get_value_list_string ("phoebe_lc_indweight", i - rvno)),
			OUTPUT_STANDARD_WEIGHT
			);

		if (status < 0) return status;
		}
	}

	return marked_tba;
	}

int wd_dci_parameters_free (WD_DCI_parameters *params)
	{
	int i;

	free (params->tba);
	free (params->step);

	if (params->rv1data + params->rv2data + params->nlc != 0)
		{
		free (params->passband);
		free (params->wavelength);
		free (params->sigma);
		free (params->hla);
		free (params->cla);
		free (params->x1a);
		free (params->y1a);
		free (params->x2a);
		free (params->y2a);
		free (params->el3);
		free (params->opsf);
		free (params->levweight);
		}
	if (params->spot1no != 0)
		{
		free (params->spot1lat);
		free (params->spot1long);
		free (params->spot1rad);
		free (params->spot1temp);
		}
	if (params->spot2no != 0)
		{
		free (params->spot2lat);
		free (params->spot2long);
		free (params->spot2rad);
		free (params->spot2temp);
		}
	for (i = 0; i < params->rv1data + params->rv2data + params->nlc; i++)
		phoebe_data_free (params->observations[i]);

	free (params);
	}

int read_in_ephemeris_parameters (double *hjd0, double *period, double *dpdt, double *pshift)
	{
	/* This function speeds up the ephemeris readout:                           */

	*hjd0   = phoebe_get_value_double ("phoebe_hjd0");
	*period = phoebe_get_value_double ("phoebe_period");
	*dpdt   = phoebe_get_value_double ("phoebe_dpdt");
	*pshift = phoebe_get_value_double ("phoebe_pshift");

	return 0;
	}

int read_in_adjustable_parameters (double **values, int **indices)
	{
	/* This function cross-checks the global parameter table for all adjustable */
	/* parameters, queries the adjustment switch state, step size and the ini-  */
	/* tial parameter value; it then fills in the values to the list variable.  */
	/* The memory is allocated in this call and it is up to the user to free it */
	/* after using it. It returns the number of adjustable parameters.          */
	
	int i, j, k;
	int lcno = phoebe_get_value_int ("phoebe_lcno");

	bool calchla = phoebe_get_value_bool ("phoebe_calculate_hla");
	int calchla_idx = phoebe_index_from_qualifier ("phoebe_hla");

	bool calcvga = phoebe_get_value_bool ("phoebe_calculate_vga");
	int calcvga_idx = phoebe_index_from_qualifier ("phoebe_vga");

	bool asini    = phoebe_get_value_bool ("phoebe_asini_switch");
	int  sma_idx  = phoebe_index_from_qualifier ("phoebe_sma");

	int multi;

	*values = NULL;
	*indices = NULL;

	j = 0;
	for (i = 0; i < PHOEBE_parameters_no; i++)
		if (PHOEBE_parameters[i].kind == KIND_ADJUSTABLE && PHOEBE_parameters[i].tba == YES)
			{
			/* First, let's check whether the user tries something silly like ad-   */
			/* justing HLAs when they should be calculated:                         */
			if (i == calchla_idx && calchla == TRUE)
				{
				phoebe_warning ("light levels (HLAs) cannot be adjusted if the calculation\n");
				phoebe_warning ("switch 'phoebe_calculate_hla' is turned on. Ignoring HLAs.\n");
				continue;
				}
			if (i == calcvga_idx && calcvga == TRUE)
				{
				phoebe_warning ("gamma velocity (VGA) cannot be adjusted if the calculation\n");
				phoebe_warning ("switch 'phoebe_calculate_vga' is turned on. Ignoring VGA.\n");
				continue;
				}
			if (i == sma_idx && asini == TRUE)
				{
				bool smaadj = PHOEBE_parameters[sma_idx].tba;
				int  incidx = phoebe_index_from_qualifier ("phoebe_incl");
				bool incadj = PHOEBE_parameters[incidx].tba;

				if (smaadj == TRUE && incadj == TRUE)
					{
					phoebe_warning ("semi-major axis and inclination cannot both be adjusted\n");
					phoebe_warning ("when a sin(i) = const constraint is used. Ignoring SMA.\n");
					continue;
					}
				}

			/* Now we have to see whether it's a system parameter (one for all      */
			/* light curves) or is it a filter-dependent parameter (one for each    */
			/* light curve):                                                        */

			if (PHOEBE_parameters[i].type == TYPE_DOUBLE)
				{
				j++;
				multi = 0;
				}
			if (PHOEBE_parameters[i].type == TYPE_DOUBLE_ARRAY)
				{
				j += lcno;
				multi = 1;
				}

			/* Next, we (re)allocate memory:                                        */
			*values   = phoebe_realloc (*values,   j * sizeof (**values));
			*indices  = phoebe_realloc (*indices,  j * sizeof (**indices));

			/* Finally, we fill in all values: if multi = 0, then only (j-1). field */
			/* is updated, otherwise all filter-dependent fields are updated:       */
			for (k = j - (multi * (lcno-1)) - 1; k < j; k++)
				{
				(*indices)[k] = i;
				if (PHOEBE_parameters[i].type == TYPE_DOUBLE)
					(*values)[k] = PHOEBE_parameters[i].value.d;
				if (PHOEBE_parameters[i].type == TYPE_DOUBLE_ARRAY)
					(*values)[k] = PHOEBE_parameters[i].value.vec->val[k - (j - (multi * (lcno-1)) - 1)];
				}
			}

	return j;
	}

int get_passband_id (char *passband)
	{
	/* This function returns the passband id of the passed passband name.       */
	/*                                                                          */
	/* Return values:                                                           */
	/*   id  ..  passband id if recognition was successful                      */

	int passid = -1;

	if (strncmp (passband, "350nm (u)",     9) == 0) passid = 1;
	if (strncmp (passband, "411nm (v)",     9) == 0) passid = 2;
	if (strncmp (passband, "467nm (b)",     9) == 0) passid = 3;
	if (strncmp (passband, "547nm (y)",     9) == 0) passid = 4;
	if (strncmp (passband, "360nm (U)",     9) == 0) passid = 5;
	if (strncmp (passband, "440nm (B)",     9) == 0) passid = 6;
	if (strncmp (passband, "550nm (V)",     9) == 0) passid = 7;
	if (strncmp (passband, "700nm (R)",     9) == 0) passid = 8;
	if (strncmp (passband, "900nm (I)",     9) == 0) passid = 9;
	if (strncmp (passband, "1250nm (J)",   10) == 0) passid = 10;
	if (strncmp (passband, "2200nm (K)",   10) == 0) passid = 11;
	if (strncmp (passband, "3400nm (L)",   10) == 0) passid = 12;
	if (strncmp (passband, "5000nm (M)",   10) == 0) passid = 13;
	if (strncmp (passband, "10200nm (N)",  11) == 0) passid = 14;
	if (strncmp (passband, "647nm (Rc)",   10) == 0) passid = 15;
	if (strncmp (passband, "786nm (Ic)",   10) == 0) passid = 16;
	if (strncmp (passband, "419nm (Bt)",   10) == 0) passid = 23;
	if (strncmp (passband, "523nm (Vt)",   10) == 0) passid = 24;
	if (strncmp (passband, "505nm (Hp)",   10) == 0) passid = 25;
	if (strncmp (passband, "861nm (RVIJ)", 12) == 0)
		{
		phoebe_warning ("861nm filter not supported by WD2003, Johnson I adopted.\n");
		passid = 9;
		}
	if (passid == -1)
		{
		phoebe_warning ("passband not recognized, Johnson V filter assumed.\n");
		passid = 7;
		}

	return passid;
	}

double get_passband_wavelength (char *passband)
	{
	/* This function returns the passband wavelength of the passed filter name. */

	double wavelength;

	if (sscanf (passband, "%lf%*s", &wavelength) == 1) return wavelength;
	else
		{
		phoebe_warning ("passband name invalid, wavelength cannot be parsed; assuming 550nm.");
		return 550.0;
		}
	}

int get_level_weighting_id (char *type)
	{
	/* This function returns the id of the passed level weighting PDF.          */

	int id = -1;

	if (strcmp (type, "No level-dependent weighting") == 0) id = 0;
	if (strcmp (type, "Poissonian scatter")           == 0) id = 1;
	if (strcmp (type, "Low light scatter")            == 0) id = 2;

	if (id == -1)
		{
		phoebe_warning ("level weighting type invalid, assuming Poissonian scatter.\n");
		return 1;
		}
	return id;
	}
