#include <errno.h>
#include <locale.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <sys/stat.h>

#include "phoebe_build_config.h"
#include "phoebe_global.h"

#include "phoebe_accessories.h"
#include "phoebe_error_handling.h"
#include "phoebe_fitting.h"
#include "phoebe_keywords.h"
#include "phoebe_scripter_ast.h"

#ifdef HAVE_LIBGSL
	#ifndef PHOEBE_GSL_DISABLED
		/* This part initializes PHOEBE randomizer, but only if GSL is present.   */
		#include <gsl/gsl_rng.h>
	#endif
#endif

/* GNU readline library may be used for configuration:                        */
#ifdef HAVE_LIBREADLINE
	#ifndef PHOEBE_READLINE_DISABLED
		#include <readline/readline.h>
		#include <readline/history.h>
	#endif
#endif

void parse_startup_line (int argc, char *argv[])
	{
	/* This function parses the command line and looks for known switches; it   */
	/* doesn't really do anything with them, it just writes what it found to    */
	/* the PHOEBE_args variable that is supposed to be used for real actions.   */

	int i;

	for (i = 1; i < argc; i++)
		{
		if ( (strcmp (argv[i],  "-h"   ) == 0) ||
		     (strcmp (argv[i],  "-?"   ) == 0) ||
				 (strcmp (argv[i], "--help") == 0) )
			{
			printf ("\nPHOEBE %s command line arguments: [-hsv] [keyword_file]\n\n", PHOEBE_VERSION_NUMBER);
			printf ("  -c, --configure     ..  configure PHOEBE\n");
			printf ("  -e, --execute       ..  execute PHOEBE script\n");
			printf ("  -h, --help, -?      ..  this help screen\n");
			printf ("  -s, --scripter      ..  Run PHOEBE scripter interactively\n");
			printf ("  -v, --version       ..  display PHOEBE version and exit\n");
			printf ("\n");
			exit (0);
			}

		if ( (strcmp (argv[i],  "-v"      ) == 0) ||
				 (strcmp (argv[i], "--version") == 0) )
			{
			printf ("\n%s, %s by Andrej Prsa\n", PHOEBE_VERSION_NUMBER, PHOEBE_VERSION_DATE);
			printf ("  Send comments and/or requests to andrej.prsa@fmf.uni-lj.si\n\n");
			exit (0);
			}

		if ( (strcmp (argv[i],  "-c"        ) == 0) ||
				 (strcmp (argv[i], "--configure") == 0) )
			{
			PHOEBE_args.CONFIGURE_SWITCH = 1;
			}

		if ( (strcmp (argv[i],  "-e"      ) == 0) ||
				 (strcmp (argv[i], "--execute") == 0) )
			{
			/* We cannot start the script from here because the widgets aren't ini- */
			/* tialized yet; thus we just initialize the proper switches:           */
			PHOEBE_args.SCRIPT_SWITCH = 1;
			
			/* If someone forgot to put the script name, we should warn him and     */
			/* exit:                                                                */
			if ( (i+1 == argc) || (argv[i+1][0] == '-') )
				{
				printf ("\n%s command line arguments: [-hsv] [keyword_file]\n\n", PHOEBE_VERSION_NUMBER);
				printf ("  -e, --execute       ..  execute PHOEBE script\n");
				printf ("  -h, --help, -?      ..  this help screen\n");
				printf ("  -s, --scripter      ..  Run PHOEBE scripter interactively\n");
				printf ("  -v, --version       ..  display PHOEBE version and exit\n");
				printf ("\n");
				phoebe_fatal ("there is no argument given to the '-e' switch.\n\n");
				exit (0);
				}

			/* Otherwise let's read in the script's name:                           */
			PHOEBE_args.SCRIPT_NAME = strdup (argv[i+1]);
			i++; i++;                            /* This will skip the script name. */
			if (i >= argc) break;   /* If this was the last switch, break the loop. */
			}

		if ( (strcmp (argv[i],  "-s"       ) == 0) ||
				 (strcmp (argv[i], "--scripter") == 0) )
			PHOEBE_args.INTERACTIVE_SCRIPTER_SWITCH = 1;
		else
			PHOEBE_args.INTERACTIVE_SCRIPTER_SWITCH = 0;

		if ( argv[i][0] != '-' )
			{
			/* This means that the command line argument doesn't contain '-'; thus  */
			/* it should be a keyword file. All other arguments to - and -- swit-   */
			/* ches (without -/--) are read in before and they should never evalua- */
			/* te here.                                                             */

			PHOEBE_args.KEYWORD_SWITCH = 1;
			PHOEBE_args.KEYWORD = strdup (argv[i]);
			}
		}
	}

int variables_init ()
	{
	/* This function initializes all global literal strings, so that they may   */
	/* be used fearlessly by other functions.                                   */
	
	/* Let's declare a global keyword filename string to "Undefined". This is   */
	/* needed for command line keyword filenames' incorporation.                */
	PHOEBE_KEYWORD_FILENAME = strdup ("Undefined");

	/* The following are global parameter variables which are identified by the */
	/* widget, qualifier, keyword and description. Since they will be dynamica- */
	/* lly stored by realloc, we need to set it to NULL.                        */
	PHOEBE_parameters_no = 0;
	PHOEBE_parameters    = NULL;

	/* Much the same as above, we initialize command line argument strings:     */
	PHOEBE_args.KEYWORD          = NULL;
	PHOEBE_args.SCRIPT_NAME      = NULL;
	}

int phoebe_init ()
	{
	/* This function initializes all core parameters.                           */

	char working_string[255];
	char *working_str = working_string;

	char keyword_string[255];
	char *keyword_str = keyword_string;

	FILE *config_file;

	/* Assign a current directory (i.e. the directory from which PHOEBE was     */
	/* started to PHOEBE_STARTUP_DIR; this is used for resolving relative       */
	/* pathnames.                                                               */
	getcwd (working_str, 255);
	PHOEBE_STARTUP_DIR = strdup (working_str);

	/* Get the current locale decimal point, store it to the global variable,   */
	/* change it to "C" and change it back to original upon exit.               */
	PHOEBE_INPUT_LOCALE = strdup (setlocale (LC_NUMERIC, NULL));
	setlocale (LC_NUMERIC, "C");

	if (getenv ("HOME") == NULL)
		phoebe_fatal ("Environment variable $HOME is not defined. PHOEBE needs this variable\n              to install its configuration files. Please define it with e.g.:\n\n                export HOME=/home/user\n\n              and restart PHOEBE.\n");
	USER_HOME_DIR = strdup (getenv ("HOME"));

	/* Although it sounds silly, let's check full permissions of the home di-   */
	/* rectory:                                                                 */
	if (!filename_has_full_permissions (USER_HOME_DIR))
		phoebe_fatal ("Your home directory (defined by $HOME variable) doesn't exist or it doesn't\n              have proper permissions set (e.g. 755). Please correct this and restart\n              PHOEBE.\n");

	/* Everything OK, let's initialize PHOEBE environment file:                 */
	sprintf (working_str, "%s/.phoebe2", USER_HOME_DIR);
	PHOEBE_HOME_DIR = strdup (working_str);
	sprintf (working_str, "%s/phoebe.config", PHOEBE_HOME_DIR);
	PHOEBE_CONFIG = strdup (working_str);

	/* Let's presume there's no config file. If there is, we'll set it to 1:    */
	PHOEBE_CONFIG_EXISTS = 0;

	/* Read out the configuration from the config file.                         */
	if (filename_exists (PHOEBE_CONFIG))
		{
		PHOEBE_CONFIG_EXISTS = 1;
		config_file = fopen (PHOEBE_CONFIG, "r");

		while (!feof (config_file))
			{
			/* The following line reads the line from the input file and checks if  */
			/* everything went OK; if en error occured (e.g. EoF reached), it bre-  */
			/* aks the loop.                                                        */

			if (fgets (keyword_str, 254, config_file) == NULL) break;

			if (strstr (keyword_str, "PHOEBE_BASE_DIR") != NULL)
				{
				if (sscanf (keyword_str, "PHOEBE_BASE_DIR %s", working_str) != 1) PHOEBE_BASE_DIR = strdup ("");
				else PHOEBE_BASE_DIR = strdup (working_str);
				}
			if (strstr (keyword_str, "PHOEBE_SOURCE_DIR") != NULL)
				{
				if (sscanf (keyword_str, "PHOEBE_SOURCE_DIR %s", working_str) != 1) PHOEBE_SOURCE_DIR = strdup ("");
				else PHOEBE_SOURCE_DIR = strdup (working_str);
				}
			if (strstr (keyword_str, "PHOEBE_DEFAULTS_DIR") != NULL)
				{
				if (sscanf (keyword_str, "PHOEBE_DEFAULTS_DIR %s", working_str) != 1) PHOEBE_DEFAULTS_DIR = strdup ("");
				else PHOEBE_DEFAULTS_DIR = strdup (working_str);
				}
			if (strstr (keyword_str, "PHOEBE_TEMP_DIR") != NULL)
				{
				if (sscanf (keyword_str, "PHOEBE_TEMP_DIR %s", working_str) != 1) PHOEBE_TEMP_DIR = strdup ("");
				else PHOEBE_TEMP_DIR = strdup (working_str);
				}
			if (strstr (keyword_str, "PHOEBE_DATA_DIR") != NULL)
				{
				if (sscanf (keyword_str, "PHOEBE_DATA_DIR %s", working_str) != 1) PHOEBE_DATA_DIR = strdup ("");
				else PHOEBE_DATA_DIR = strdup (working_str);
				}
			if (strstr (keyword_str, "PHOEBE_FF_DIR") != NULL)
				{
				if (sscanf (keyword_str, "PHOEBE_FF_DIR %s", working_str) != 1) PHOEBE_FF_DIR = strdup ("");
				else PHOEBE_FF_DIR = strdup (working_str);
				}
			if (strstr (keyword_str, "PHOEBE_PLOTTING_PACKAGE") != NULL)
				{
				if (sscanf (keyword_str, "PHOEBE_PLOTTING_PACKAGE %s", working_str) != 1) PHOEBE_PLOTTING_PACKAGE = strdup ("");
				else PHOEBE_PLOTTING_PACKAGE = strdup (working_str);
				}
			if (strstr (keyword_str, "PHOEBE_LD_SWITCH") != NULL)
				if (sscanf (keyword_str, "PHOEBE_LD_SWITCH %d", &PHOEBE_LD_SWITCH) != 1) PHOEBE_LD_SWITCH = 0;
			if (strstr (keyword_str, "PHOEBE_LD_DIR") != NULL)
				{
				if (sscanf (keyword_str, "PHOEBE_LD_DIR %s", working_str) != 1) PHOEBE_LD_DIR = strdup ("");
				else PHOEBE_LD_DIR = strdup (working_str);
				}
			if (strstr (keyword_str, "PHOEBE_KURUCZ_SWITCH") != NULL)
				if (sscanf (keyword_str, "PHOEBE_KURUCZ_SWITCH %d", &PHOEBE_KURUCZ_SWITCH) != 1) PHOEBE_KURUCZ_SWITCH = 0;
			if (strstr (keyword_str, "PHOEBE_KURUCZ_DIR") != NULL)
				{
				if (sscanf (keyword_str, "PHOEBE_KURUCZ_DIR %s", working_str) != 1) PHOEBE_KURUCZ_DIR = strdup ("");
				else PHOEBE_KURUCZ_DIR = strdup (working_str);
				}
			if (strstr (keyword_str, "PHOEBE_3D_PLOT_CALLBACK_OPTION") != NULL)
				if (sscanf (keyword_str, "PHOEBE_3D_PLOT_CALLBACK_OPTION %d", &PHOEBE_3D_PLOT_CALLBACK_OPTION) != 1) PHOEBE_3D_PLOT_CALLBACK_OPTION = 0;
			if (strstr (keyword_str, "PHOEBE_CONFIRM_ON_SAVE") != NULL)
				if (sscanf (keyword_str, "PHOEBE_CONFIRM_ON_SAVE %d", &PHOEBE_CONFIRM_ON_SAVE) != 1) PHOEBE_CONFIRM_ON_SAVE = 1;
			if (strstr (keyword_str, "PHOEBE_CONFIRM_ON_QUIT") != NULL)
				if (sscanf (keyword_str, "PHOEBE_CONFIRM_ON_QUIT %d", &PHOEBE_CONFIRM_ON_QUIT) != 1) PHOEBE_CONFIRM_ON_QUIT = 1;
			if (strstr (keyword_str, "PHOEBE_WARN_ON_SYNTHETIC_SCATTER") != NULL)
				if (sscanf (keyword_str, "PHOEBE_WARN_ON_SYNTHETIC_SCATTER %d", &PHOEBE_WARN_ON_SYNTHETIC_SCATTER) != 1) PHOEBE_WARN_ON_SYNTHETIC_SCATTER = 1;
			}
		fclose (config_file);
		}

	if (!filename_exists (PHOEBE_CONFIG) || PHOEBE_args.CONFIGURE_SWITCH == 1)
		{
		create_configuration_file ();
		}

	/* All PHOEBE parameters are referenced globally, so we have to initialize  */
	/* them before they could be used.                                          */
	declare_all_parameters ();

	/* Choose a randomizer seed:                                                */
	srand (time (0));

	/* Allocate memory for Nelder & Mead's Simplex options struct:              */
	PHOEBE_nms_params = phoebe_malloc (sizeof (*PHOEBE_nms_params));

	/* Initialize PHOEBE randomizer:                                            */
	#ifdef HAVE_LIBGSL
		#ifndef PHOEBE_GSL_DISABLED
			PHOEBE_randomizer = gsl_rng_alloc (gsl_rng_mt19937);
			gsl_rng_set (PHOEBE_randomizer, time (0));
		#endif
	#endif

	/* Open stdout for writing:                                                 */
	PHOEBE_output = stdout;
	}

int create_configuration_file ()
	{
	int errcode;
	char answer;
	char *dir;

	printf ("\nWelcome to %s!\n\n", PHOEBE_VERSION_NUMBER);
	printf ("I am about to create a top-level configuration directory:\n\n\t%s\n\n", PHOEBE_HOME_DIR);
	printf ("and guide you through the setup process. Do you wish to continue? [Y/n] ");
	answer = fgetc (stdin);

	if (answer != 'y' && answer != 'Y' && answer != '\n')
		{
		printf ("\n");
		exit (0);
		}

	errcode = mkdir (PHOEBE_HOME_DIR, 0755);
	if (errcode == -1)
		{
		if (errno == EEXIST)
			printf ("\nDirectory %s already exists, proceeding to configuration.\n\n", PHOEBE_HOME_DIR);
		else if (errno == EACCES)
			{
			printf ("\nYou don't have permissions to create %s.\n", PHOEBE_HOME_DIR);
			printf ("Please enable write access to your home directory and restart PHOEBE.\n\n");
			exit (-1);
			}
		else if (errno == ENOSPC)
			{
			printf ("\nYour hard disk is out of space. Please make some room and try again.\n\n");
			exit (-1);
			}
		else
			{
			printf ("\nAn undefined error has occured while creating configuration directory.\n");
			printf ("Please try to resolve the issue and restart PHOEBE.\n\n");
			exit (-1);
			}
		}
	else
		printf ("\nPHOEBE configuration directory created.\n\n");

	/* If libreadline is present, use it:                                       */
	#ifdef HAVE_LIBREADLINE
		#ifndef PHOEBE_READLINE_DISABLED
			{
			char *basedir, *srcdir, *defdir, *tempdir, *datadir, *lddir;
			char *ldswitch, *yn;
			char prompt[255], defaultdir[255];

			printf ("Please supply names of directories to be used by PHOEBE:\n");
			printf ("  note that you may use tab-completion features to access directories;\n");
			printf ("  if you are happy with defaults enclosed in [...], just press ENTER.\n\n");

			/*************************** BASE DIRECTORY *****************************/

			if (PHOEBE_CONFIG_EXISTS == 1 && PHOEBE_BASE_DIR != NULL)
				{
				sprintf (prompt, "PHOEBE base directory      [%s]: ", PHOEBE_BASE_DIR);
				basedir = readline (prompt);
				if (strcmp (basedir, "") == 0) basedir = strdup (PHOEBE_BASE_DIR);
				}
			else
				{
				sprintf (prompt, "PHOEBE base directory      [/usr/local/share/phoebe2]: ");
				basedir = readline (prompt);
				if (strcmp (basedir, "") == 0) basedir = strdup ("/usr/local/share/phoebe2");
				}
			if (basedir[strlen(basedir)-1] == '/') basedir[strlen(basedir)-1] = '\0';

			/************************** SOURCE DIRECTORY ****************************/

			if (PHOEBE_CONFIG_EXISTS == 1 && PHOEBE_SOURCE_DIR != NULL)
				{
				sprintf (prompt, "PHOEBE source directory    [%s]: ", PHOEBE_SOURCE_DIR);
				srcdir = readline (prompt);
				if (strcmp (srcdir, "") == 0) srcdir = strdup (PHOEBE_SOURCE_DIR);
				}
			else
				{
				sprintf (prompt, "PHOEBE source directory    [%s/src]: ", basedir);
				srcdir = readline (prompt);
				if (strcmp (srcdir, "") == 0)
					{
					sprintf (defaultdir, "%s/src", basedir);
					srcdir = strdup (defaultdir);
					}
				}
			if (srcdir[strlen(srcdir)-1] == '/') srcdir[strlen(srcdir)-1] = '\0';

			/************************* DEFAULTS DIRECTORY ***************************/

			if (PHOEBE_CONFIG_EXISTS == 1 && PHOEBE_DEFAULTS_DIR != NULL)
				{
				sprintf (prompt, "PHOEBE defaults directory  [%s]: ", PHOEBE_DEFAULTS_DIR);
				defdir = readline (prompt);
				if (strcmp (defdir, "") == 0) defdir = strdup (PHOEBE_DEFAULTS_DIR);
				}
			else
				{
				sprintf (prompt, "PHOEBE defaults directory  [%s/defaults]: ", basedir);
				defdir = readline (prompt);
				if (strcmp (defdir, "") == 0)
					{
					sprintf (defaultdir, "%s/defaults", basedir);
					defdir = strdup (defaultdir);
					}
				}
			if (defdir[strlen(defdir)-1] == '/') defdir[strlen(defdir)-1] = '\0';

			/*************************** TEMP DIRECTORY *****************************/

			if (PHOEBE_CONFIG_EXISTS == 1 && PHOEBE_TEMP_DIR != NULL)
				{
				sprintf (prompt, "PHOEBE temporary directory [%s]: ", PHOEBE_TEMP_DIR);
				tempdir = readline (prompt);
				if (strcmp (tempdir, "") == 0) tempdir = strdup (PHOEBE_TEMP_DIR);
				}
			else
				{
				sprintf (prompt, "PHOEBE temporary directory [/tmp]: ");
				tempdir = readline (prompt);
				if (strcmp (tempdir, "") == 0) tempdir = strdup ("/tmp");
				}
			if (tempdir[strlen(tempdir)-1] == '/') tempdir[strlen(tempdir)-1] = '\0';

			/*************************** DATA DIRECTORY *****************************/

			if (PHOEBE_CONFIG_EXISTS == 1 && PHOEBE_DATA_DIR != NULL)
				{
				sprintf (prompt, "PHOEBE data directory      [%s]: ", PHOEBE_DATA_DIR);
				datadir = readline (prompt);
				if (strcmp (datadir, "") == 0) datadir = strdup (PHOEBE_DATA_DIR);
				}
			else
				{
				sprintf (prompt, "PHOEBE data directory      [%s/phoebe/data]: ", USER_HOME_DIR);
				datadir = readline (prompt);
				if (strcmp (datadir, "") == 0)
					{
					sprintf (defaultdir, "%s/phoebe/data", USER_HOME_DIR);
					datadir = strdup (defaultdir);
					}
				}
			if (datadir[strlen(datadir)-1] == '/') datadir[strlen(datadir)-1] = '\0';

			/****************************** LD TABLES *******************************/

			ldswitch = readline ("Are Van Hamme (1993) limb darkening tables present on your system [y/N]? ");
			if (strcmp (ldswitch, "") == 0 || ldswitch[0] == 'n' || ldswitch[0] == 'N') ;
			else
				{
				if (PHOEBE_CONFIG_EXISTS == 1 && PHOEBE_LD_DIR != NULL)
					{
					sprintf (prompt, "PHOEBE LD directory        [%s]: ", PHOEBE_LD_DIR);
					lddir = readline (prompt);
					if (strcmp (lddir, "") == 0) lddir = strdup (PHOEBE_LD_DIR);
					}
				else
					{
					sprintf (prompt, "PHOEBE LD directory        [%s/catalogs/ld]: ", USER_HOME_DIR);
					lddir = readline (prompt);
					if (strcmp (lddir, "") == 0)
						{
						sprintf (defaultdir, "%s/catalogs/ld", USER_HOME_DIR);
						lddir = strdup (defaultdir);
						}
					}
				if (lddir[strlen(lddir)-1] == '/') lddir[strlen(lddir)-1] = '\0';
				}

			/************************************************************************/

			printf ("\nConfiguration summary:\n----------------------\n");
			printf ("PHOEBE base directory:       %s\n", basedir);
			printf ("PHOEBE source directory:     %s\n", srcdir);
			printf ("PHOEBE defaults directory:   %s\n", defdir);
			printf ("PHOEBE temporary directory:  %s\n", tempdir);
			printf ("PHOEBE data directory:       %s\n", datadir);

			if (strcmp (ldswitch, "") == 0 || ldswitch[0] == 'n' || ldswitch[0] == 'N')
				printf ("Limb darkening tables:       not present\n");
			else
				{
				printf ("Limb darkening tables:       present\n");
				printf ("PHOEBE LD directory:         %s\n", lddir);
				}

			yn = readline ("\nAre these settings ok [Y/n]? ");
			if (strcmp (yn, "") == 0 || yn[0] == 'y' || yn[0] == 'Y')
				{
				FILE *config = fopen (PHOEBE_CONFIG, "w");
				fprintf (config, "PHOEBE_BASE_DIR\t\t%s\n", basedir);
				fprintf (config, "PHOEBE_SOURCE_DIR\t%s\n", srcdir);
				fprintf (config, "PHOEBE_DEFAULTS_DIR\t%s\n", defdir);
				fprintf (config, "PHOEBE_TEMP_DIR\t\t%s\n", tempdir);
				fprintf (config, "PHOEBE_DATA_DIR\t\t%s\n", datadir);

				if (PHOEBE_BASE_DIR     != NULL) free (PHOEBE_BASE_DIR);
				if (PHOEBE_SOURCE_DIR   != NULL) free (PHOEBE_SOURCE_DIR);
				if (PHOEBE_DEFAULTS_DIR != NULL) free (PHOEBE_DEFAULTS_DIR);
				if (PHOEBE_TEMP_DIR     != NULL) free (PHOEBE_TEMP_DIR);
				if (PHOEBE_DATA_DIR     != NULL) free (PHOEBE_DATA_DIR);
				if (PHOEBE_LD_DIR       != NULL) free (PHOEBE_LD_DIR);

				PHOEBE_BASE_DIR     = strdup (basedir);
				PHOEBE_SOURCE_DIR   = strdup (srcdir);
				PHOEBE_DEFAULTS_DIR = strdup (defdir);
				PHOEBE_TEMP_DIR     = strdup (tempdir);
				PHOEBE_DATA_DIR     = strdup (datadir);

				if (strcmp (ldswitch, "") == 0 || ldswitch[0] == 'n' || ldswitch[0] == 'N')
					{
					fprintf (config, "PHOEBE_LD_SWITCH\t0\n");
					fprintf (config, "PHOEBE_LD_DIR\t\t\n");
					PHOEBE_LD_SWITCH = 0;
					PHOEBE_LD_DIR = strdup ("");
					}
				else
					{
					fprintf (config, "PHOEBE_LD_SWITCH\t1\n");
					fprintf (config, "PHOEBE_LD_DIR\t\t%s\n", lddir);
					PHOEBE_LD_SWITCH = 1;
					PHOEBE_LD_DIR = strdup (lddir);
					}
				fclose (config);


				printf ("\nConfiguration file written. If you ever want to change your settings,\n");
				printf ("simply edit %s/phoebe.config or start PHOEBE\n", PHOEBE_HOME_DIR);
				printf ("with -c (--configure) switch.\n");
				return 0;
				}
			else
				{
				printf ("\nAborting configuration. Please restart PHOEBE to try again.\n\n");
				exit (0);
				}
			}
		#endif
	#endif

		/* This part is executed if GNU readline isn't found:                     */
			{
			char basedir[255], srcdir[255], defdir[255], tempdir[255], datadir[255], lddir[255];
			char ldswitch[255], yn[255];

			printf ("Please supply names of directories to be used by PHOEBE:\n");
			printf ("  if you are happy with defaults enclosed in [...], just press ENTER.\n\n");

			/*************************** BASE DIRECTORY *****************************/

			if (PHOEBE_CONFIG_EXISTS == 1 && PHOEBE_BASE_DIR != NULL)
				{
				printf ("PHOEBE base directory      [%s]: ", PHOEBE_BASE_DIR);
				fgets (basedir, 255, stdin); basedir[strlen(basedir)-1] = '\0';
				if (strcmp (basedir, "") == 0) sprintf (basedir, "%s", PHOEBE_BASE_DIR);
				}
			else
				{
				printf ("PHOEBE base directory      [/usr/local/share/phoebe2]: ");
				fgets (basedir, 255, stdin); basedir[strlen(basedir)-1] = '\0';
				if (strcmp (basedir, "") == 0) sprintf (basedir, "/usr/local/share/phoebe2");
				}
			if (basedir[strlen(basedir)-1] == '/') basedir[strlen(basedir)-1] = '\0';

			/************************** SOURCE DIRECTORY ****************************/

			if (PHOEBE_CONFIG_EXISTS == 1 && PHOEBE_SOURCE_DIR != NULL)
				{
				printf ("PHOEBE source directory    [%s]: ", PHOEBE_SOURCE_DIR);
				fgets (srcdir, 255, stdin); srcdir[strlen(srcdir)-1] = '\0';
				if (strcmp (srcdir, "") == 0) sprintf (srcdir, "%s", PHOEBE_SOURCE_DIR);
				}
			else
				{
				printf ("PHOEBE source directory    [%s/src]: ", basedir);
				fgets (srcdir, 255, stdin); srcdir[strlen(srcdir)-1] = '\0';
				if (strcmp (srcdir, "") == 0) sprintf (srcdir, "%s/src", basedir);
				}
			if (srcdir[strlen(srcdir)-1] == '/') srcdir[strlen(srcdir)-1] = '\0';

			/************************* DEFAULTS DIRECTORY ***************************/

			if (PHOEBE_CONFIG_EXISTS == 1 && PHOEBE_DEFAULTS_DIR != NULL)
				{
				printf ("PHOEBE defaults directory  [%s]: ", PHOEBE_DEFAULTS_DIR);
				fgets (defdir, 255, stdin); defdir[strlen(defdir)-1] = '\0';
				if (strcmp (defdir, "") == 0) sprintf (defdir, "%s", PHOEBE_DEFAULTS_DIR);
				}
			else
				{
				printf ("PHOEBE defaults directory  [%s/defaults]: ", basedir);
				fgets (defdir, 255, stdin); defdir[strlen(defdir)-1] = '\0';
				if (strcmp (defdir, "") == 0) sprintf (defdir, "%s/defaults", basedir);
				}
			if (defdir[strlen(defdir)-1] == '/') defdir[strlen(defdir)-1] = '\0';

			/*************************** TEMP DIRECTORY *****************************/

			if (PHOEBE_CONFIG_EXISTS == 1 && PHOEBE_TEMP_DIR != NULL)
				{
				printf ("PHOEBE temporary directory [%s]: ", PHOEBE_TEMP_DIR);
				fgets (tempdir, 255, stdin); tempdir[strlen(tempdir)-1] = '\0';
				if (strcmp (tempdir, "") == 0) sprintf (tempdir, "%s", PHOEBE_TEMP_DIR);
				}
			else
				{
				printf ("PHOEBE temporary directory [/tmp]: ");
				fgets (tempdir, 255, stdin); tempdir[strlen(tempdir)-1] = '\0';
				if (strcmp (tempdir, "") == 0) sprintf (tempdir, "/tmp");
				}
			if (tempdir[strlen(tempdir)-1] == '/') tempdir[strlen(tempdir)-1] = '\0';

			/*************************** DATA DIRECTORY *****************************/

			if (PHOEBE_CONFIG_EXISTS == 1 && PHOEBE_DATA_DIR != NULL)
				{
				printf ("PHOEBE data directory      [%s]: ", PHOEBE_DATA_DIR);
				fgets (datadir, 255, stdin); datadir[strlen(datadir)-1] = '\0';
				if (strcmp (datadir, "") == 0) sprintf (datadir, "%s", PHOEBE_DATA_DIR);
				}
			else
				{
				printf ("PHOEBE data directory      [%s/data]: ", basedir);
				fgets (datadir, 255, stdin); datadir[strlen(datadir)-1] = '\0';
				if (strcmp (datadir, "") == 0) sprintf (datadir, "%s/data", basedir);
				}
			if (datadir[strlen(datadir)-1] == '/') datadir[strlen(datadir)-1] = '\0';

			/****************************** LD TABLES *******************************/

			printf ("Are Van Hamme (1993) limb darkening tables present on your system [y/N]? ");
			fgets (ldswitch, 255, stdin); ldswitch[strlen(ldswitch)-1] = '\0';

			if (strcmp (ldswitch, "") == 0 || ldswitch[0] == 'n' || ldswitch[0] == 'N') ;
			else
				{
				if (PHOEBE_CONFIG_EXISTS == 1 && PHOEBE_LD_DIR != NULL)
					{
					printf ("PHOEBE LD directory        [%s]: ", PHOEBE_LD_DIR);
					fgets (lddir, 255, stdin); lddir[strlen(lddir)-1] = '\0';
					if (strcmp (lddir, "") == 0) sprintf (lddir, "%s", PHOEBE_LD_DIR);
					}
				else
					{
					printf ("PHOEBE LD directory        [%s/catalogs/ld]: ", USER_HOME_DIR);
					fgets (lddir, 255, stdin); lddir[strlen(lddir)-1] = '\0';
					if (strcmp (lddir, "") == 0) sprintf (lddir, "%s/catalogs/ld", USER_HOME_DIR);
					}
				if (lddir[strlen(lddir)-1] == '/') lddir[strlen(lddir)-1] = '\0';
				}

			/************************************************************************/

			printf ("\nConfiguration summary:\n----------------------\n");
			printf ("PHOEBE base directory:       %s\n", basedir);
			printf ("PHOEBE source directory:     %s\n", srcdir);
			printf ("PHOEBE defaults directory:   %s\n", defdir);
			printf ("PHOEBE temporary directory:  %s\n", tempdir);
			printf ("PHOEBE data directory:       %s\n", datadir);

			if (strcmp (ldswitch, "") == 0 || ldswitch[0] == 'n' || ldswitch[0] == 'N')
				printf ("Limb darkening tables:       not present\n");
			else
				{
				printf ("Limb darkening tables:       present\n");
				printf ("PHOEBE LD directory:         %s\n", lddir);
				}

			printf ("\nAre these settings ok [Y/n]? ");
			fgets (yn, 255, stdin); yn[strlen(yn)-1] = '\0';

			if (strcmp (yn, "") == 0 || yn[0] == 'y' || yn[0] == 'Y')
				{
				FILE *config = fopen (PHOEBE_CONFIG, "w");
				fprintf (config, "PHOEBE_BASE_DIR\t\t%s\n", basedir);
				fprintf (config, "PHOEBE_SOURCE_DIR\t%s\n", srcdir);
				fprintf (config, "PHOEBE_DEFAULTS_DIR\t%s\n", defdir);
				fprintf (config, "PHOEBE_TEMP_DIR\t\t%s\n", tempdir);
				fprintf (config, "PHOEBE_DATA_DIR\t\t%s\n", datadir);

				if (PHOEBE_BASE_DIR     != NULL) free (PHOEBE_BASE_DIR);
				if (PHOEBE_SOURCE_DIR   != NULL) free (PHOEBE_SOURCE_DIR);
				if (PHOEBE_DEFAULTS_DIR != NULL) free (PHOEBE_DEFAULTS_DIR);
				if (PHOEBE_TEMP_DIR     != NULL) free (PHOEBE_TEMP_DIR);
				if (PHOEBE_DATA_DIR     != NULL) free (PHOEBE_DATA_DIR);
				if (PHOEBE_LD_DIR       != NULL) free (PHOEBE_LD_DIR);

				PHOEBE_BASE_DIR     = strdup (basedir);
				PHOEBE_SOURCE_DIR   = strdup (srcdir);
				PHOEBE_DEFAULTS_DIR = strdup (defdir);
				PHOEBE_TEMP_DIR     = strdup (tempdir);
				PHOEBE_DATA_DIR     = strdup (datadir);

				if (strcmp (ldswitch, "") == 0 || ldswitch[0] == 'n' || ldswitch[0] == 'N')
					{
					fprintf (config, "PHOEBE_LD_SWITCH\t0\n");
					fprintf (config, "PHOEBE_LD_DIR\t\t\n");
					PHOEBE_LD_SWITCH = 0;
					PHOEBE_LD_DIR = strdup ("");
					}
				else
					{
					fprintf (config, "PHOEBE_LD_SWITCH\t1\n");
					fprintf (config, "PHOEBE_LD_DIR\t\t%s\n", lddir);
					PHOEBE_LD_SWITCH = 1;
					PHOEBE_LD_DIR = strdup (lddir);
					}
				fclose (config);

				printf ("\nConfiguration file written. If you ever want to change your settings,\n");
				printf ("simply edit %s/phoebe.config or start PHOEBE\n", PHOEBE_HOME_DIR);
				printf ("with -c (--configure) switch.\n");
				return 0;
				}
			else
				{
				printf ("\nAborting configuration. Please restart PHOEBE to try again.\n\n");
				exit (0);
				}
			}

	}
