#include <math.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "phoebe_build_config.h"

#include "phoebe_accessories.h"
#include "phoebe_connectivity.h"
#include "phoebe_error_handling.h"
#include "phoebe_global.h"
#include "phoebe_keywords.h"
#include "phoebe_type_handling.h"

void declare_parameter (char *qualifier, char *dependency, char *keyword, char *description, PHOEBE_parameter_kind kind, double min, double max, double step, bool tba, ...)
	{
	/* This function adds an empty record to the PHOEBE_parameter struct (defi- */
	/* ned in phoebe_global.h dynamically and fills it in with values given as  */
	/* arguments to this function. It should be called *only* by declare_all_   */
	/* _parameters () function.                                                 */
	/*                                                                          */
	/* This function handles two types of parameters - native types (all ints,  */
	/* doubles, char *s, ...) and parameter arrays. In case of the latter, no   */
	/* initialization is performed, since the array sizes depend on particular  */
	/* dependency parameters (e.g. LC number, ...).                             */

	va_list args;
	int i = PHOEBE_parameters_no;

	PHOEBE_parameters_no++;
	PHOEBE_parameters = realloc (PHOEBE_parameters, PHOEBE_parameters_no * (sizeof (*PHOEBE_parameters)));

	PHOEBE_parameters[i].qualifier   = strdup (qualifier);
	PHOEBE_parameters[i].dependency  = strdup (dependency);
	PHOEBE_parameters[i].keyword     = strdup (keyword);
	PHOEBE_parameters[i].description = strdup (description);
	PHOEBE_parameters[i].kind        = kind;
	PHOEBE_parameters[i].min         = min;
	PHOEBE_parameters[i].max         = max;
	PHOEBE_parameters[i].step        = step;
	PHOEBE_parameters[i].tba         = tba;

	va_start (args, tba);
	PHOEBE_parameters[i].type        = va_arg (args, PHOEBE_type);

	switch (PHOEBE_parameters[i].type)
		{
		case TYPE_INT:
			PHOEBE_parameters[i].value.i = va_arg (args, int);
		break;
		case TYPE_DOUBLE:
			PHOEBE_parameters[i].value.d = va_arg (args, double);
		break;
		case TYPE_BOOL:
			PHOEBE_parameters[i].value.b = va_arg (args, bool);
		break;
		case TYPE_STRING:
			{
			char *str = va_arg (args, char *);
			PHOEBE_parameters[i].value.str = phoebe_malloc (strlen (str) + 1);
			strcpy (PHOEBE_parameters[i].value.str, str);
			}
		break;
		case TYPE_INT_ARRAY:
			PHOEBE_parameters[i].value.iarray = NULL;
			PHOEBE_parameters[i].defaultvalue.i = va_arg (args, int);
		break;
		case TYPE_DOUBLE_ARRAY:
			PHOEBE_parameters[i].value.vec = phoebe_vector_new ();
			PHOEBE_parameters[i].defaultvalue.d = va_arg (args, double);
		break;
		case TYPE_BOOL_ARRAY:
			PHOEBE_parameters[i].value.barray = NULL;
			PHOEBE_parameters[i].defaultvalue.b = va_arg (args, bool);
		break;
		case TYPE_STRING_ARRAY:
			{
			char *str = va_arg (args, char *);
			PHOEBE_parameters[i].defaultvalue.str = strdup (str);
			}
		break;
		}
	va_end (args);
	}

void update_parameter_arrays (char *dependency, int oldval)
	{
	/* This function is called whenever the dimension of parameter arrays must  */
	/* be changed. Typically this happens when the number of observed data cur- */
	/* ves is changed, the number of spots is changed etc.                      */

	int dim = PHOEBE_parameters[phoebe_index_from_qualifier(dependency)].value.i;
	int i, j;
	PHOEBE_type type;

	if (oldval == dim) return;

	for (i = 0; i < PHOEBE_parameters_no; i++)
		if (strcmp (PHOEBE_parameters[i].dependency, dependency) == 0)
			{
			phoebe_debug ("resizing array %s to dim %d\n", PHOEBE_parameters[i].qualifier, dim);
			type = PHOEBE_parameters[i].type;
			switch (type)
				{
				case TYPE_INT_ARRAY:
					PHOEBE_parameters[i].value.iarray = phoebe_realloc (PHOEBE_parameters[i].value.iarray, dim * sizeof (int));
					for (j = oldval; j < dim; j++)
						PHOEBE_parameters[i].value.iarray[j] = PHOEBE_parameters[i].defaultvalue.i;
				break;
				case TYPE_BOOL_ARRAY:
					PHOEBE_parameters[i].value.barray = phoebe_realloc (PHOEBE_parameters[i].value.barray, dim * sizeof (bool));
					for (j = oldval; j < dim; j++)
						PHOEBE_parameters[i].value.barray[j] = PHOEBE_parameters[i].defaultvalue.b;
				break;
				case TYPE_DOUBLE_ARRAY:
					phoebe_vector_realloc (PHOEBE_parameters[i].value.vec, dim);
					for (j = oldval; j < dim; j++)
						PHOEBE_parameters[i].value.vec->val[j] = PHOEBE_parameters[i].defaultvalue.d;
				break;
				case TYPE_STRING_ARRAY:
					PHOEBE_parameters[i].value.strarray = phoebe_realloc (PHOEBE_parameters[i].value.strarray, dim * sizeof (char *));
					for (j = oldval; j < dim; j++)
						PHOEBE_parameters[i].value.strarray[j] = strdup (PHOEBE_parameters[i].defaultvalue.str);
				break;
				}
			}
	}

void declare_all_parameters ()
	{
	/* This (and only this) function holds all parameters that are accessible   */
	/* to PHOEBE; there is no GUI connection or any other plug-in connection in */
	/* this function, only native PHOEBE parameters.                            */

	/* ***********************   Model parameters   *************************** */

	declare_parameter ("phoebe_name",                        "",                 "phoebe_name",                "Star name",                                  KIND_PARAMETER,    0.0,    0.0,    0.0, NO, TYPE_STRING,  "");
	declare_parameter ("phoebe_indep",                       "",                 "phoebe_indep",               "Independent variable",                       KIND_MENU,         0.0,    0.0,    0.0, NO, TYPE_STRING,  "Phase");
	declare_parameter ("phoebe_model",                       "",                 "phoebe_model",               "The model (morphological constraint)",       KIND_MENU,         0.0,    0.0,    0.0, NO, TYPE_STRING,  "Unconstrained binary system");

	declare_parameter ("phoebe_lcno",                        "dependency",       "phoebe_lcno",                "Number of observed light curves",            KIND_MODIFIER,     0.0,    0.0,    0.0, NO, TYPE_INT,     0);
	declare_parameter ("phoebe_rvno",                        "dependency",       "phoebe_rvno",                "Number of observed RV curves",               KIND_MODIFIER,     0.0,    0.0,    0.0, NO, TYPE_INT,     0);
	declare_parameter ("phoebe_spno",                        "dependency",       "phoebe_spno",                "Number of observed spectra",                 KIND_MODIFIER,     0.0,    0.0,    0.0, NO, TYPE_INT,     0);

	/* ***********************   Model constraints   ************************** */

	declare_parameter ("phoebe_asini_switch",                "",                 "phoebe_asini_switch",        "(a sin i) is kept constant",                 KIND_SWITCH,       0.0,    0.0,    0.0, NO, TYPE_BOOL,    NO);
	declare_parameter ("phoebe_asini_value",                 "",                 "phoebe_asini_value",         "(a sin i) constant",                         KIND_PARAMETER,    0.0,   1E10,    0.0, NO, TYPE_DOUBLE,  10.0);

	declare_parameter ("phoebe_msc1_switch", "", "phoebe_msc1_switch", "Main-sequence constraint for star 1", KIND_PARAMETER, 0, 0, 0, NO, TYPE_BOOL, NO);
	declare_parameter ("phoebe_msc2_switch", "", "phoebe_msc2_switch", "Main-sequence constraint for star 2", KIND_PARAMETER, 0, 0, 0, NO, TYPE_BOOL, NO);

	/* ************************   Data parameters   *************************** */

	declare_parameter ("phoebe_lc_filename",          "phoebe_lcno", "phoebe_lc_filename",          "Observed LC data filename",                  KIND_PARAMETER,    0.0,    0.0,    0.0, NO, TYPE_STRING_ARRAY, "Undefined");
	declare_parameter ("phoebe_lc_sigma",             "phoebe_lcno", "phoebe_lc_sigma",             "Observed LC data standard deviation",        KIND_PARAMETER,    0.0,    0.0,    0.0, NO, TYPE_DOUBLE_ARRAY, 0.01);
	declare_parameter ("phoebe_lc_filter",            "phoebe_lcno", "phoebe_lc_filter",            "Observed LC data filter",                    KIND_MENU,         0.0,    0.0,    0.0, NO, TYPE_STRING_ARRAY, "440nm (B)");
	declare_parameter ("phoebe_lc_indep",             "phoebe_lcno", "phoebe_lc_indep",             "Observed LC data independent variable",      KIND_MENU,         0.0,    0.0,    0.0, NO, TYPE_STRING_ARRAY, "Time (HJD)");
	declare_parameter ("phoebe_lc_dep",               "phoebe_lcno", "phoebe_lc_dep",               "Observed LC data dependent variable",        KIND_MENU,         0.0,    0.0,    0.0, NO, TYPE_STRING_ARRAY, "Magnitude");
	declare_parameter ("phoebe_lc_indweight",         "phoebe_lcno", "phoebe_lc_indweight",         "Observed LC data individual weighting",      KIND_MENU,         0.0,    0.0,    0.0, NO, TYPE_STRING_ARRAY, "Standard deviation (sigma)");
	declare_parameter ("phoebe_lc_levweight",         "phoebe_lcno", "phoebe_lc_levweight",         "Observed LC data level weighting",           KIND_MENU,         0.0,    0.0,    0.0, NO, TYPE_STRING_ARRAY, "Poissonian scatter");
	declare_parameter ("phoebe_lc_active",            "phoebe_lcno", "phoebe_lc_active",            "Observed LC data is used",                   KIND_SWITCH,       0.0,    0.0,    0.0, NO, TYPE_BOOL_ARRAY,   YES);

	declare_parameter ("phoebe_rv_filename",          "phoebe_rvno", "phoebe_rv_filename",          "Observed RV data filename",                  KIND_PARAMETER,    0.0,    0.0,    0.0, NO, TYPE_STRING_ARRAY, "Undefined");
	declare_parameter ("phoebe_rv_sigma",             "phoebe_rvno", "phoebe_rv_sigma",             "Observed RV data standard deviation",        KIND_PARAMETER,    0.0,    0.0,    0.0, NO, TYPE_DOUBLE_ARRAY, 1.0);
	declare_parameter ("phoebe_rv_filter",            "phoebe_rvno", "phoebe_rv_filter",            "Observed RV data filter",                    KIND_MENU,         0.0,    0.0,    0.0, NO, TYPE_STRING_ARRAY, "861nm (RVIJ)");
	declare_parameter ("phoebe_rv_indep",             "phoebe_rvno", "phoebe_rv_indep",             "Observed RV data independent variable",      KIND_MENU,         0.0,    0.0,    0.0, NO, TYPE_STRING_ARRAY, "Time (HJD)");
	declare_parameter ("phoebe_rv_dep",               "phoebe_rvno", "phoebe_rv_dep",               "Observed RV data dependent variable",        KIND_MENU,         0.0,    0.0,    0.0, NO, TYPE_STRING_ARRAY, "Primary RV");
	declare_parameter ("phoebe_rv_indweight",         "phoebe_rvno", "phoebe_rv_indweight",         "Observed RV data individual weighting",      KIND_MENU,         0.0,    0.0,    0.0, NO, TYPE_STRING_ARRAY, "Standard deviation (sigma)");
	declare_parameter ("phoebe_rv_active",            "phoebe_rvno", "phoebe_rv_active",            "Observed RV data is used",                   KIND_SWITCH,       0.0,    0.0,    0.0, NO, TYPE_BOOL_ARRAY,   YES);

	declare_parameter ("phoebe_mnorm",                "",            "phoebe_mnorm",                "Flux-normalizing magnitude",                 KIND_PARAMETER,    0.0,    0.0,    0.0, NO, TYPE_DOUBLE,  10.0);

	declare_parameter ("phoebe_bins_switch",          "",            "phoebe_bins_switch",          "Data binning",                               KIND_SWITCH,       0.0,    0.0,    0.0, NO, TYPE_BOOL,    NO);
	declare_parameter ("phoebe_bins",                 "",            "phoebe_bins",                 "Number of bins",                             KIND_PARAMETER,    0.0,    0.0,    0.0, NO, TYPE_INT,     100);

	declare_parameter ("phoebe_ie_switch",            "",            "phoebe_ie_switch",            "Interstellar extinction (reddening)",        KIND_SWITCH,       0.0,    0.0,    0.0, NO, TYPE_BOOL,    NO);
	declare_parameter ("phoebe_ie_factor",            "",            "phoebe_ie_factor",            "Interstellar extinction coefficient",        KIND_PARAMETER,    0.0,    0.0,    0.0, NO, TYPE_DOUBLE,  3.1);
	declare_parameter ("phoebe_ie_excess",            "",            "phoebe_ie_excess",            "Interstellar extinction color excess value", KIND_PARAMETER,    0.0,    0.0,    0.0, NO, TYPE_DOUBLE,  0.0);

	declare_parameter ("phoebe_proximity_rv1_switch", "",            "phoebe_proximity_rv1_switch", "Proximity effects for primary star RV",      KIND_SWITCH,       0.0,    0.0,    0.0, NO, TYPE_BOOL,    YES);
	declare_parameter ("phoebe_proximity_rv2_switch", "",            "phoebe_proximity_rv2_switch", "Proximity effects for secondary star RV",    KIND_SWITCH,       0.0,    0.0,    0.0, NO, TYPE_BOOL,    YES);

	/* ***********************   System parameters   ************************** */

	declare_parameter ("phoebe_hjd0",                        "",                 "phoebe_hjd0",      "Origin of HJD time",                         KIND_ADJUSTABLE, -1E10,   1E10, 0.0001, NO, TYPE_DOUBLE,    0.0);
	declare_parameter ("phoebe_period",                      "",                 "phoebe_period",    "Orbital period in days",                     KIND_ADJUSTABLE,   0.0,   1E10, 0.0001, NO, TYPE_DOUBLE,    1.0);
	declare_parameter ("phoebe_dpdt",                        "",                 "phoebe_dpdt",      "First time derivative of period (days/day)", KIND_ADJUSTABLE,  -1.0,    1.0,   1E-6, NO, TYPE_DOUBLE,    0.0);
	declare_parameter ("phoebe_pshift",                      "",                 "phoebe_pshift",    "Phase shift",                                KIND_ADJUSTABLE,  -0.5,    0.5,   0.01, NO, TYPE_DOUBLE,    0.0);
	declare_parameter ("phoebe_sma",                         "",                 "phoebe_sma",       "Semi-major axis in solar radii",             KIND_ADJUSTABLE,   0.0,   1E10,   0.01, NO, TYPE_DOUBLE,   10.0);
	declare_parameter ("phoebe_rm",                          "",                 "phoebe_rm",        "Mass ratio (secondary over primary",         KIND_ADJUSTABLE,   0.0,   1E10,   0.01, NO, TYPE_DOUBLE,    1.0);
	declare_parameter ("phoebe_incl",                        "",                 "phoebe_incl",      "Inclination in degrees",                     KIND_ADJUSTABLE,   0.0,  180.0,   0.01, NO, TYPE_DOUBLE,   80.0);
	declare_parameter ("phoebe_vga",                         "",                 "phoebe_vga",       "Center-of-mass velocity in km/s",            KIND_ADJUSTABLE,  -1E3,    1E3,    1.0, NO, TYPE_DOUBLE,    0.0);

	/* *********************   Component parameters   ************************* */

	declare_parameter ("phoebe_teff1",                       "",                 "phoebe_teff1",         "Primary star effective temperature in K",    KIND_ADJUSTABLE,  3500,  50000,     10, NO, TYPE_DOUBLE, 6000.0);
	declare_parameter ("phoebe_teff2",                       "",                 "phoebe_teff2",         "Secondary star effective temperature in K",  KIND_ADJUSTABLE,  3500,  50000,     10, NO, TYPE_DOUBLE, 6000.0);
	declare_parameter ("phoebe_pot1",                        "",                 "phoebe_pot1",          "Primary star surface potential",             KIND_ADJUSTABLE,   0.0,   1E10,   0.01, NO, TYPE_DOUBLE,   10.0);
	declare_parameter ("phoebe_pot2",                        "",                 "phoebe_pot2",          "Secondary star surface potential",           KIND_ADJUSTABLE,   0.0,   1E10,   0.01, NO, TYPE_DOUBLE,   10.0);
	declare_parameter ("phoebe_logg1",                       "",                 "phoebe_logg1",         "Primary star surface potential",             KIND_ADJUSTABLE,   0.0,   10.0,   0.01, NO, TYPE_DOUBLE,    4.3);
	declare_parameter ("phoebe_logg2",                       "",                 "phoebe_logg2",         "Primary star surface potential",             KIND_ADJUSTABLE,   0.0,   10.0,   0.01, NO, TYPE_DOUBLE,    4.3);
	declare_parameter ("phoebe_met1",                        "",                 "phoebe_met1",          "Primary star metallicity",                   KIND_ADJUSTABLE, -10.0,   10.0,   0.01, NO, TYPE_DOUBLE,    0.0);
	declare_parameter ("phoebe_met2",                        "",                 "phoebe_met2",          "Secondary star metallicity",                 KIND_ADJUSTABLE, -10.0,   10.0,   0.01, NO, TYPE_DOUBLE,    0.0);
	declare_parameter ("phoebe_f1",                          "",                 "phoebe_f1",            "Primary star synchronicity parameter",       KIND_ADJUSTABLE,   0.0,    1.0,   0.01, NO, TYPE_DOUBLE,    0.0);
	declare_parameter ("phoebe_f2",                          "",                 "phoebe_f2",            "Secondary star synchronicity parameter",     KIND_ADJUSTABLE,   0.0,    1.0,   0.01, NO, TYPE_DOUBLE,    0.0);
	declare_parameter ("phoebe_alb1",                        "",                 "phoebe_alb1",          "Primary star surface albedo",                KIND_ADJUSTABLE,   0.0,    1.0,   0.01, NO, TYPE_DOUBLE,    0.6);
	declare_parameter ("phoebe_alb2",                        "",                 "phoebe_alb2",          "Secondary star surface albedo",              KIND_ADJUSTABLE,   0.0,    1.0,   0.01, NO, TYPE_DOUBLE,    0.6);
	declare_parameter ("phoebe_grb1",                        "",                 "phoebe_grb1",          "Primary star gravity brightening",           KIND_ADJUSTABLE,   0.0,    1.0,   0.01, NO, TYPE_DOUBLE,   0.32);
	declare_parameter ("phoebe_grb2",                        "",                 "phoebe_grb2",          "Primary star gravity brightening",           KIND_ADJUSTABLE,   0.0,    1.0,   0.01, NO, TYPE_DOUBLE,   0.32);

	/* ***********************   Orbit parameters   *************************** */

	declare_parameter ("phoebe_ecc",                         "",                 "phoebe_ecc",           "Orbital eccentricity",                       KIND_ADJUSTABLE,   0.0,    1.0,   0.01, NO, TYPE_DOUBLE,    0.0);
	declare_parameter ("phoebe_perr0",                       "",                 "phoebe_perr0",         "Argument of periastron",                     KIND_ADJUSTABLE,   0.0, 2*M_PI,   0.01, NO, TYPE_DOUBLE,    0.0);
	declare_parameter ("phoebe_dperdt",                      "",                 "phoebe_dperdt",        "First time derivative of periastron",        KIND_ADJUSTABLE,  -1.0,    1.0,   1E-6, NO, TYPE_DOUBLE,    0.0);

	/* **********************   Surface parameters   ************************** */

	declare_parameter ("phoebe_hla",                         "phoebe_lcno",      "phoebe_hla",           "LC primary star flux leveler",               KIND_ADJUSTABLE,   0.0,   1E10,   0.01, NO, TYPE_DOUBLE_ARRAY, 10.0);
	declare_parameter ("phoebe_cla",                         "phoebe_lcno",      "phoebe_cla",           "LC secondary star flux leveler",             KIND_ADJUSTABLE,   0.0,   1E10,   0.01, NO, TYPE_DOUBLE_ARRAY, 10.0);
	declare_parameter ("phoebe_el3",                         "phoebe_lcno",      "phoebe_el3",           "Third light contribution",                   KIND_ADJUSTABLE,   0.0,   1E10,   0.01, NO, TYPE_DOUBLE_ARRAY, 0.0);
	declare_parameter ("phoebe_opsf",                        "phoebe_lcno",      "phoebe_opsf",          "Third light contribution",                   KIND_ADJUSTABLE,   0.0,   1E10,   0.01, NO, TYPE_DOUBLE_ARRAY, 0.0);

	declare_parameter ("phoebe_atm1_switch",                 "",                 "phoebe_atm1_switch",          "Use Kurucz's models for primary star",       KIND_SWITCH,       0.0,    0.0,    0.0, NO, TYPE_BOOL,      YES);
	declare_parameter ("phoebe_atm2_switch",                 "",                 "phoebe_atm2_switch",          "Use Kurucz's models for secondary star",     KIND_SWITCH,       0.0,    0.0,    0.0, NO, TYPE_BOOL,      YES);
	declare_parameter ("phoebe_reffect_switch",              "",                 "phoebe_reffect_switch",     "Detailed reflection effect",                 KIND_SWITCH,       0.0,    0.0,    0.0, NO, TYPE_BOOL,       NO);
	declare_parameter ("phoebe_reffect_reflections",         "",                 "phoebe_reffect_reflections",     "Number of detailed reflections",             KIND_PARAMETER,      2,     10,      1, NO, TYPE_INT,         2);

	declare_parameter ("phoebe_usecla_switch",               "",                 "phoebe_usecla_switch",   "Decouple CLAs from temperature",             KIND_SWITCH,       0.0,    0.0,    0.0, NO, TYPE_BOOL,       NO);

	/* **********************   Fitting parameters   ************************** */

	declare_parameter ("phoebe_grid_finesize1",              "",                 "phoebe_grid_finesize1",           "Fine grid size on primary star",             KIND_PARAMETER,      5,     60,      1, NO, TYPE_INT,        20);
	declare_parameter ("phoebe_grid_finesize2",              "",                 "phoebe_grid_finesize2",           "Fine grid size on secondary star",           KIND_PARAMETER,      5,     60,      1, NO, TYPE_INT,        20);
	declare_parameter ("phoebe_grid_coarsesize1",            "",                 "phoebe_grid_coarsesize1",           "Coarse grid size on primary star",           KIND_PARAMETER,      5,     60,      1, NO, TYPE_INT,         5);
	declare_parameter ("phoebe_grid_coarsesize2",            "",                 "phoebe_grid_coarsesize2",           "Coarse grid size on secondary star",         KIND_PARAMETER,      5,     60,      1, NO, TYPE_INT,         5);

	declare_parameter ("phoebe_calculate_hla",               "",                 "phoebe_calculate_hla",  "Calculate HLA levels",                       KIND_SWITCH,       0.0,    0.0,    0.0, NO, TYPE_BOOL,      YES);
	declare_parameter ("phoebe_calculate_vga",               "",                 "phoebe_calculate_vga",  "Calculate gamma velocity",                   KIND_SWITCH,       0.0,    0.0,    0.0, NO, TYPE_BOOL,      YES);

	/* ***********************   DC fit parameters   ************************** */

	declare_parameter ("phoebe_dc_symder_switch", "", "phoebe_dc_symder_switch", "Should symmetrical derivatives be used in DC", KIND_SWITCH,         0,      0,      0, NO,   TYPE_BOOL,    YES);
	declare_parameter ("phoebe_dc_lambda",        "", "phoebe_dc_lambda",        "Levenberg-Marquardt multiplier for DC",        KIND_PARAMETER,    0.0,    1.0,   1e-3, NO, TYPE_DOUBLE,   1e-3);

	declare_parameter ("phoebe_dc_spot1src", "", "phoebe_dc_spot1src", "Adjusted spot 1 source (at which star is the spot)", KIND_PARAMETER, 1,   2, 1, NO, TYPE_INT, 1);
	declare_parameter ("phoebe_dc_spot2src", "", "phoebe_dc_spot2src", "Adjusted spot 2 source (at which star is the spot)", KIND_PARAMETER, 1,   2, 1, NO, TYPE_INT, 2);
	declare_parameter ("phoebe_dc_spot1id",  "", "phoebe_dc_spot1id",  "Adjusted spot 1 id (which spot is to be adjusted)",  KIND_PARAMETER, 1, 100, 1, NO, TYPE_INT, 1);
	declare_parameter ("phoebe_dc_spot2id",  "", "phoebe_dc_spot2id",  "Adjusted spot 2 id (which spot is to be adjusted)",  KIND_PARAMETER, 1, 100, 1, NO, TYPE_INT, 1);

	/* ********************   Perturbations parameters   ********************** */

	declare_parameter ("phoebe_ld_model", "",            "phoebe_ld_model",     "Limb darkening model",                       KIND_MENU,           0,      0,      0, NO, TYPE_STRING, "Logarithmic law");
	declare_parameter ("phoebe_ld_xbol1", "",            "phoebe_ld_xbol1",     "Primary star bolometric LD coefficient x",   KIND_PARAMETER,    0.0,    1.0,   0.01, NO, TYPE_DOUBLE,  0.5);
	declare_parameter ("phoebe_ld_ybol1", "",            "phoebe_ld_ybol1",     "Secondary star bolometric LD coefficient x", KIND_PARAMETER,    0.0,    1.0,   0.01, NO, TYPE_DOUBLE,  0.5);
	declare_parameter ("phoebe_ld_xbol2", "",            "phoebe_ld_xbol2",     "Primary star bolometric LD coefficient y",   KIND_PARAMETER,    0.0,    1.0,   0.01, NO, TYPE_DOUBLE,  0.5);
	declare_parameter ("phoebe_ld_ybol2", "",            "phoebe_ld_ybol2",     "Secondary star bolometric LD coefficient y", KIND_PARAMETER,    0.0,    1.0,   0.01, NO, TYPE_DOUBLE,  0.5);
	declare_parameter ("phoebe_ld_lcx1",  "phoebe_lcno", "phoebe_ld_lcx1",      "Primary star bandpass LD coefficient x",     KIND_ADJUSTABLE,   0.0,    1.0,   0.01, NO, TYPE_DOUBLE_ARRAY, 0.5);
	declare_parameter ("phoebe_ld_lcx2",  "phoebe_lcno", "phoebe_ld_lcx2",      "Secondary star bandpass LD coefficient x",   KIND_ADJUSTABLE,   0.0,    1.0,   0.01, NO, TYPE_DOUBLE_ARRAY, 0.5);
	declare_parameter ("phoebe_ld_lcy1",  "phoebe_lcno", "phoebe_ld_lcy1",      "Primary star bandpass LD coefficient y",     KIND_PARAMETER,    0.0,    1.0,   0.01, NO, TYPE_DOUBLE_ARRAY, 0.5);
	declare_parameter ("phoebe_ld_lcy2",  "phoebe_lcno", "phoebe_ld_lcy2",      "Secondary star bandpass LD coefficient y",   KIND_PARAMETER,    0.0,    1.0,   0.01, NO, TYPE_DOUBLE_ARRAY, 0.5);
	declare_parameter ("phoebe_ld_rvx1",  "phoebe_rvno", "phoebe_ld_rvx1",      "Primary RV bandpass LD coefficient x",       KIND_PARAMETER,    0.0,    1.0,   0.01, NO, TYPE_DOUBLE_ARRAY, 0.5);
	declare_parameter ("phoebe_ld_rvx2",  "phoebe_rvno", "phoebe_ld_rvx2",      "Secondary RV bandpass LD coefficient x",     KIND_PARAMETER,    0.0,    1.0,   0.01, NO, TYPE_DOUBLE_ARRAY, 0.5);
	declare_parameter ("phoebe_ld_rvy1",  "phoebe_rvno", "phoebe_ld_rvy1",      "Primary RV bandpass LD coefficient y",       KIND_PARAMETER,    0.0,    1.0,   0.01, NO, TYPE_DOUBLE_ARRAY, 0.5);
	declare_parameter ("phoebe_ld_rvy2",  "phoebe_rvno", "phoebe_ld_rvy2",      "Secondary RV bandpass LD coefficient y",     KIND_PARAMETER,    0.0,    1.0,   0.01, NO, TYPE_DOUBLE_ARRAY, 0.5);

	declare_parameter ("phoebe_spots_no1",   "dependency",       "phoebe_spots_no1",   "Number of spots on primary star",            KIND_PARAMETER,      0,    100,      1, NO, TYPE_INT,       0);
	declare_parameter ("phoebe_spots_no2",   "dependency",       "phoebe_spots_no2",   "Number of spots on secondary star",          KIND_PARAMETER,      0,    100,      1, NO, TYPE_INT,       0);
	declare_parameter ("phoebe_spots_move1", "",                 "phoebe_spots_move1", "Spots on primary star move in longitude",    KIND_SWITCH,         0,      0,      0, NO, TYPE_BOOL,    YES);
	declare_parameter ("phoebe_spots_move2", "",                 "phoebe_spots_move2", "Spots on secondary star move in longitude",  KIND_SWITCH,         0,      0,      0, NO, TYPE_BOOL,    YES);
	declare_parameter ("phoebe_spots_lat1",  "phoebe_spots_no1", "phoebe_spots_lat1",  "Latitude of the spot on primary star",       KIND_PARAMETER,    0.0,   M_PI,   0.01, NO, TYPE_DOUBLE_ARRAY, 0.0);
	declare_parameter ("phoebe_spots_long1", "phoebe_spots_no1", "phoebe_spots_long1", "Longitude of the spot on primary star",      KIND_PARAMETER,    0.0, 2*M_PI,   0.01, NO, TYPE_DOUBLE_ARRAY, 0.0);
	declare_parameter ("phoebe_spots_rad1",  "phoebe_spots_no1", "phoebe_spots_rad1",  "Radius of the spot on primary star",         KIND_PARAMETER,    0.0,   M_PI,   0.01, NO, TYPE_DOUBLE_ARRAY, 0.2);
	declare_parameter ("phoebe_spots_temp1", "phoebe_spots_no1", "phoebe_spots_temp1", "Temperature of the spot on primary star",    KIND_PARAMETER,    0.0,    100,   0.01, NO, TYPE_DOUBLE_ARRAY, 0.9);
	declare_parameter ("phoebe_spots_lat2",  "phoebe_spots_no2", "phoebe_spots_lat2",  "Latitude of the spot on secondary star",     KIND_PARAMETER,    0.0,   M_PI,   0.01, NO, TYPE_DOUBLE_ARRAY, 0.0);
	declare_parameter ("phoebe_spots_long2", "phoebe_spots_no2", "phoebe_spots_long2", "Longitude of the spot on secondary star",    KIND_PARAMETER,    0.0, 2*M_PI,   0.01, NO, TYPE_DOUBLE_ARRAY, 0.0);
	declare_parameter ("phoebe_spots_rad2",  "phoebe_spots_no2", "phoebe_spots_rad2",  "Radius of the spot on secondary star",       KIND_PARAMETER,    0.0,   M_PI,   0.01, NO, TYPE_DOUBLE_ARRAY, 0.2);
	declare_parameter ("phoebe_spots_temp2", "phoebe_spots_no2", "phoebe_spots_temp2", "Temperature of the spot on secondary star",  KIND_PARAMETER,    0.0,    100,   0.01, NO, TYPE_DOUBLE_ARRAY, 0.9);

	/* **********************   Utilities parameters   ************************ */

	declare_parameter ("phoebe_synscatter_switch",    "", "phoebe_synscatter_switch", "Synthetic scatter",                          KIND_SWITCH,         0,      0,      0, NO, TYPE_BOOL,     NO);
	declare_parameter ("phoebe_synscatter_sigma",     "", "phoebe_synscatter_sigma",  "Synthetic scatter standard deviation",       KIND_PARAMETER,    0.0,  100.0,   0.01, NO, TYPE_DOUBLE, 0.01);
	declare_parameter ("phoebe_synscatter_seed",      "", "phoebe_synscatter_seed",   "Synthetic scatter seed",                     KIND_PARAMETER,    1E8,    1E9,      1, NO, TYPE_DOUBLE, 150000000.0);
	declare_parameter ("phoebe_synscatter_levweight", "", "phoebe_synscatter_levweight", "Synthetic scatter weighting",                KIND_MENU,           0,      0,      0, NO, TYPE_STRING, "Poissonian scatter");

	/* **********************   Scripter parameters   ************************* */

	declare_parameter ("scripter_verbosity_level",    "", "scripter_verbosity_level", "Scripter verbosity level",                   KIND_PARAMETER,      0,      1,      1, NO, TYPE_INT,     1);


	/* ****************** Obsolete and not-yet-implemented: ******************* */
/*
	declare_parameter ("phoebe_synthetic_reddening",         "",                 "SYNREDSWITCH",  "Synthetic interstellar reddening",           KIND_SWITCH,         0,      0,      0, NO, TYPE_BOOL,     NO);
	declare_parameter ("phoebe_synthetic_reddening_factor",  "",                 "SYNREDFACTOR",  "Synthetic interstellar reddening factor",    KIND_PARAMETER,      0,     10,   0.01, NO, TYPE_DOUBLE,  3.1);
	declare_parameter ("phoebe_synthetic_reddening_excess",  "",                 "SYNREDEXCESS",  "Synthetic interstellar reddening excess",    KIND_PARAMETER,      0,     10,   0.01, NO, TYPE_DOUBLE,  0.0);

	declare_parameter ("phoebe_lc_ptsno_value",      "phoebe_lcno", "LCPTSNO",   "Experimental data points number",             TYPE_INT_ARRAY);
	declare_parameter ("phoebe_lc_sum_value",        "phoebe_lcno", "LCSUM",     "Experimental data points sum",                TYPE_DOUBLE_ARRAY);
	declare_parameter ("phoebe_lc_average_value",    "phoebe_lcno", "LCAVERAGE", "Experimental data points average",            TYPE_DOUBLE_ARRAY);
	declare_parameter ("phoebe_lc_median_value",     "phoebe_lcno", "LCMEDIAN",  "Experimental data points median",             TYPE_DOUBLE_ARRAY);
	declare_parameter ("phoebe_rv_ptsno_value",      "phoebe_rvno", "RVPTSNO",   "Experimental data points number",             TYPE_DOUBLE_ARRAY);
	declare_parameter ("phoebe_rv_sum_value",        "phoebe_rvno", "RVSUM",     "Experimental data points sum",                TYPE_DOUBLE_ARRAY);
	declare_parameter ("phoebe_rv_average_value",    "phoebe_rvno", "RVAVERAGE", "Experimental data points average",            TYPE_DOUBLE_ARRAY);
	declare_parameter ("phoebe_rv_median_value",     "phoebe_rvno", "RVMEDIAN",  "Experimental data points median",             TYPE_DOUBLE_ARRAY);

	declare_parameter ("phoebe_plot_lc_column_2",    "phoebe_lcno", "PLOTLCCOLUMN2", "",                                                 TYPE_STRING_ARRAY);
	declare_parameter ("phoebe_plot_lc_column_3",    "phoebe_lcno", "PLOTLCCOLUMN3", "",                                                 TYPE_STRING_ARRAY);
	declare_parameter ("phoebe_plot_lc_shift",       "phoebe_lcno", "PLOTLCSHIFT", "",                                                   TYPE_DOUBLE_ARRAY);
	declare_parameter ("phoebe_plot_lc_nexp_value",  "phoebe_lcno", "LCSTATNEXP",  "Experimental data points number",         TYPE_DOUBLE_ARRAY);
	declare_parameter ("phoebe_plot_lc_aexp_value",  "phoebe_lcno", "LCSTATAEXP",  "Experimental data points average",        TYPE_DOUBLE_ARRAY);
	declare_parameter ("phoebe_plot_lc_mexp_value",  "phoebe_lcno", "LCSTATMEXP",  "Experimental data points median",         TYPE_DOUBLE_ARRAY);
	declare_parameter ("phoebe_plot_lc_asyn_value",  "phoebe_lcno", "LCSTATASYN",  "Synthetic data points average",           TYPE_DOUBLE_ARRAY);
	declare_parameter ("phoebe_plot_lc_msyn_value",  "phoebe_lcno", "LCSTATMSYN",  "Synthetic data points median",            TYPE_DOUBLE_ARRAY);
	declare_parameter ("phoebe_plot_lc_sigma_value", "phoebe_lcno", "LCSTATSIGMA", "Unweighted standard deviation",           TYPE_DOUBLE_ARRAY);
	declare_parameter ("phoebe_plot_lc_wsig_value",  "phoebe_lcno", "LCSTATWSIG",  "Weighted standard deviation",             TYPE_DOUBLE_ARRAY);

	declare_parameter ("output_lc_indep",                    "",                 "OUTLCINDEP",     "Independent LC variable used for output",    KIND_MENU,           0,      0,      0, NO, TYPE_STRING,  "Phase");
	declare_parameter ("output_lc_dep",                      "",                 "OUTLCDEP",       "Dependent LC variable used for output",      KIND_MENU,           0,      0,      0, NO, TYPE_STRING,  "Total flux");
	declare_parameter ("output_lc_individual_weighting",     "",                 "OUTLCINDWEIGHT", "Individual LC weighting used for output",    KIND_MENU,           0,      0,      0, NO, TYPE_STRING,  "Standard deviation");
	declare_parameter ("output_lc_phase_start",              "",                 "OUTLCPHSTART",   "LC phase lower limit used for output",       KIND_PARAMETER,  -1E10,   1E10,   0.01, NO, TYPE_DOUBLE,  -0.6);
	declare_parameter ("output_lc_phase_end",                "",                 "OUTLCPHEND",     "LC phase upper limit used for output",       KIND_PARAMETER,  -1E10,   1E10,   0.01, NO, TYPE_DOUBLE,   0.6);
	declare_parameter ("output_lc_vertices",                 "",                 "OUTLCVERTICES",  "Number of LC vertices used for output",      KIND_PARAMETER,      2,    1E5,      1, NO, TYPE_INT,      300);
	declare_parameter ("output_lc_aliasing",                 "",                 "OUTLCALIASING",  "Alias LC observations on all phases",        KIND_SWITCH,         0,      0,      0, NO, TYPE_BOOL,     YES);
	declare_parameter ("output_lc_residuals",                "",                 "OUTLCRESIDUALS", "Plot RV residuals (O-C)",                    KIND_SWITCH,         0,      0,      0, NO, TYPE_BOOL,      NO);

	declare_parameter ("output_rv_indep",                    "",                 "OUTRVINDEP",     "Independent RV variable used for output",    KIND_MENU,           0,      0,      0, NO, TYPE_STRING,  "Phase");
	declare_parameter ("output_rv_dep",                      "",                 "OUTRVDEP",       "Dependent RV variable used for output",      KIND_MENU,           0,      0,      0, NO, TYPE_STRING,  "Both RVs in km/s");
	declare_parameter ("output_rv_individual_weighting",     "",                 "OUTRVINDWEIGHT", "Individual RV weighting used for output",    KIND_MENU,           0,      0,      0, NO, TYPE_STRING,  "Standard deviation");
	declare_parameter ("output_rv_phase_start",              "",                 "OUTRVPHSTART",   "RV phase lower limit used for output",       KIND_PARAMETER,  -1E10,   1E10,   0.01, NO, TYPE_DOUBLE,  -0.6);
	declare_parameter ("output_rv_phase_end",                "",                 "OUTRVPHEND",     "RV phase upper limit used for output",       KIND_PARAMETER,  -1E10,   1E10,   0.01, NO, TYPE_DOUBLE,   0.6);
	declare_parameter ("output_rv_vertices",                 "",                 "OUTRVVERTICES",  "Number of RV vertices used for output",      KIND_PARAMETER,      2,    1E5,      1, NO, TYPE_INT,       50);
	declare_parameter ("output_rv_aliasing",                 "",                 "OUTRVALIASING",  "Alias RV observations on all phases",        KIND_SWITCH,         0,      0,      0, NO, TYPE_BOOL,     YES);
	declare_parameter ("output_rv_residuals",                "",                 "OUTRVRESIDUALS", "Plot RV residuals (O-C)",                    KIND_SWITCH,         0,      0,      0, NO, TYPE_BOOL,      NO);
*/

	declare_parameter ("phoebe_cc",                          "",                 "phoebe_cc",                  "Conditional constraints",                    KIND_MENU,         0.0,    0.0,    0.0, NO, TYPE_STRING,  "None");
	}

char *phoebe_qualifier_from_keyword (char *keyword)
	{
	/* This function returns the qualifier of the given keyword.                */
	
	int i;
	
	for (i = 0; i < PHOEBE_parameters_no; i++)
		if ( strcmp (PHOEBE_parameters[i].keyword, keyword) == 0 )
			return PHOEBE_parameters[i].qualifier;

	phoebe_warning ("phoebe_qualifier_from_keyword () called with keyword\n                '%s', which doesn't exist!\n", keyword);
	return NULL;
	}

char *phoebe_qualifier_from_description (char *description)
	{
	/* This function returns the qualifier of the given description.            */
	
	int i;
	
	for (i = 0; i < PHOEBE_parameters_no; i++)
		if ( strcmp (PHOEBE_parameters[i].description, description) == 0 )
			return PHOEBE_parameters[i].qualifier;

	phoebe_warning ("phoebe_qualifier_from_description () called with description\n                '%s', which doesn't exist!\n", description);
	return NULL;
	}

char *phoebe_keyword_from_qualifier (char *qualifier)
	{
	/* This function returns the keyword of the given qualifier.                */
	
	int i;
	
	for (i = 0; i < PHOEBE_parameters_no; i++)
		if ( strcmp (PHOEBE_parameters[i].qualifier, qualifier) == 0 )
			return PHOEBE_parameters[i].keyword;

	phoebe_warning ("phoebe_keyword_from_qualifier () called with qualifier\n                '%s', which doesn't exist!\n", qualifier);
	return NULL;
	}

char *phoebe_keyword_from_description (char *description)
	{
	/* This function returns the keyword of the given description.              */
	
	int i;
	
	for (i = 0; i < PHOEBE_parameters_no; i++)
		if ( strcmp (PHOEBE_parameters[i].description, description) == 0 )
			return PHOEBE_parameters[i].keyword;

	phoebe_warning ("phoebe_keyword_from_description () called with description\n                '%s', which doesn't exist!\n", description);
	return NULL;
	}

char *phoebe_description_from_qualifier (char *qualifier)
	{
	/* This function returns the description of the given qualifier.            */
	
	int i;
	
	for (i = 0; i < PHOEBE_parameters_no; i++)
		if ( strcmp (PHOEBE_parameters[i].qualifier, qualifier) == 0 )
			return PHOEBE_parameters[i].description;

	phoebe_warning ("phoebe_description_from_qualifier () called with qualifier\n                '%s', which doesn't exist!\n", qualifier);
	return NULL;
	}

char *phoebe_description_from_keyword (char *keyword)
	{
	/* This function returns the description of the given keyword.              */
	
	int i;
	
	for (i = 0; i < PHOEBE_parameters_no; i++)
		if ( strcmp (PHOEBE_parameters[i].keyword, keyword) == 0 )
			return PHOEBE_parameters[i].description;

	phoebe_warning ("phoebe_description_from_keyword () called with keyword\n                '%s', which doesn't exist!\n", keyword);
	return NULL;
	}

int phoebe_index_from_qualifier (char *qualifier)
	{
	/* This function returns the parameter table index of the given qualifier.  */
	
	int i;
	
	for (i = 0; i < PHOEBE_parameters_no; i++)
		if ( strcmp (PHOEBE_parameters[i].qualifier, qualifier) == 0 )
			return i;

	phoebe_warning ("phoebe_index_from_qualifier () called with qualifier\n                '%s', which doesn't exist!\n", qualifier);
	return -1;
	}

int phoebe_index_from_keyword (char *keyword)
	{
	/* This function returns the parameter table index of the given keyword.    */
	
	int i;
	
	for (i = 0; i < PHOEBE_parameters_no; i++)
		if ( strcmp (PHOEBE_parameters[i].keyword, keyword) == 0 )
			return i;

	phoebe_warning ("phoebe_index_from_keyword () called with keyword\n                '%s', which doesn't exist!\n", keyword);
	return -1;
	}

int phoebe_index_from_description (char *description)
	{
	/* This function returns the parameter table index of the given description.*/
	
	int i;
	
	for (i = 0; i < PHOEBE_parameters_no; i++)
		if ( strcmp (PHOEBE_parameters[i].description, description) == 0 )
			return i;

	phoebe_warning ("phoebe_index_from_description () called with description\n                '%s', which doesn't exist!\n", description);
	return -1;
	}

PHOEBE_type phoebe_type_from_qualifier (char *qualifier)
	{
	/* This function returns the parameter type of the given qualifier. If the  */
	/* qualifier doesn't exist, -1 is returned.                                 */
	
	int i;
	
	for (i = 0; i < PHOEBE_parameters_no; i++)
		if ( strcmp (PHOEBE_parameters[i].qualifier, qualifier) == 0 )
			return PHOEBE_parameters[i].type;

	return -1;
	}

PHOEBE_type phoebe_type_from_keyword (char *keyword)
	{
	/* This function returns the parameter type of the given keyword.           */
	
	int i;
	
	for (i = 0; i < PHOEBE_parameters_no; i++)
		if ( strcmp (PHOEBE_parameters[i].keyword, keyword) == 0 )
			return PHOEBE_parameters[i].type;

	return -1;
	}

PHOEBE_type phoebe_type_from_description (char *description)
	{
	/* This function returns the parameter type of the given description.       */
	
	int i;
	
	for (i = 0; i < PHOEBE_parameters_no; i++)
		if ( strcmp (PHOEBE_parameters[i].description, description) == 0 )
			return PHOEBE_parameters[i].type;

	phoebe_warning ("phoebe_type_from_description () called with description\n                '%s', which doesn't exist!\n", description);
	return -1;
	}

PHOEBE_parameter_kind phoebe_kind_from_qualifier (char *qualifier)
	{
	int i;
	
	for (i = 0; i < PHOEBE_parameters_no; i++)
		if ( strcmp (PHOEBE_parameters[i].qualifier, qualifier) == 0 )
			return PHOEBE_parameters[i].kind;

	phoebe_warning ("phoebe_kind_from_qualifier () called with qualifier\n                '%s', which doesn't exist!\n", qualifier);
	return -1;
	}

/* ************************************************************************** */

int phoebe_get_value_int (char *qualifier)
	{
	return PHOEBE_parameters[phoebe_index_from_qualifier (qualifier)].value.i;
	}

bool phoebe_get_value_bool (char *qualifier)
	{
	return PHOEBE_parameters[phoebe_index_from_qualifier (qualifier)].value.b;
	}

double phoebe_get_value_double (char *qualifier)
	{
	return PHOEBE_parameters[phoebe_index_from_qualifier (qualifier)].value.d;
	}

const char *phoebe_get_value_string (char *qualifier)
	{
	return PHOEBE_parameters[phoebe_index_from_qualifier (qualifier)].value.str;
	}

char *phoebe_get_value_list_string (char *qualifier, int row)
	{
	return PHOEBE_parameters[phoebe_index_from_qualifier (qualifier)].value.strarray[row];
	}

double phoebe_get_value_list_double (char *qualifier, int row)
	{
	return PHOEBE_parameters[phoebe_index_from_qualifier (qualifier)].value.vec->val[row];
	}

bool phoebe_get_value_list_bool (char *qualifier, int row)
	{
	return PHOEBE_parameters[phoebe_index_from_qualifier (qualifier)].value.barray[row];
	}

int phoebe_get_value_list_int (char *qualifier, int row)
	{
	return PHOEBE_parameters[phoebe_index_from_qualifier (qualifier)].value.iarray[row];
	}

int phoebe_get_parameter_value (char *qualifier, ...)
	{
	/* This is the public function for changing qualifier values. It is the on- */
	/* ly function that should be used for this purpose, all other functions    */
	/* should be regarded as internal and should not be used.                   */
	/*                                                                          */
	/* Synopsis:                                                                */
	/*                                                                          */
	/*   phoebe_get_parameter_value (qualifier, [index, ], &value)              */
	/*                                                                          */
	/* Return values:                                                           */
	/*                                                                          */
	/*   -2  ..  array subscript is out of range                                */
	/*    0  ..  everything ok, continue                                        */

	int index = 0;
	va_list args;

	va_start (args, qualifier);
	switch (phoebe_type_from_qualifier (qualifier))
		{
		case TYPE_INT:
			{
			int *value = va_arg (args, int *);
			*value = phoebe_get_value_int (qualifier);
			}
		break;
		case TYPE_BOOL:
			{
			bool *value = va_arg (args, bool *);
			*value = phoebe_get_value_bool (qualifier);
			}
		break;
		case TYPE_DOUBLE:
			{
			double *value = va_arg (args, double *);
			*value = phoebe_get_value_double (qualifier);
			}
		break;
		case TYPE_STRING:
			{
			const char **value = va_arg (args, const char **);
			*value = phoebe_get_value_string (qualifier);
			}
		break;
		case TYPE_INT_ARRAY:
			index = va_arg (args, int);
			{
			int *value = va_arg (args, int *);
			int i = phoebe_index_from_qualifier (qualifier);
			int range = phoebe_get_value_int (PHOEBE_parameters[i].dependency);
			if (index < 0 || index > range-1) return -2;

			*value = phoebe_get_value_list_int (qualifier, index);
			}
		break;
		case TYPE_BOOL_ARRAY:
			index = va_arg (args, int);
			{
			bool *value = va_arg (args, bool *);
			int i = phoebe_index_from_qualifier (qualifier);
			int range = phoebe_get_value_int (PHOEBE_parameters[i].dependency);
			if (index < 0 || index > range-1) return -2;

			*value = phoebe_get_value_list_bool (qualifier, index);
			}
		break;
		case TYPE_DOUBLE_ARRAY:
			index = va_arg (args, int);
			{
			double *value = va_arg (args, double *);
			int i = phoebe_index_from_qualifier (qualifier);
			int range = phoebe_get_value_int (PHOEBE_parameters[i].dependency);
			if (index < 0 || index > range-1) return -2;

			*value = phoebe_get_value_list_double (qualifier, index);
			}
		break;
		case TYPE_STRING_ARRAY:
			index = va_arg (args, int);
			{
			const char **value = va_arg (args, const char **);
			int i = phoebe_index_from_qualifier (qualifier);
			int range = phoebe_get_value_int (PHOEBE_parameters[i].dependency);
			if (index < 0 || index > range-1) return -2;

			*value = phoebe_get_value_list_string (qualifier, index);
			}
		break;
		}
	va_end (args);

	return 0;
	}

/* ************************************************************************** */

int intern_set_value_int (char *qualifier, int value)
	{
	int oldval = phoebe_get_value_int (qualifier);
	int i = phoebe_index_from_qualifier (qualifier);

	PHOEBE_parameters[i].value.i = value;
	if (strcmp (PHOEBE_parameters[i].dependency, "dependency") == 0)
		update_parameter_arrays (qualifier, oldval);

	return 0;
	}

int intern_set_value_double (char *qualifier, double value)
	{
	int i = phoebe_index_from_qualifier (qualifier);
	PHOEBE_parameters[i].value.d = value;

	return 0;
	}

int intern_set_value_bool (char *qualifier, bool value)
	{
	int i = phoebe_index_from_qualifier (qualifier);
	PHOEBE_parameters[i].value.b = value;

	return 0;
	}

int intern_set_value_string (char *qualifier, const char *value)
	{
	int i = phoebe_index_from_qualifier (qualifier);
	free (PHOEBE_parameters[i].value.str);
	PHOEBE_parameters[i].value.str = phoebe_malloc (strlen (value) + 1);
	strcpy (PHOEBE_parameters[i].value.str, value);

	return 0;
	}

int intern_set_value_list_string (char *qualifier, int row, const char *value)
	{
	/* Return values:                                                           */
	/*                                                                          */
	/*   -2  ..  the array index is out of bounds                               */
	/*   -1  ..  (reserved for non-existent qualifier signal)                   */
	/*    0  ..  set command successful                                         */

	int i = phoebe_index_from_qualifier (qualifier);
	int range;

	range = phoebe_get_value_int (PHOEBE_parameters[i].dependency);
	if (row < 0 || row > range-1) return -2;

	free (PHOEBE_parameters[i].value.strarray[row]);
	PHOEBE_parameters[i].value.strarray[row] = phoebe_malloc (strlen (value) + 1);
	strcpy (PHOEBE_parameters[i].value.strarray[row], value);
	return 0;
	}

int intern_set_value_list_int (char *qualifier, int row, int value)
	{
	/* Return values:                                                           */
	/*                                                                          */
	/*   -2  ..  the array index is out of bounds                               */
	/*   -1  ..  (reserved for non-existent qualifier signal)                   */
	/*    0  ..  set command successful                                         */

	int i = phoebe_index_from_qualifier (qualifier);
	int range;

	range = phoebe_get_value_int (PHOEBE_parameters[i].dependency);
	if (row < 0 || row > range-1) return -2;

	PHOEBE_parameters[i].value.iarray[row] = value;
	return 0;
	}

int intern_set_value_list_double (char *qualifier, int row, double value)
	{
	/* Return values:                                                           */
	/*                                                                          */
	/*   -2  ..  the array index is out of bounds                               */
	/*   -1  ..  (reserved for non-existent qualifier signal)                   */
	/*    0  ..  set command successful                                         */

	int i = phoebe_index_from_qualifier (qualifier);
	int range;

	range = phoebe_get_value_int (PHOEBE_parameters[i].dependency);
	if (row < 0 || row > range-1) return -2;

	PHOEBE_parameters[i].value.vec->val[row] = value;
	return 0;
	}

int intern_set_value_list_bool (char *qualifier, int row, bool value)
	{
	/* Return values:                                                           */
	/*                                                                          */
	/*   -2  ..  the array index is out of bounds                               */
	/*   -1  ..  (reserved for non-existent qualifier signal)                   */
	/*    0  ..  set command successful                                         */

	int i = phoebe_index_from_qualifier (qualifier);
	int range;

	range = phoebe_get_value_int (PHOEBE_parameters[i].dependency);
	if (row < 0 || row > range-1) return -2;

	PHOEBE_parameters[i].value.barray[row] = value;
	return 0;
	}

int phoebe_set_parameter_value (char *qualifier, ...)
	{
	/* This is the public function for changing qualifier values. It is the on- */
	/* ly function that should be used for this purpose, all other functions    */
	/* should be regarded as internal and should not be used.                   */
	/*                                                                          */
	/* Synopsis:                                                                */
	/*                                                                          */
	/*   phoebe_set_parameter_value (qualifier, [curve, ] value)                */

	int index = 0;
	va_list args;
	int errno;

	PHOEBE_type type = phoebe_type_from_qualifier (qualifier);

	if (type == -1) return -1;

	va_start (args, qualifier);
	switch (phoebe_type_from_qualifier (qualifier))
		{
		case TYPE_INT:
			{
			int value = va_arg (args, int);
			errno = intern_set_value_int (qualifier, value);
			}
		break;
		case TYPE_BOOL:
			{
			bool value = va_arg (args, bool);
			errno = intern_set_value_bool (qualifier, value);
			}
		break;
		case TYPE_DOUBLE:
			{
			double value = va_arg (args, double);
			errno = intern_set_value_double (qualifier, value);
			}
		break;
		case TYPE_STRING:
			{
			char *value = va_arg (args, char *);
			errno = intern_set_value_string (qualifier, value);
			}
		break;
		case TYPE_INT_ARRAY:
			index = va_arg (args, int);
			{
			int value = va_arg (args, int);
			errno = intern_set_value_list_int (qualifier, index, value);
			}
		break;
		case TYPE_BOOL_ARRAY:
			index = va_arg (args, int);
			{
			bool value = va_arg (args, bool);
			errno = intern_set_value_list_bool (qualifier, index, value);
			}
		break;
		case TYPE_DOUBLE_ARRAY:
			index = va_arg (args, int);
			{
			double value = va_arg (args, double);
			errno = intern_set_value_list_double (qualifier, index, value);
			}
		break;
		case TYPE_STRING_ARRAY:
			index = va_arg (args, int);
			{
			char *value = va_arg (args, char *);
			errno = intern_set_value_list_string (qualifier, index, value);
			}
		break;
		}
	va_end (args);

	/* Since the global lookup table has changed, we have to signal the plugins */
	/* to update their own tables.                                              */
	if (errno == 0)
		emit_parameter_changed_signal_for_plugins (qualifier, index);

	return errno;
	}

int phoebe_set_parameter_tba (char *qualifier, bool state)
	{
	/* This is the public function for changing qualifier's TBA (To Be Adjus-   */
	/* ted) marker.                                                             */
	/*                                                                          */
	/* Error codes:                                                             */
	/*                                                                          */
	/*   -1   ..   qualifier doesn't exist                                      */
	/*    0   ..   assignment successful                                        */

	int i = phoebe_index_from_qualifier (qualifier);
	if (i == -1) return -1;

	PHOEBE_parameters[i].tba = state;

	return 0;
	}

bool phoebe_get_parameter_tba (char *qualifier)
	{
	int i = phoebe_index_from_qualifier (qualifier);
	if (i == -1) return -1;

	return PHOEBE_parameters[i].tba;
	}

double phoebe_get_parameter_step (char *qualifier)
	{
	int i = phoebe_index_from_qualifier (qualifier);
	if (i == -1) return -1;

	return PHOEBE_parameters[i].step;
	}

int phoebe_set_parameter_limits (char *qualifier, double valmin, double valmax)
	{
	/* This is the public function for changing qualifier limits.               */
	/*                                                                          */
	/* Error codes:                                                             */
	/*                                                                          */
	/*   -1   ..   qualifier doesn't exist                                      */
	/*    0   ..   assignment successful                                        */

	int i = phoebe_index_from_qualifier (qualifier);
	if (i == -1) return -1;

	PHOEBE_parameters[i].min = valmin;
	PHOEBE_parameters[i].max = valmax;

	return 0;
	}

int phoebe_set_parameter_step (char *qualifier, double step)
	{
	/* This is the public function for changing qualifier step.                 */
	/*                                                                          */
	/* Error codes:                                                             */
	/*                                                                          */
	/*   -1   ..   qualifier doesn't exist                                      */
	/*    0   ..   assignment successful                                        */

	int i = phoebe_index_from_qualifier (qualifier);
	if (i == -1) return -1;

	PHOEBE_parameters[i].step = step;

	return 0;
	}

PHOEBE_input_indep get_input_independent_variable (char *value)
	{
	/* This function returns the input independent variable enumeration.        */
	/*                                                                          */
	/* Return values:                                                           */
	/*                                                                          */
	/*     -1  ..  recognition failed                                           */
	/*   enum  ..  the recognition was successful                               */

	if (strcmp (value, "Phase")      == 0) return INPUT_PHASE;
	if (strcmp (value, "Time (HJD)") == 0) return INPUT_HJD;
	
	return -1;
	}

PHOEBE_input_dep get_input_dependent_variable (char *value)
	{
	/* This function returns the input dependent variable enumeration.          */
	/*                                                                          */
	/* Return values:                                                           */
	/*                                                                          */
	/*     -1  ..  recognition failed                                           */
	/*   enum  ..  the recognition was successful                               */

	if (strcmp (value, "Magnitude")    == 0) return INPUT_MAGNITUDE;
	if (strcmp (value, "Flux")         == 0) return INPUT_FLUX;
	if (strcmp (value, "Primary RV")   == 0) return INPUT_PRIMARY_RV;
	if (strcmp (value, "Secondary RV") == 0) return INPUT_SECONDARY_RV;
	
	return -1;
	}

PHOEBE_input_weight get_input_weight (char *value)
	{
	/* This function returns the input weight variable enumeration.             */
	/*                                                                          */
	/* Return values:                                                           */
	/*                                                                          */
	/*     -1  ..  recognition failed                                           */
	/*   enum  ..  the recognition was successful                               */

	if (strcmp (value, "Standard weight (integer)")  == 0) return INPUT_STANDARD_WEIGHT;
	if (strcmp (value, "Standard weight (double)")   == 0) return INPUT_STANDARD_WEIGHT;
	if (strcmp (value, "Standard deviation (sigma)") == 0) return INPUT_STANDARD_DEVIATION;
	if (strcmp (value, "Unavailable")                == 0) return INPUT_UNAVAILABLE;
	
	return -1;
	}

PHOEBE_output_indep get_output_independent_variable (char *value)
	{
	/* This function returns the output independent variable enumeration.       */
	/*                                                                          */
	/* Return values:                                                           */
	/*                                                                          */
	/*     -1  ..  recognition failed                                           */
	/*   enum  ..  the recognition was successful                               */
	/*                                                                          */

	if (strcmp (value, "Phase")      == 0) return OUTPUT_PHASE;
	if (strcmp (value, "Time (HJD)") == 0) return OUTPUT_HJD;

	return -1;
	}

PHOEBE_output_dep get_output_dependent_variable (char *value)
	{
	/* This function returns the output dependent variable enumeration.         */
	/*                                                                          */
	/* Return values:                                                           */
	/*                                                                          */
	/*     -1  ..  recognition failed                                           */
	/*   enum  ..  the recognition was successful                               */

	if (strcmp (value, "Magnitude")               == 0) return OUTPUT_MAGNITUDE;
	if (strcmp (value, "Total flux")              == 0) return OUTPUT_TOTAL_FLUX;
	if (strcmp (value, "Primary star flux")       == 0) return OUTPUT_PRIMARY_FLUX;
	if (strcmp (value, "Secondary star flux")     == 0) return OUTPUT_SECONDARY_FLUX;
	if (strcmp (value, "Primary RV")              == 0) return OUTPUT_PRIMARY_RV;
	if (strcmp (value, "Secondary RV")            == 0) return OUTPUT_SECONDARY_RV;
	if (strcmp (value, "Both RVs")                == 0) return OUTPUT_BOTH_RVS;
	if (strcmp (value, "Primary normalized RV")   == 0) return OUTPUT_PRIMARY_NORMALIZED_RV;
	if (strcmp (value, "Secondary normalized RV") == 0) return OUTPUT_SECONDARY_NORMALIZED_RV;

	return -1;
	}

PHOEBE_output_weight get_output_weight (char *value)
	{
	/* This function returns the output weighting variable enumeration.         */
	/*                                                                          */
	/* Return values:                                                           */
	/*                                                                          */
	/*     -1  ..  recognition failed                                           */
	/*   enum  ..  the recognition was successful                               */

	if (strcmp (value, "Standard weight")            == 0) return OUTPUT_STANDARD_WEIGHT;
	if (strcmp (value, "Standard deviation (sigma)") == 0) return OUTPUT_STANDARD_DEVIATION;
	if (strcmp (value, "Unavailable")                == 0) return OUTPUT_UNAVAILABLE;

	return -1;
	}

void get_from_keyword_file (char *qualifier, char *value_str)
	{
	int i = phoebe_index_from_qualifier (qualifier);
	int errno;

	switch (PHOEBE_parameters[i].type)
		{
		case TYPE_INT:
			errno = phoebe_set_parameter_value (qualifier, atoi (value_str));
		break;
		case TYPE_BOOL:
			errno = phoebe_set_parameter_value (qualifier, atoi (value_str));
		break;
		case TYPE_DOUBLE:
			errno = phoebe_set_parameter_value (qualifier, atof (value_str));
		break;
		case TYPE_STRING:
			/* Strip the string of quotes if necessary:                             */
			while (value_str[0] == '"') value_str++;
			while (value_str[strlen(value_str)-1] == '"') value_str[strlen(value_str)-1] = '\0';
			errno = phoebe_set_parameter_value (qualifier, value_str);
		break;
		}

	if (errno != 0)
		phoebe_warning ("qualifier '%s' cannot be set!\n", qualifier);
	}

void get_wavelength_dependent_parameter_from_keyword_file (char *qualifier, int idx, char *value_str)
	{
	/* Much like the function above, only that this one takes three arguments,  */
	/* the 2nd one being the index of the wavelength-dependent parameter.       */

	int i = phoebe_index_from_qualifier (qualifier);
	int errno;

	switch (PHOEBE_parameters[i].type)
		{
		case TYPE_INT_ARRAY:
			errno = phoebe_set_parameter_value (qualifier, idx, atoi (value_str));
		break;
		case TYPE_BOOL_ARRAY:
			errno = phoebe_set_parameter_value (qualifier, idx, atoi (value_str));
		break;
		case TYPE_DOUBLE_ARRAY:
			errno = phoebe_set_parameter_value (qualifier, idx, atof (value_str));
		break;
		case TYPE_STRING_ARRAY:
			while (value_str[0] == '"') value_str++;
			while (value_str[strlen(value_str)-1] == '"') value_str[strlen(value_str)-1] = '\0';
			errno = phoebe_set_parameter_value (qualifier, idx, value_str);
		break;
		}

	if (errno != 0)
		phoebe_warning ("qualifier '%s[%d]' cannot be set!\n", qualifier, idx+1);
	}

int open_keyword_file (const char *filename)
	{
	/* This function opens PHOEBE 0.3x keyword files. The return value is:      */
	/*                                                                          */
	/*   -5 .. keyword file is an empty file                                    */
	/*   -4 .. invalid header line of a keyword file (PHOEBE without version).  */
	/*   -3 .. the specified file isn't a regular file (probably a directory).  */
	/*   -2 .. the specified file doesn't have read permissions.                */
	/*   -1 .. the specified file doesn't exist.                                */
	/*    0 .. everything OK, file readout successful.                          */
	/*    1 .. keywords within a file don't contain '=' signs; this is usually  */
	/*         the case when old keyword files or data files are being opened.  */

	char readout_string[256];
	char *readout_str = readout_string;

	char keyword_string[256];
	char *keyword_str = keyword_string;

	char *value_str;

	int i;
	int lineno = 0;

	FILE *keyword_file;

	phoebe_debug ("entering function 'open_keyword_file ()'\n");

	/* First a checkup if everything is OK with the filename:                   */
	if (!filename_is_regular_file ((char *) filename))      return -3;
	if (!filename_has_read_permissions ((char *) filename)) return -2;
	if (!filename_exists ((char *) filename))               return -1;

	keyword_file = fopen (filename, "r");

	/* Keyword files start with a commented header line; see if it's there:   	*/

	fgets (readout_str, 255, keyword_file); lineno++;
	if (feof (keyword_file)) return -5;
	readout_str[strlen(readout_str)-1] = '\0';

	if (strstr (readout_str, "PHOEBE") != 0)
		{
		/* Yep, it's there! Read out version number and if it's a legacy file,    */
		/* call a suitable function for opening legacy keyword files.             */

		char *version_str = strstr (readout_str, "PHOEBE");
		double version;
			
		if (sscanf (version_str, "PHOEBE %lf", &version) != 1)
			{
			/* Just in case, if the header line is invalid.                       */
			phoebe_warning ("Invalid header line in %s, aborting.\n", filename);
			return -4;
			}
		if (version < 0.30)
			{
			fclose (keyword_file);
			phoebe_notice ("opening legacy keyword file.\n");
			open_legacy_keyword_file (filename);
			return 1;
			}
		phoebe_debug ("PHOEBE keyword file version: %2.2lf\n", version);
		}

	while (!feof (keyword_file))
		{
		fgets (readout_str, 255, keyword_file); lineno++;
		if (feof (keyword_file)) break;

		/* Clear the read string of leading and trailing spaces, tabs, newlines,  */
		/* comments and empty lines:                                              */
		readout_str[strlen(readout_str)-1] = '\0';
		if (strchr (readout_str, '#') != 0) (strchr (readout_str, '#'))[0] = '\0';
		while (readout_str[0] == ' ' || readout_str[0] == '\t') readout_str++;
		while (readout_str[strlen(readout_str)-1] == ' ' || readout_str[strlen(readout_str)-1] == '\t') readout_str[strlen(readout_str)-1] = '\0';
		if (strlen (readout_str) == 0) continue;

		/* Read out the keyword only. We can't read out the value here because it */
		/* may be a string containing spaces and that would result in an incom-   */
		/* plete readout.                                                         */
		if (sscanf (readout_str, "%s = %*s", keyword_str) != 1)
			{
			phoebe_warning ("line %d of the keyword file is invalid.\n", lineno);
			continue;
			}
		value_str = strchr (readout_str, '=');
		if (value_str == NULL)
			{
			/* If the keyword doesn't have '=', it will be skipped.                 */
			phoebe_warning ("keyword %s initialization (line %d) is invalid.\n", keyword_str, lineno);
			continue;
			}

		/* value_str now points to '=', we need the next character:               */
		value_str++;

		/* Eat all empty spaces at the beginning and at the end:                  */
		while (value_str[0] == ' ' || value_str[0] == '\t') value_str++;
		while (value_str[strlen (value_str)-1] == ' ' || value_str[strlen (value_str)-1] == '\t') value_str[strlen(value_str)-1] = '\0';

		phoebe_debug ("keyword: %s, value: %s\n", keyword_str, value_str);

		if (strchr (keyword_str, '.') != NULL)
			{
			/* This means we have an adjustable parameter which needs to be handled */
			/* differently from other parameters:                                   */

			for (i = 0; i < PHOEBE_parameters_no; i++)
				if (strncmp (keyword_str, PHOEBE_parameters[i].keyword, strlen (PHOEBE_parameters[i].keyword)) == 0) break;
			if ( (i == PHOEBE_parameters_no - 1) && (strncmp (keyword_str, PHOEBE_parameters[i].keyword, strlen (PHOEBE_parameters[i].keyword)) != 0) )
				phoebe_warning ("keyword %s is unkown, skipping.\n", keyword_str);

			if (PHOEBE_parameters[i].type == TYPE_DOUBLE)
				{
				if (strstr (keyword_str, ".VAL")  != NULL) PHOEBE_parameters[i].value.d = atof (value_str);
				if (strstr (keyword_str, ".MIN")  != NULL) PHOEBE_parameters[i].min     = atof (value_str);
				if (strstr (keyword_str, ".MAX")  != NULL) PHOEBE_parameters[i].max     = atof (value_str);
				if (strstr (keyword_str, ".STEP") != NULL) PHOEBE_parameters[i].step    = atof (value_str);
				if (strstr (keyword_str, ".ADJ")  != NULL)
					{
					if (strcmp (value_str, "YES") == 0 || strcmp (value_str, "TRUE") == 0 || strcmp (value_str, "1") == 0)
						PHOEBE_parameters[i].tba = YES;
					else
						PHOEBE_parameters[i].tba = NO;
					}
				}
			if (PHOEBE_parameters[i].type == TYPE_DOUBLE_ARRAY)
				{
				if (strstr (keyword_str, ".MIN")  != NULL) PHOEBE_parameters[i].min     = atof (value_str);
				if (strstr (keyword_str, ".MAX")  != NULL) PHOEBE_parameters[i].max     = atof (value_str);
				if (strstr (keyword_str, ".STEP") != NULL) PHOEBE_parameters[i].step    = atof (value_str);
				if (strstr (keyword_str, ".ADJ")  != NULL)
					{
					if (strcmp (value_str, "YES") == 0 || strcmp (value_str, "TRUE") == 0 || strcmp (value_str, "1") == 0)
						PHOEBE_parameters[i].tba = YES;
					else
						PHOEBE_parameters[i].tba = NO;
					}
				if (strstr (keyword_str, ".VAL")  != NULL)
					{
					char *index = strchr (keyword_str, '[');
					int idx;
					sscanf (index, "[%d]", &idx);
					PHOEBE_parameters[i].value.vec->val[idx-1] = atof (value_str);
					}
				}
			}
		else
			{
			for (i = 0; i < PHOEBE_parameters_no; i++)
				{
				if (strcmp (keyword_str, PHOEBE_parameters[i].keyword) == 0)
					{
					get_from_keyword_file (PHOEBE_parameters[i].qualifier, value_str);
					break;
					}
				if (strncmp (keyword_str, PHOEBE_parameters[i].keyword, strlen (PHOEBE_parameters[i].keyword)) == 0)
					{
					/* This is appopriate for wavelength-dependent parameters; a conse- */
					/* cutive number is attached to the end of the keyword for distinc- */
					/* tion.                                                            */

					char *index = strchr (keyword_str, '[');
					int idx;

					if (index == NULL)
						{
						phoebe_warning ("keyword %s is invalid, ignoring.\n", keyword_str);
						break;
						}

					sscanf (index, "[%d]", &idx);
					get_wavelength_dependent_parameter_from_keyword_file (PHOEBE_parameters[i].qualifier, idx-1, value_str);
					break;
					}
				}

			if ( (i == PHOEBE_parameters_no) && (strncmp (keyword_str, PHOEBE_parameters[i-1].keyword, strlen (PHOEBE_parameters[i-1].keyword)) != 0) )
				phoebe_warning ("keyword '%s' isn't recognized by PHOEBE.\n", keyword_str);
			}

		/* As explained above, we have to repoint strings:                        */
		readout_str = readout_string;
		keyword_str = keyword_string;
		}

	fclose (keyword_file);

	phoebe_debug ("leaving function 'open_keyword_file ()'\n");

	return 0;
	}

void save_to_keyword_file (char *keyword, FILE *file)
	{
	int i = phoebe_index_from_keyword (keyword);
	int j;

	if (PHOEBE_parameters[i].kind == KIND_ADJUSTABLE)
		{
		switch (PHOEBE_parameters[i].type)
			{
			case TYPE_DOUBLE:
				fprintf (file, "%s.VAL  = %lf\n", keyword, PHOEBE_parameters[i].value.d);
			break;
			case TYPE_DOUBLE_ARRAY:
				for (j = 0; j < phoebe_get_value_int (PHOEBE_parameters[i].dependency); j++)
					fprintf (file, "%s[%d].VAL  = %lf\n", keyword, j+1, PHOEBE_parameters[i].value.vec->val[j]);
			break;
			}
		fprintf (file, "%s.ADJ  = %d\n",  keyword, PHOEBE_parameters[i].tba);
		fprintf (file, "%s.STEP = %lf\n", keyword, PHOEBE_parameters[i].step);
		fprintf (file, "%s.MIN  = %lf\n", keyword, PHOEBE_parameters[i].min);
		fprintf (file, "%s.MAX  = %lf\n", keyword, PHOEBE_parameters[i].max);
		}
	else
		{
		switch (PHOEBE_parameters[i].type)
			{
			case TYPE_INT:
				fprintf (file, "%s = %d\n", keyword, PHOEBE_parameters[i].value.i);
			break;
			case TYPE_BOOL:
				fprintf (file, "%s = %d\n", keyword, PHOEBE_parameters[i].value.b);
			break;
			case TYPE_DOUBLE:
				fprintf (file, "%s = %lf\n", keyword, PHOEBE_parameters[i].value.d);
			break;
			case TYPE_STRING:
				fprintf (file, "%s = \"%s\"\n", keyword, PHOEBE_parameters[i].value.str);
			break;
			case TYPE_INT_ARRAY:
				for (j = 0; j < phoebe_get_value_int (PHOEBE_parameters[i].dependency); j++)
					fprintf (file, "%s[%d] = %d\n", keyword, j+1, PHOEBE_parameters[i].value.iarray[j]);
			break;
			case TYPE_BOOL_ARRAY:
				for (j = 0; j < phoebe_get_value_int (PHOEBE_parameters[i].dependency); j++)
					fprintf (file, "%s[%d] = %d\n", keyword, j+1, PHOEBE_parameters[i].value.barray[j]);
			break;
			case TYPE_DOUBLE_ARRAY:
				for (j = 0; j < phoebe_get_value_int (PHOEBE_parameters[i].dependency); j++)
					fprintf (file, "%s[%d] = %lf\n", keyword, j+1, PHOEBE_parameters[i].value.vec->val[j]);
			break;
			case TYPE_STRING_ARRAY:
				for (j = 0; j < phoebe_get_value_int (PHOEBE_parameters[i].dependency); j++)
					fprintf (file, "%s[%d] = \"%s\"\n", keyword, j+1, PHOEBE_parameters[i].value.strarray[j]);
			break;
			}
		}
	}

int save_keyword_file (const char *filename)
	{
	/* This function saves PHOEBE 0.3x keyword files. The return value is:      */
	/*                                                                          */
	/*   -3 .. the specified file isn't a regular file (probably a directory).  */
	/*   -2 .. the specified file doesn't have write permissions.               */
	/*    0 .. everything OK, file readout successful.                          */

	FILE *keyword_file;
	int i;

	/* First a checkup if everything is OK with the filename:                   */
	if (!filename_is_regular_file ((char *) filename))       return -3;
	if (!filename_has_write_permissions ((char *) filename)) return -2;

	keyword_file = fopen (filename, "w");

	/* Write a version header:                                                  */
	fprintf (keyword_file, "# Parameter file conforming to %s\n", PHOEBE_VERSION_NUMBER);

	for (i = 0; i < PHOEBE_parameters_no; i++)
		save_to_keyword_file (PHOEBE_parameters[i].keyword, keyword_file);

	fclose (keyword_file);
	return 0;
	}

int open_legacy_keyword_file (const char *filename)
	{
	phoebe_warning ("Not yet implemented!\n");
	return -1;
	}
