/* This is PHOEBE scripter's Abstract Syntax Tree (AST) implementation.       */

#include "phoebe_build_config.h"

#include <limits.h>
#include <math.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "phoebe_global.h"

#include "phoebe_scripter_ast.h"
#include "phoebe_scripter_commands.h"
#include "phoebe_scripter_directives.h"
#include "phoebe_scripter_grammar.h"

#include "phoebe_accessories.h"
#include "phoebe_allocations.h"
#include "phoebe_calculations.h"
#include "phoebe_error_handling.h"
#include "phoebe_fortran_interface.h"
#include "phoebe_keywords.h"
#include "phoebe_type_handling.h"

scripter_ast *scripter_ast_add_int (const int val)
	{
	scripter_ast *out  = malloc (sizeof (*out));
	out->type          = ast_int;
	out->value.integer = val;
	return out;
	}

scripter_ast *scripter_ast_add_double (const double val)
	{
	scripter_ast *out = malloc (sizeof (*out));
	out->type         = ast_double;
	out->value.real   = val;
	return out;
	}

scripter_ast *scripter_ast_add_bool (const bool val)
	{
	scripter_ast *out  = malloc (sizeof (*out));
	out->type          = ast_bool;
	out->value.boolean = val;
	return out;
	}

scripter_ast *scripter_ast_add_string (const char *val)
	{
	scripter_ast *out = malloc (sizeof (*out));
	out->type         = ast_string;
	out->value.string = strdup (val);
	return out;
	}

scripter_ast *scripter_ast_add_array (PHOEBE_array *array)
	{
	scripter_ast *out = malloc (sizeof (*out));
	out->type         = ast_array;
	out->value.array  = phoebe_array_duplicate (array);
	return out;
	}

scripter_ast *scripter_ast_add_vector (PHOEBE_vector *vec)
	{
	scripter_ast *out;

	out = phoebe_malloc (sizeof (*out));
	out->type         = ast_vector;
	out->value.vec    = phoebe_vector_duplicate (vec);
	return out;
	}

scripter_ast *scripter_ast_add_data (PHOEBE_data *data)
	{
	phoebe_scripter_output ("1) IMPLEMENT ME!\n");
	}

scripter_ast *scripter_ast_add_variable (const char *val)
	{
	scripter_ast *out   = malloc (sizeof (*out));
	out->type           = ast_variable;
	out->value.variable = strdup (val);
	return out;
	}

scripter_ast *scripter_ast_add_function (const char *val)
	{
	scripter_ast *out   = malloc (sizeof (*out));
	out->type           = ast_function;
	out->value.variable = strdup (val);
	return out;
	}

scripter_ast *scripter_ast_add_node (const scripter_ast_kind kind, scripter_ast_list *args)
	{
	scripter_ast *out    = malloc (sizeof (*out));
	out->type            = ast_node;
	out->value.node.kind = kind;
	out->value.node.args = args;
	return out;
	}

scripter_ast_list *scripter_ast_construct_list (scripter_ast *ast, scripter_ast_list *list)
	{
	scripter_ast_list *out = malloc (sizeof (*out));
	out->elem = ast;
	out->next = list;
	return out;
	}

scripter_ast_list *scripter_ast_reverse_list (scripter_ast_list *r, scripter_ast_list *s)
	{
	/* When defining a function, the list is filled from the last element to    */
	/* the first, e.g. f(x,y,z) would create z->y->x parentage. This isn't what */
	/* we want and the following function reverses this order through recursion */
	/* by calling itself. The result after that is x->y->z, as it should be.    */

	if (r == 0) return s;
	return scripter_ast_reverse_list (r->next, scripter_ast_construct_list (r->elem, s));
	}

int scripter_ast_list_length (scripter_ast_list *in)
	{
	int i = 0;
	for (; in != 0; in = in->next) i++;
	return i;
	}

void scripter_ast_print_list (scripter_ast_list *in)
	{
	if (in == 0) return;
	fprintf (PHOEBE_output, "\t");
	scripter_ast_print (in->elem);
	scripter_ast_print_list (in->next);
	};

void scripter_ast_print (scripter_ast *in)
	{
	switch (in->type)
		{
		case ast_double:
			fprintf (PHOEBE_output, "%lf", in->value.real);
		break;
		case ast_variable:
			fprintf (PHOEBE_output, "%s", in->value.variable);
		break;
		case ast_string:
			fprintf (PHOEBE_output, "%s", in->value.string);
		break;
		case ast_node:
			scripter_ast_print_list (in->value.node.args);
		break;
		}
	}

scripter_ast_value scripter_ast_evaluate (scripter_ast *in)
	{
	/* This is the most important function of the AST implementation. Based on  */
	/* the AST leaf type, it decides what to do and how to do it. It returns    */
	/* any type (the union) and the function must decide which union field to   */
	/* use based on the AST type.                                               */

	scripter_ast_value out;

	if (in == NULL)
		{
		out.type = type_void;
		return out;
		}

	switch (in->type)
		{
		case ast_int:
			{
			/* It is an unsigned integer value. Return it.                          */
			out.type    = type_int;
			out.value.i = in->value.integer;
			return out;
			}
		case ast_double:
			{
			/* It is a numeric value. Return it without any complications.          */
			out.type    = type_double;
			out.value.d = in->value.real;
			return out;
			}
		case ast_bool:
			{
			/* It's a boolean (switched) value, which may be TRUE or FALSE.         */
			out.type    = type_bool;
			out.value.b = in->value.boolean;
			return out;
			}
		case ast_string:
			{
			out.type      = type_string;
			out.value.str = strdup (in->value.string);
			return out;
			}
		case ast_element:
			{
			/* It is an element of a vector given by the variable name and index.   */
			phoebe_scripter_output ("2) IMPLEMENT ME!\n");
			}
		case ast_array:
			{
			out.type        = type_array;
			out.value.array = phoebe_array_duplicate (in->value.array);
			return out;
			}
		case ast_vector:
			{
			out.type      = type_vector;
			out.value.vec = phoebe_vector_duplicate (in->value.vec);
			return out;
			}
		case ast_data:
			{
			phoebe_scripter_output ("3) IMPLEMENT ME!\n");
			}
		case ast_variable:
			{
			/* It is a variable.                                                    */
			scripter_symbol *s = scripter_st_identifier_lookup (in->value.variable);
			if (s == 0)
				{
				phoebe_scripter_output ("variable '%s' is not initialized.\n", in->value.variable);
				out.type = type_void; return out;
				}
			if (s->link == 0)
				{
				phoebe_scripter_output ("variable '%s' is empty.\n", in->value.variable);
				out.type = type_void; return out;
				}
			if (s->link->type == ast_string)
				{
				out.type      = type_string;
				out.value.str = s->link->value.string;
				}
			else if (s->link->type == ast_int)
				{
				out.type      = type_int;
				out.value.i   = s->link->value.integer;
				}
			else if (s->link->type == ast_double)
				{
				out.type      = type_double;
				out.value.d   = s->link->value.real;
				}
			else if (s->link->type == ast_bool)
				{
				out.type      = type_bool;
				out.value.b   = s->link->value.boolean;
				}
			else if (s->link->type == ast_array)
				{
				out.type        = type_array;
				out.value.array = s->link->value.array;
				}
			else if (s->link->type == ast_vector)
				{
				out.type      = type_vector;
				out.value.vec = s->link->value.vec;
				}
			else if (s->link->type == ast_data)
				{
				phoebe_scripter_output ("4) IMPLEMENT ME!\n");
				}
			else
				{
				phoebe_scripter_output ("variable '%s' doesn't hold a numeric or string value.\n", in->value.variable);
				out.type      = type_void;
				}

			return out;
			}
		case ast_node:
			if (in->value.node.kind == kind_statement)
				{
				/* This is the most elementary building block of PHOEBE scripter.     */

				if (in->value.node.args->next == NULL)
					scripter_ast_evaluate (in->value.node.args->elem);
				else
					{
					scripter_ast_value fname = scripter_ast_evaluate (in->value.node.args->next->elem);

					if (fname.type != type_string)
						{
						phoebe_scripter_output ("attempted redirection to non-string identifier, reverting to stdout.\n");
						PHOEBE_output = stdout;
						scripter_ast_evaluate (in->value.node.args->elem);
						out.type = type_void;
						return out;
						}

					PHOEBE_output = fopen (fname.value.str, in->value.node.args->next->next->elem->value.string);

					if (PHOEBE_output == NULL)
						{
				    PHOEBE_output = stdout;
						phoebe_scripter_output ("redirection to \"%s\" failed, reverting to stdout.\n", in->value.node.args->next->elem->value.string);
						scripter_ast_evaluate (in->value.node.args->elem);
						}
					else
						{
						scripter_ast_evaluate (in->value.node.args->elem);
				    fclose (PHOEBE_output);
				    PHOEBE_output = stdout;
						}
					}

				out.type = type_void;
				return out;
				}
			if (in->value.node.kind == kind_block)
				{
				/* This means that the user passed the whole block of statements en-  */
				/* closed in curly braces. If the return directive is encountered,    */
				/* a block gets a non-void value; otherwise it is void.               */

				scripter_ast_list *s = in->value.node.args;
				scripter_ast_value val;

				if (s == 0)
					{
					phoebe_scripter_output ("redundant block.\n");
					out.type = type_void;
					return out;
					}
				while (s != 0)
					{
					if (scripter_return_flag == YES) break;
					val = scripter_ast_evaluate (s->elem);
					s = s->next;
					}

				out.type = type_void;
				return out;
				}
			if (in->value.node.kind == kind_return)
				{
				scripter_ast_value val = scripter_ast_evaluate (in->value.node.args->elem);
				scripter_return_flag = YES;

				switch (val.type)
					{
					case type_int:
						scripter_return_value.type = type_int;
						scripter_return_value.value.i = val.value.i;
					break;
					case type_double:
						scripter_return_value.type = type_double;
						scripter_return_value.value.d = val.value.d;
					break;
					case type_array:
						scripter_return_value.type = type_array;
						scripter_return_value.value.array = phoebe_array_duplicate (val.value.array);
					break;
					case type_vector:
						scripter_return_value.type = type_vector;
						scripter_return_value.value.vec = phoebe_vector_duplicate (val.value.vec);
					break;
					default:
						scripter_return_value.type = type_void;
					break;
					}

				out.type = type_void;
				return out;
				}
			if (in->value.node.kind == kind_system_call)
				{
				/* This means the user used the '!' delimeter to execute a system     */
				/* command. The return value is always void.                          */

				char *syscommand = in->value.node.args->elem->value.string;
				system (syscommand);
				out.type = type_void; return out;
				}
			if (in->value.node.kind == kind_os_ls)
				{
				/* This is the 'ls' command.                                          */
				out.type = type_void;

				if (strlen (in->value.node.args->elem->value.string) == 0)
					system ("ls --color=auto");
				else
					{
					char com[255];
					sprintf (com, "ls --color=auto %s", in->value.node.args->elem->value.string);
					system (com);
					return out;
					}
				}
			if (in->value.node.kind == kind_os_cd)
				{
				out.type = type_void;

				if (strlen (in->value.node.args->elem->value.string) == 0)
					{
					int errno = chdir (USER_HOME_DIR);
					if (errno != 0)
						phoebe_scripter_output ("Changing directory to $HOME failed, aborting.\n");
					return out;
					}
				else
					{
					int errno = chdir (in->value.node.args->elem->value.string);
					if (errno != 0)
						phoebe_scripter_output ("changing directory to %s failed, aborting.\n", in->value.node.args->elem->value.string);
					return out;
					}
				}
			if (in->value.node.kind == kind_os_pwd)
				{
				char cwd[255];
				out.type = type_void;
				getcwd (cwd, 255);
				fprintf (PHOEBE_output, "%s\n", cwd);
				}
			if (in->value.node.kind == kind_while)
				{
				/* This is the 'while' loop. The return type is always void.          */

				scripter_ast_value condition = scripter_ast_evaluate (in->value.node.args->elem);
				out.type = type_void;

				if (condition.type != type_bool)
					{
					phoebe_scripter_output ("the 'while' loop condition is not boolean.\n");
					return out;
					}

				while (condition.value.b == TRUE)
					{
					scripter_ast_evaluate (in->value.node.args->next->elem);
					condition = scripter_ast_evaluate (in->value.node.args->elem);
					}
				return out;
				}
			if (in->value.node.kind == kind_for)
				{
				/* This is the 'for' loop. It always returns void.                    */

				/* The following line will always evaluate successfully because of    */
				/* the grammar rules:                                                 */
				char  *ident = in->value.node.args->elem->value.variable;
				scripter_ast_value value = scripter_ast_evaluate (in->value.node.args->next->elem);
				scripter_ast_value cond;

				scripter_ast *action = in->value.node.args->next->next->next->elem;
				scripter_ast *block  = in->value.node.args->next->next->next->next->elem;

				out.type = type_void;

				/* Check whether a value is a numeric value:                          */
				if (value.type != type_int && value.type != type_double)
					{
					phoebe_scripter_output ("the initial value is not a numeric value, aborting.\n");
					return out;
					}

				/* Create a symbol for iteration and initialize it; since this is a   */
				/* index variable, its scope is local.                                */

				scripter_ast_flow_begin_scope ();

				if (value.type == type_int)
					scripter_st_identifier_commit (ident, scripter_ast_add_int (value.value.i), scripter_flow_scope_level);
				if (value.type == type_double)
					scripter_st_identifier_commit (ident, scripter_ast_add_double (value.value.d), scripter_flow_scope_level);

				/* Evaluate a condition:                                              */
				cond = scripter_ast_evaluate (in->value.node.args->next->next->elem);
				if (cond.type != type_bool)
					{
					phoebe_scripter_output ("the condition is not boolean, so it cannot be evaluated.\n");
					return out;
					}

				/* Enter the loop:                                                    */
				while (cond.value.b == TRUE)
					{
					scripter_ast_evaluate (block);
					scripter_ast_evaluate (action);
					cond = scripter_ast_evaluate (in->value.node.args->next->next->elem);
					}

				/* Close the loop scope:                                              */
				scripter_ast_flow_end_scope ();

				/* Wrap it up and exit:                                               */
				return out;
				}

			if (in->value.node.kind == kind_unaryp)
				{
				/* This means that we have a unary '+'.                               */
				scripter_ast_value val = scripter_ast_evaluate (in->value.node.args->elem);

				switch (val.type)
					{
					case type_int:
						return val;
					break;
					case type_double:
						return val;
					break;
					case type_vector:
						return val;
					break;
					}

				phoebe_scripter_output ("unary '+' operator operates only on numeric types.\n");
				val.type = type_void;
				return val;
				}
			if (in->value.node.kind == kind_unarym)
				{
				/* This means that we have a unary '-'.                               */
				scripter_ast_value val = scripter_ast_evaluate (in->value.node.args->elem);

				switch (val.type)
					{
					case type_int:
						out.type = type_int;
						out.value.i = -val.value.i;
						return out;
					break;
					case type_double:
						out.type = type_double;
						out.value.d = -val.value.d;
						return out;
					break;
					case type_vector:
						out.type = type_vector;
						out.value.vec = phoebe_vector_duplicate (val.value.vec);
						phoebe_vector_multiply_by (out.value.vec, -1.0);
						return out;
					break;
					}

				phoebe_scripter_output ("unary '-' operator operates only on numeric types.\n");
				out.type = type_void;
				return out;
				}

			if (in->value.node.kind == kind_not)
				{
				/* This means that we have a logical NOT.                             */

				scripter_ast_value val = scripter_ast_evaluate (in->value.node.args->elem);

				if (val.type != type_bool)
					{
					phoebe_scripter_output ("the unary '!' operates only on boolean expressions, aborting.\n");
					out.type = type_void;
					return out;
					}

				out.type = type_bool;
				if ( val.value.b == TRUE ) out.value.b = FALSE;
					else out.value.b = TRUE;

				return out;
				}

			if (in->value.node.kind == kind_inc)
				{
				/* This means that we have the increment operator on the variable.    */

				char *ident = in->value.node.args->elem->value.variable;
				scripter_symbol *symb = scripter_st_identifier_lookup (ident);
				scripter_ast_value value;

				out.type = type_void;
				if (symb == 0)
					{
					phoebe_scripter_output ("identifier '%s' not initialized, aborting.\n", ident);
					return out;
					}

				value = scripter_ast_evaluate (symb->link);
				if (value.type != type_int && value.type != type_double)
					{
					phoebe_scripter_output ("identifier '%s' doesn't hold a numeric value.\n", ident);
					return out;
					}

				if (value.type == type_int)
					{
					int i = value.value.i + 1;
					scripter_st_identifier_commit (ident, scripter_ast_add_int (i), scripter_flow_scope_level);
					}
				if (value.type == type_double)
					{
					double d = value.value.d + 1.0;
					scripter_st_identifier_commit (ident, scripter_ast_add_double (d), scripter_flow_scope_level);
					}

				return out;
				}
			if (in->value.node.kind == kind_dec)
				{
				/* This means that we have the decrement operator on the variable.    */

				char *ident = in->value.node.args->elem->value.variable;
				scripter_symbol *symb = scripter_st_identifier_lookup (ident);
				scripter_ast_value value;

				out.type = type_void;
				if (symb == 0)
					{
					phoebe_scripter_output ("identifier '%s' not initialized, aborting.\n", ident);
					return out;
					}

				value = scripter_ast_evaluate (symb->link);
				if (value.type != type_int && value.type != type_double)
					{
					phoebe_scripter_output ("identifier '%s' doesn't hold a numeric value.\n", ident);
					return out;
					}

				if (value.type == type_int)
					{
					int i = value.value.i - 1;
					scripter_st_identifier_commit (ident, scripter_ast_add_int (i), scripter_flow_scope_level);
					}
				if (value.type == type_double)
					{
					double d = value.value.d - 1.0;
					scripter_st_identifier_commit (ident, scripter_ast_add_double (d), scripter_flow_scope_level);
					}

				return out;
				}
			if (in->value.node.kind == kind_incby)
				{
				/* This means that we have the increment-by operator on the variable. */

				char   *ident         = in->value.node.args->elem->value.variable;
				scripter_symbol *symb = scripter_st_identifier_lookup (ident);
				scripter_ast_value by = scripter_ast_evaluate (in->value.node.args->next->elem);
				scripter_ast_value value;

				out.type = type_void;
				if (symb == 0)
					{
					phoebe_scripter_output ("identifier '%s' not initialized, aborting.\n", ident);
					return out;
					}
				if (by.type != type_int && by.type != type_double)
					{
					phoebe_scripter_output ("increment doesn't hold a numeric value.\n", ident);
					return out;
					}
				value = scripter_ast_evaluate (symb->link);
				if (value.type != type_int && value.type != type_double)
					{
					phoebe_scripter_output ("identifier '%s' doesn't hold a numeric value.\n", ident);
					return out;
					}

				if (value.type == type_int)
					{
					int i;
					if (by.type == type_double)
						{
						int val = (int) by.value.d; by.type = type_int; by.value.i = val;
						}
					i = value.value.i + by.value.i;
					scripter_st_identifier_commit (ident, scripter_ast_add_int (i), scripter_flow_scope_level);
					}
				if (value.type == type_double)
					{
					double d;
					if (by.type == type_int)
						{
						double val = (double) by.value.i; by.type = type_double; by.value.d = val;
						}
					d = value.value.d + by.value.d;
					scripter_st_identifier_commit (ident, scripter_ast_add_double (d), scripter_flow_scope_level);
					}

				return out;
				}
			if (in->value.node.kind == kind_decby)
				{
				/* This means that we have the decrement-by operator on the variable. */

				char   *ident         = in->value.node.args->elem->value.variable;
				scripter_symbol *symb = scripter_st_identifier_lookup (ident);
				scripter_ast_value by = scripter_ast_evaluate (in->value.node.args->next->elem);
				scripter_ast_value value;

				out.type = type_void;
				if (symb == 0)
					{
					phoebe_scripter_output ("identifier '%s' not initialized, aborting.\n", ident);
					return out;
					}
				if (by.type != type_int && by.type != type_double)
					{
					phoebe_scripter_output ("increment doesn't hold a numeric value.\n", ident);
					return out;
					}
				value = scripter_ast_evaluate (symb->link);
				if (value.type != type_int && value.type != type_double)
					{
					phoebe_scripter_output ("identifier '%s' doesn't hold a numeric value.\n", ident);
					return out;
					}

				if (value.type == type_int)
					{
					int i;
					if (by.type == type_double)
						{
						int val = (int) by.value.d; by.type = type_int; by.value.i = val;
						}
					i = value.value.i - by.value.i;
					scripter_st_identifier_commit (ident, scripter_ast_add_int (i), scripter_flow_scope_level);
					}
				if (value.type == type_double)
					{
					double d;
					if (by.type == type_int)
						{
						double val = (double) by.value.i; by.type = type_double; by.value.d = val;
						}
					d = value.value.d - by.value.d;
					scripter_st_identifier_commit (ident, scripter_ast_add_double (d), scripter_flow_scope_level);
					}

				return out;
				}
			if (in->value.node.kind == kind_multby)
				{
				/* This means that we have the multiply-by operator on the variable.  */

				char   *ident         = in->value.node.args->elem->value.variable;
				scripter_symbol *symb = scripter_st_identifier_lookup (ident);
				scripter_ast_value by = scripter_ast_evaluate (in->value.node.args->next->elem);
				scripter_ast_value value;

				out.type = type_void;
				if (symb == 0)
					{
					phoebe_scripter_output ("identifier '%s' not initialized, aborting.\n", ident);
					return out;
					}
				if (by.type != type_int && by.type != type_double)
					{
					phoebe_scripter_output ("increment doesn't hold a numeric value.\n", ident);
					return out;
					}
				value = scripter_ast_evaluate (symb->link);
				if (value.type != type_int && value.type != type_double)
					{
					phoebe_scripter_output ("identifier '%s' doesn't hold a numeric value.\n", ident);
					return out;
					}

				if (value.type == type_int)
					{
					int i;
					if (by.type == type_double)
						{
						int val = (int) by.value.d; by.type = type_int; by.value.i = val;
						}
					i = value.value.i * by.value.i;
					scripter_st_identifier_commit (ident, scripter_ast_add_int (i), scripter_flow_scope_level);
					}
				if (value.type == type_double)
					{
					double d;
					if (by.type == type_int)
						{
						double val = (double) by.value.i; by.type = type_double; by.value.d = val;
						}
					d = value.value.d * by.value.d;
					scripter_st_identifier_commit (ident, scripter_ast_add_double (d), scripter_flow_scope_level);
					}

				return out;
				}
			if (in->value.node.kind == kind_divby)
				{
				/* This means that we have the divide-by operator on the variable.    */

				char   *ident         = in->value.node.args->elem->value.variable;
				scripter_symbol *symb = scripter_st_identifier_lookup (ident);
				scripter_ast_value by = scripter_ast_evaluate (in->value.node.args->next->elem);
				scripter_ast_value value;

				out.type = type_void;
				if (symb == 0)
					{
					phoebe_scripter_output ("identifier '%s' not initialized, aborting.\n", ident);
					return out;
					}
				if (by.type != type_int && by.type != type_double)
					{
					phoebe_scripter_output ("increment doesn't hold a numeric value.\n", ident);
					return out;
					}
				value = scripter_ast_evaluate (symb->link);
				if (value.type != type_int && value.type != type_double)
					{
					phoebe_scripter_output ("identifier '%s' doesn't hold a numeric value.\n", ident);
					return out;
					}

				if (value.type == type_int)
					{
					int i;
					if (by.type == type_int)
						scripter_st_identifier_commit (ident, scripter_ast_add_double ((double) value.value.i / (double) by.value.i), scripter_flow_scope_level);
					if (by.type == type_double)
						scripter_st_identifier_commit (ident, scripter_ast_add_double ((double) value.value.i / by.value.d), scripter_flow_scope_level);
					}
				if (value.type == type_double)
					{
					double d;
					if (by.type == type_int)
						{
						double val = (double) by.value.i; by.type = type_double; by.value.d = val;
						}
					d = value.value.d / by.value.d;
					scripter_st_identifier_commit (ident, scripter_ast_add_double (d), scripter_flow_scope_level);
					}

				return out;
				}

			if (in->value.node.kind == kind_element)
				{
				/* This function evaluates the 'var[index]' expression to the corres- */
				/* ponding double.                                                    */

				char *ident = in->value.node.args->elem->value.variable;
				scripter_ast_value index = scripter_ast_evaluate (in->value.node.args->next->elem);
				scripter_symbol *var = scripter_st_identifier_lookup (ident);

				if (var == 0)
					{
					phoebe_scripter_output ("variable '%s' is not initialized, aborting\n", ident);
					out.type = type_void; return out;
					}
				if (var->link->type != ast_vector)
					{
					phoebe_scripter_output ("variable '%s' is not a vector, aborting\n", ident);
					out.type = type_void; return out;
					}
				if (index.type != type_int)
					{
					phoebe_scripter_output ("element index is not an integer, aborting.\n");
					out.type = type_void;
					return out;
					}
				if (index.value.i < 1 || index.value.i > var->link->value.vec->dim)
					{
					phoebe_scripter_output ("index %d of vector %s is out of range [%d, %d], aborting.\n", index.value.i, ident, 1, var->link->value.vec->dim);
					out.type = type_void; return out;
					}

				out.type = type_double;
				out.value.d = var->link->value.vec->val[index.value.i-1];
				return out;
				}

			if (in->value.node.kind == kind_export)
				{
				/* This is the 'export' directive. Its purpose is to globalise local  */
				/* variables used e.g. in loops, macro definitions, functions etc.    */

				scripter_symbol *s;
				int hash;

				/* The following assignment cannot fail because of the grammar:       */
				char *ident = in->value.node.args->elem->value.variable;

				/* The result of export is always void:                               */
				out.type = type_void;

				/* Check whether by any chance the user is trying to export a global  */
				/* variable:                                                          */
				if (scripter_flow_scope_level == 0)
					{
					phoebe_scripter_output ("attempting to use export in global scope, ignoring.\n");
					return out;
					}

				hash = scripter_st_identifier_hash (ident);

				s = scripter_st[hash];
				while (s != NULL)
					{
					if ( strcmp (s->name, ident) == 0 && s->depth == scripter_flow_scope_level ) break;
					s = s->next;
					}

				if (s == NULL)
					{
					phoebe_scripter_output ("variable %s is not locally declared, aborting.\n", ident);
					return out;
					}
				else
					scripter_st_identifier_commit (ident, s->link, scripter_flow_scope_level - 1);
				}

			if (in->value.node.kind == kind_set)
				{
				/* This is a set statement which assigns a value to the variable.     */
				/* There are two modes of this directive operation: one is direct,    */
				/* i.e. set i=1, and the other mode is by using a variable operator,  */
				/* i.e. set i++. The return value is always void.                     */

				scripter_ast_value value;
				out.type = type_void;

				/* First the direct method; it means it has more than 1 argument.     */
				if (in->value.node.args->next != 0)
					{
					/* The following line always evaluates because of the grammar:      */
					char *ident = in->value.node.args->elem->value.variable;
					value = scripter_ast_evaluate (in->value.node.args->next->elem);

					if (value.type == type_int)
						{
						scripter_st_identifier_commit (ident, scripter_ast_add_int (value.value.i), scripter_flow_scope_level);
						return out;
						}
					if (value.type == type_bool)
						{
						scripter_st_identifier_commit (ident, scripter_ast_add_bool (value.value.b), scripter_flow_scope_level);
						return out;
						}
					if (value.type == type_double)
						{
						scripter_st_identifier_commit (ident, scripter_ast_add_double (value.value.d), scripter_flow_scope_level);
						return out;
						}
					if (value.type == type_string)
						{
						scripter_st_identifier_commit (ident, scripter_ast_add_string (value.value.str), scripter_flow_scope_level);
						return out;
						}
					if (value.type == type_array)
						{
						scripter_st_identifier_commit (ident, scripter_ast_add_array (value.value.array), scripter_flow_scope_level);
						return out;
						}
					if (value.type == type_vector)
						{
						scripter_st_identifier_commit (ident, scripter_ast_add_vector (value.value.vec), scripter_flow_scope_level);
						return out;
						}
					if (value.type == type_data)
						{
						phoebe_scripter_output ("5) IMPLEMENT ME!\n");
						}
					if (value.type == type_void)
						{
						phoebe_scripter_output ("assignment failed.\n");
						return out;
						}
					}

				/* And now the second method; individual symbol table updates are     */
				/* called in dedicated functions (for ++, --, ...).                   */
				if (in->value.node.args->next == 0)
					value = scripter_ast_evaluate (in->value.node.args->elem);

				return out;
				}

			if (in->value.node.kind == kind_set_element)
				{
				char *ident = in->value.node.args->elem->value.variable;
				scripter_ast_value index = scripter_ast_evaluate (in->value.node.args->next->elem);
				scripter_ast_value value = scripter_ast_evaluate (in->value.node.args->next->next->elem);
				scripter_symbol *var = scripter_st_identifier_lookup (ident);

				if (index.type != type_int)
					{
					phoebe_scripter_output ("vector index is not an integer, aborting.\n");
					out.type = type_void;
					return out;
					}
				if (value.type != type_int && value.type != type_double)
					{
					phoebe_scripter_output ("element value must be a real value, aborting.\n");
					out.type = type_void;
					return out;
					}
				if (value.type == type_int)
					{
					/* Propagate this value to double:                                  */
					int val = value.value.i;
					value.type = type_double;
					value.value.d = (double) val;
					}
				if (var == 0)
					{
					phoebe_scripter_output ("variable '%s' is not initialized, aborting\n", ident);
					out.type = type_void; return out;
					}
				if (var->link->type != ast_vector)
					{
					phoebe_scripter_output ("variable '%s' is not a vector, aborting\n", ident);
					out.type = type_void; return out;
					}
				if (index.value.i < 1 || index.value.i > var->link->value.vec->dim)
					{
					phoebe_scripter_output ("index %d for '%s' is out of range [%d, %d], aborting.\n", index.value.i, ident, 1, var->link->value.vec->dim);
					out.type = type_void; return out;
					}
				index.value.i--;

				phoebe_debug ("%s[%d] = %f\n", ident, index.value.i, value.value.d);
				var->link->value.vec->val[index.value.i] = value.value.d;
				}

			if (in->value.node.kind == kind_set_vector)
				{
				/* This part is executed when 'set ident = array (dim)' is issued.    */

				char *ident = in->value.node.args->elem->value.variable;
				scripter_ast_value dim = scripter_ast_evaluate (in->value.node.args->next->elem);

				PHOEBE_vector *vec;
				int status;

				/* Vectors shouldn't return any values, they are simply added to the  */
				/* symbol list. Thus we should always return type_void.               */
				out.type = type_void;

				if (dim.type != type_int)
					{
					phoebe_scripter_output ("type: %d\n", dim.type);
					phoebe_scripter_output ("vector() called with non-integer dimension, aborting.\n");
					return out;
					}

				/* Initialize a new vector, allocate memory and pad it with 0es:      */
				vec    = phoebe_vector_new ();
				status = phoebe_vector_alloc (vec, dim.value.i);
				if (status == -1)
					{
					phoebe_scripter_output ("vector() called with invalid dimension, aborting.\n");
					phoebe_vector_free (vec);
					return out;
					}

				phoebe_vector_pad (vec, 0.0);

				scripter_st_identifier_commit (ident, scripter_ast_add_vector (vec), scripter_flow_scope_level);
				phoebe_vector_free (vec);
				return out;
				}

			if (in->value.node.kind == kind_data)
				{
				phoebe_scripter_output ("5) IMPLEMENT ME!\n");
				}

			if (in->value.node.kind == kind_builtin)
				{
				/* This means we have a built-in mathematical function. The returned  */
				/* value is double if everything is ok and void if an error occured.  */

				/* This is uniquely determined, so the assignment cannot fail:        */
				char *func  = in->value.node.args->elem->value.function;
				scripter_ast_value value = scripter_ast_evaluate (in->value.node.args->next->elem);

				switch (value.type)
					{
					case type_int:
						{
						/* All functions act on double precision numbers, so all integers */
						/* must be propagated to a double type:                           */
						int val = value.value.i;
						value.type = type_double;
						value.value.d = (double) val;
						}
					/* Don't break it here, otherwise the propagation won't work! */
					case type_double:
						if (strcmp (func,   "sin") == 0) out.value.d = sin   (value.value.d);
						if (strcmp (func,   "cos") == 0) out.value.d = cos   (value.value.d);
						if (strcmp (func,   "tan") == 0) out.value.d = tan   (value.value.d);
						if (strcmp (func,  "asin") == 0) out.value.d = asin  (value.value.d);
						if (strcmp (func,  "acos") == 0) out.value.d = acos  (value.value.d);
						if (strcmp (func,  "atan") == 0) out.value.d = atan  (value.value.d);
						if (strcmp (func,   "exp") == 0) out.value.d = exp   (value.value.d);
						if (strcmp (func,    "ln") == 0) out.value.d = log   (value.value.d);
						if (strcmp (func,   "log") == 0) out.value.d = log10 (value.value.d);
						if (strcmp (func,  "sqrt") == 0) out.value.d = sqrt  (value.value.d);
						if (strcmp (func,  "norm") == 0) out.value.d =        value.value.d;
						if (strcmp (func,  "rand") == 0) out.value.d = (double) value.value.d * rand () / RAND_MAX;
						if (strcmp (func,   "int") == 0) out.value.d = (double) ((int) value.value.d);
						if (strcmp (func,  "frac") == 0) out.value.d = fabs (value.value.d - (int) value.value.d);
						if (strcmp (func,   "abs") == 0) out.value.d = fabs (value.value.d);
						out.type = type_double;
						if (strcmp (func,   "dim") == 0)
							{
							out.value.i = 1;
							out.type = type_int;
							}
						if (strcmp (func, "trunc") == 0)
							{
							out.value.i = (int) value.value.d;
							out.type = type_int;
							}
						if (strcmp (func, "round") == 0)
							{
							if (fabs (value.value.d - (int) value.value.d) < 0.5)
								out.value.i = (int) value.value.d;
							else
								if (value.value.d > 0)
									out.value.i = (int) value.value.d + 1;
								else
									out.value.i = (int) value.value.d - 1;
							out.type = type_int;
							}
					break;
					case type_vector:
						out.value.vec = phoebe_vector_new ();
						if (strcmp (func,  "sin") == 0) phoebe_vector_submit (out.value.vec, value.value.vec, sin);
						if (strcmp (func,  "cos") == 0) phoebe_vector_submit (out.value.vec, value.value.vec, cos);
						if (strcmp (func,  "tan") == 0) phoebe_vector_submit (out.value.vec, value.value.vec, tan);
						if (strcmp (func, "asin") == 0) phoebe_vector_submit (out.value.vec, value.value.vec, asin);
						if (strcmp (func, "acos") == 0) phoebe_vector_submit (out.value.vec, value.value.vec, acos);
						if (strcmp (func, "atan") == 0) phoebe_vector_submit (out.value.vec, value.value.vec, atan);
						if (strcmp (func,  "exp") == 0) phoebe_vector_submit (out.value.vec, value.value.vec, exp);
						if (strcmp (func,   "ln") == 0) phoebe_vector_submit (out.value.vec, value.value.vec, log);
						if (strcmp (func,  "log") == 0) phoebe_vector_submit (out.value.vec, value.value.vec, log10);
						if (strcmp (func, "sqrt") == 0) phoebe_vector_submit (out.value.vec, value.value.vec, sqrt);
						if (strcmp (func,  "abs") == 0) phoebe_vector_submit (out.value.vec, value.value.vec, fabs);
						out.type = type_vector;
						if (strcmp (func, "norm") == 0)
							{
							/* This one is special, since it returns double and not vector: */
							phoebe_vector_free (out.value.vec);
							phoebe_vector_norm (&(out.value.d), value.value.vec);
							out.type = type_double;
							}
						if (strcmp (func, "dim") == 0)
							{
							/* This one is special, since it returns double and not vector: */
							phoebe_vector_free (out.value.vec);
							phoebe_vector_dim (&(out.value.i), value.value.vec);
							out.type = type_int;
							}
						if (strcmp (func,  "rand") == 0 || strcmp   (func, "int") == 0 ||
						    strcmp (func,  "frac") == 0 || strcmp (func, "trunc") == 0 ||
								strcmp (func, "round") == 0 )
							{
							phoebe_vector_free (out.value.vec);
							phoebe_scripter_output ("function %s() cannot act on arrays, aborting.\n", func);
							out.type = type_void;
							return out;
							}
					break;
					default:
						phoebe_scripter_output ("function %s() called with non-numeric argument, aborting.\n", func);
						out.type = type_void;
					break;
					}

				return out;
				}

			if (in->value.node.kind == kind_help)
				{
				if (in->value.node.args == NULL)
					scripter_directive_help ("");
				else
					{
					char *command = in->value.node.args->elem->value.string;
					scripter_directive_help (command);
					}
				}

			if (in->value.node.kind == kind_get_parameter_value)
				return scripter_get_parameter_value (in->value.node.args);

			if (in->value.node.kind == kind_minimize_using_nms)
				return scripter_minimize_using_nms (in->value.node.args);

			if (in->value.node.kind == kind_minimize_using_dc)
				return scripter_minimize_using_dc (in->value.node.args);

			if (in->value.node.kind == kind_command)
				{
				/* This is PHOEBE scripter command.                                   */
				char *command = in->value.node.args->elem->value.string;

				if (strcmp (command, "open_keyword_file") == 0)
					scripter_open_keyword_file (in->value.node.args);

				if (strcmp (command, "save_keyword_file") == 0)
					scripter_save_keyword_file (in->value.node.args);

				if (strcmp (command, "set_parameter_value") == 0)
					scripter_set_parameter_value (in->value.node.args);

				if (strcmp (command, "set_parameter_limits") == 0)
					scripter_set_parameter_limits (in->value.node.args);

				if (strcmp (command, "set_parameter_step") == 0)
					scripter_set_parameter_step (in->value.node.args);

				if (strcmp (command, "mark_for_adjustment") == 0)
					scripter_mark_for_adjustment (in->value.node.args);

				if (strcmp (command, "create_wd_lci_file") == 0)
					scripter_create_wd_lci_file (in->value.node.args);

				if (strcmp (command, "compute_lc") == 0)
					scripter_compute_lc (in->value.node.args);

				if (strcmp (command, "compute_rv") == 0)
					scripter_compute_rv (in->value.node.args);

				if (strcmp (command, "minimize_using_nms") == 0)
					scripter_minimize_using_nms (in->value.node.args->next);

				if (strcmp (command, "minimize_using_dc") == 0)
					scripter_minimize_using_dc (in->value.node.args->next);

				if (strcmp (command, "kick_parameters") == 0)
					scripter_kick_parameters (in->value.node.args);

				if (strcmp (command, "phoebe_data_list") == 0)
					scripter_phoebe_data_list (in->value.node.args);

				if (strcmp (command, "phoebe_data_stats") == 0)
					scripter_phoebe_data_stats (in->value.node.args);

				if (strcmp (command, "transform_hjd_to_phase") == 0)
					scripter_transform_hjd_to_phase (in->value.node.args);

				if (strcmp (command, "transform_flux_to_magnitude") == 0)
					scripter_transform_flux_to_magnitude (in->value.node.args);

				out.type = type_void; return out;
				}

			if (in->value.node.kind == kind_proc)
				{
				/* The following line will always evaluate because of grammar rules.  */
				char *id = in->value.node.args->elem->value.variable;
				scripter_st_identifier_commit (id, in, scripter_flow_scope_level);
				out.type = type_void; return out;
				}

			if (in->value.node.kind == kind_define)
				{
				/* The following line will always evaluate because of grammar rules.  */
				char *id = in->value.node.args->elem->value.variable;
				scripter_st_identifier_commit (id, in, scripter_flow_scope_level);
				out.type = type_void; return out;
				}

			if (in->value.node.kind == kind_func)
				{
				/* This event is called by the user-defined function (the 'define'    */
				/* directive). All user-defined functions return a numeric value.     */

				scripter_ast_list *arguments = NULL;
				scripter_ast_list *params;
				scripter_ast      *body;

				scripter_ast_list *list;

				scripter_symbol *id = scripter_st_identifier_lookup (in->value.node.args->elem->value.variable);

				if (id == 0)
					{
					phoebe_scripter_output ("function '%s' not defined.\n", in->value.node.args->elem->value.variable);
					out.type = type_void; return out;
					}
				if ( id->link->type != ast_node || id->link->value.node.kind != kind_define )
					{
					phoebe_scripter_output ("identifier '%s' is not a function.\n", in->value.node.args->elem->value.variable);
					out.type = type_void; return out;
					}

				body   = id->link->value.node.args->next->elem;
				params = id->link->value.node.args->next->next;

				/* Now we have to evaluate the function arguments:                    */
				for (list = in->value.node.args->next; list != 0; list = list->next)
					{
					scripter_ast_value val = scripter_ast_evaluate (list->elem);
					if (val.type != type_double && val.type != type_int && val.type != type_vector)
						{
						phoebe_scripter_output ("non-numerical argument in definition call, aborting.\n");
						out.type = type_void; return out;
						}
					switch (val.type)
						{
						case type_int:
							arguments = scripter_ast_construct_list (scripter_ast_add_int (val.value.i), arguments);
						break;
						case type_double:
							arguments = scripter_ast_construct_list (scripter_ast_add_double (val.value.d), arguments);
						break;
						case type_vector:
							arguments = scripter_ast_construct_list (scripter_ast_add_vector (val.value.vec), arguments);
						break;
						}
					}
				arguments = scripter_ast_reverse_list (arguments, NULL);

				if (scripter_ast_list_length (params) != scripter_ast_list_length (arguments))
					{
					phoebe_scripter_output ("function '%s' called with wrong number of parameters.\n", in->value.node.args->elem->value.variable);
					out.type = type_void; return out;
					}

				/* Start a new environment where the function will be evaluated:      */
				scripter_ast_flow_begin_scope ();

				for (list = params; list != 0; list = list->next, arguments = arguments->next)
					scripter_st_identifier_commit (list->elem->value.variable, arguments->elem, scripter_flow_scope_level);

				scripter_ast_evaluate (body);
				scripter_return_flag = NO;

				out.type = scripter_return_value.type;

				switch (out.type)
					{
					case type_int:
						out.value.i = scripter_return_value.value.i;
					break;
					case type_double:
						out.value.d = scripter_return_value.value.d;
					break;
					case type_vector:
						out.value.vec = phoebe_vector_duplicate (scripter_return_value.value.vec);
					break;
					default:
						phoebe_scripter_output ("function doesn't return a numeric value, aborting.\n");
					break;
					}

				/* The value has been calculated, end the scope:                      */
				scripter_ast_flow_end_scope ();

				/* Reset the output value to void (in case of illegal definition):    */
				scripter_return_value.type = type_void;

				return out;
				}

			if (in->value.node.kind == kind_info)
				{
				/* This is the 'info' directive. It takes one argument, either quali- */
				/* fier or a keyword. It lists relevant information for the passed    */
				/* parameter and always returns void.                                 */

				/* The following assignment will always work because of grammar rules */
				char *argument  = in->value.node.args->elem->value.variable;
				char *qualifier;
				PHOEBE_type type;
				out.type = type_void;

				do
					{
					/* 1st alternative: the argument is a qualifier:                    */
					type = phoebe_type_from_qualifier (argument);
					if (type != -1)
						{
						qualifier = strdup (argument);
						break;
						}

					/* 2nd alternative: the argument is a keyword:                      */
					type = phoebe_type_from_keyword (argument);
					if (type != -1)
						{
						qualifier = strdup (phoebe_qualifier_from_keyword (argument));
						type = phoebe_type_from_qualifier (qualifier);
						break;
						}

					phoebe_scripter_output ("the argument is not a qualifier or a keyword, aborting.\n");
					return out;
					}
				while (FALSE);

				/* First: qualifier, keyword and description:                         */
				fprintf (PHOEBE_output, "\n");
				fprintf (PHOEBE_output, "  Description:    %s\n", phoebe_description_from_qualifier (qualifier));
				fprintf (PHOEBE_output, "  Qualifier:      %s\n", qualifier);

				/* Next: type and value:                                              */
				switch (type)
					{
					case TYPE_INT:
						fprintf (PHOEBE_output, "  Type:           integer\n");
						fprintf (PHOEBE_output, "  Value:          %d\n", phoebe_get_value_int (qualifier));
					break;
					case TYPE_BOOL:
						fprintf (PHOEBE_output, "  Type:           boolean\n");
						if (phoebe_get_value_bool (qualifier) == TRUE)
							fprintf (PHOEBE_output, "  Value:          yes\n");
						else
							fprintf (PHOEBE_output, "  Value:          no\n");
					break;
					case TYPE_DOUBLE:
						fprintf (PHOEBE_output, "  Type:           real\n");
						fprintf (PHOEBE_output, "  Value:          %f\n", phoebe_get_value_double (qualifier));
					break;
					case TYPE_STRING:
						fprintf (PHOEBE_output, "  Type:           string\n");
						fprintf (PHOEBE_output, "  Value:          %s\n", phoebe_get_value_string (qualifier));
					break;
					case TYPE_INT_ARRAY:
						{
						PHOEBE_array *array = phoebe_array_new_from_qualifier (qualifier);
						fprintf (PHOEBE_output, "  Type:           integer array\n");
						if (array != NULL)
							{
							fprintf (PHOEBE_output, "  Value:          ");
							phoebe_array_print (array);
							fprintf (PHOEBE_output, "\n");
							phoebe_array_free (array);
							}
						else
							fprintf (PHOEBE_output, "  Value:          <empty array>\n");
						}
					break;
					case TYPE_BOOL_ARRAY:
						{
						PHOEBE_array *array = phoebe_array_new_from_qualifier (qualifier);
						fprintf (PHOEBE_output, "  Type:           boolean array\n");
						if (array != NULL)
							{
							fprintf (PHOEBE_output, "  Value:          ");
							phoebe_array_print (array);
							fprintf (PHOEBE_output, "\n");
							phoebe_array_free (array);
							}
						else
							fprintf (PHOEBE_output, "  Value:          <empty array>\n");
						}
					break;
					case TYPE_DOUBLE_ARRAY:
						{
						PHOEBE_vector *vec = phoebe_vector_new_from_qualifier (qualifier);
						fprintf (PHOEBE_output, "  Type:           real array\n");
						if (vec != NULL)
							{
							fprintf (PHOEBE_output, "  Value:          ");
							phoebe_vector_print (vec);
							fprintf (PHOEBE_output, "\n");
							phoebe_vector_free (vec);
							}
						else
							fprintf (PHOEBE_output, "  Value:          <empty array>\n");
						}
					break;
					case TYPE_STRING_ARRAY:
						{
						PHOEBE_array *array = phoebe_array_new_from_qualifier (qualifier);
						fprintf (PHOEBE_output, "  Type:           string array\n");
						if (array != NULL)
							{
							fprintf (PHOEBE_output, "  Value:          ");
							phoebe_array_print (array);
							fprintf (PHOEBE_output, "\n");
							phoebe_array_free (array);
							}
						else
							fprintf (PHOEBE_output, "  Value:          <empty array>\n");
						}
					break;
					}

				/* Next, is it adjustable?                                            */
					{
					PHOEBE_parameter_kind kind = phoebe_kind_from_qualifier (qualifier);
					int index = phoebe_index_from_qualifier (qualifier);
					if (kind == KIND_ADJUSTABLE)
						{
						fprintf (PHOEBE_output, "  Adjustable:     yes\n");
						fprintf (PHOEBE_output, "  | Marked TBA:   ");
						if (PHOEBE_parameters[index].tba == TRUE)
							fprintf (PHOEBE_output, "yes\n");
						else
							fprintf (PHOEBE_output, "no\n");
						fprintf (PHOEBE_output, "  | Step size:    %lf\n", PHOEBE_parameters[index].step);
						fprintf (PHOEBE_output, "  | Lower limit:  %lf\n", PHOEBE_parameters[index].min);
						fprintf (PHOEBE_output, "  | Upper limit:  %lf\n", PHOEBE_parameters[index].max);
						}
					else
						fprintf (PHOEBE_output, "  Adjustable:     no\n");
					}

				/* Finally, an empty line to finish:                                  */
				fprintf (PHOEBE_output, "\n");
				return out;
				}

			if (in->value.node.kind == kind_show)
				{
				/* This is the 'show' directive. It takes one argument, either quali- */
				/* fier or a keyword. It always returns void.                         */

				/* The following assignment will always work because of grammar rules */
				char *argument  = in->value.node.args->elem->value.variable;
				char *qualifier;
				PHOEBE_type type;
				out.type = type_void;

				do
					{
					/* 1st alternative: the argument is a qualifier:                    */
					type = phoebe_type_from_qualifier (argument);
					if (type != -1)
						{
						qualifier = strdup (argument);
						break;
						}

					/* 2nd alternative: the argument is a keyword:                      */
					type = phoebe_type_from_keyword (argument);
					if (type != -1)
						{
						qualifier = strdup (phoebe_qualifier_from_keyword (argument));
						type = phoebe_type_from_qualifier (qualifier);
						break;
						}

					phoebe_scripter_output ("the argument is not a qualifier or a keyword, aborting.\n");
					return out;
					}
				while (FALSE);

				switch (type)
					{
					case TYPE_INT:
						fprintf (PHOEBE_output, "\t%d\n", phoebe_get_value_int (qualifier));
					break;
					case TYPE_BOOL:
						if (phoebe_get_value_bool (qualifier) == TRUE)
							fprintf (PHOEBE_output, "\tYES\n");
						else
							fprintf (PHOEBE_output, "\tNO\n");
					break;
					case TYPE_DOUBLE:
						fprintf (PHOEBE_output, "\t%f\n", phoebe_get_value_double (qualifier));
					break;
					case TYPE_STRING:
						fprintf (PHOEBE_output, "\t%s\n", phoebe_get_value_string (qualifier));
					break;
					case TYPE_INT_ARRAY:
						{
						PHOEBE_array *array = phoebe_array_new_from_qualifier (qualifier);
						if (array != NULL)
							{
							fprintf (PHOEBE_output, "\t");
							phoebe_array_print (array);
							fprintf (PHOEBE_output, "\n");
							phoebe_array_free (array);
							}
						else
							fprintf (PHOEBE_output, "\t<empty array>\n");
						}
					break;
					case TYPE_BOOL_ARRAY:
						{
						PHOEBE_array *array = phoebe_array_new_from_qualifier (qualifier);
						if (array != NULL)
							{
							fprintf (PHOEBE_output, "\t");
							phoebe_array_print (array);
							fprintf (PHOEBE_output, "\n");
							phoebe_array_free (array);
							}
						else
							fprintf (PHOEBE_output, "\t<empty array>\n");
						}
					break;
					case TYPE_DOUBLE_ARRAY:
						{
						PHOEBE_vector *vec = phoebe_vector_new_from_qualifier (qualifier);
						if (vec != NULL)
							{
							fprintf (PHOEBE_output, "\t");
							phoebe_vector_print (vec);
							fprintf (PHOEBE_output, "\n");
							phoebe_vector_free (vec);
							}
						else
							fprintf (PHOEBE_output, "\t<empty array>\n");
						}
					break;
					case TYPE_STRING_ARRAY:
						{
						PHOEBE_array *array = phoebe_array_new_from_qualifier (qualifier);
						if (array != NULL)
							{
							fprintf (PHOEBE_output, "\t");
							phoebe_array_print (array);
							fprintf (PHOEBE_output, "\n");
							phoebe_array_free (array);
							}
						else
							fprintf (PHOEBE_output, "\t<empty array>\n");
						}
					break;
					}
				return out;
				}

			if (in->value.node.kind == kind_execute_macro)
				{
				scripter_ast_list *arguments = NULL;
				scripter_ast_list *params;
				scripter_ast      *body;
				scripter_ast_list *list;

				scripter_symbol *id = scripter_st_identifier_lookup (in->value.node.args->elem->value.variable);

				if (id == 0)
					{
					phoebe_scripter_output ("macro '%s' not defined.\n", in->value.node.args->elem->value.variable);
					out.type = type_void; return out;
					}
				if ( (id->link->type != ast_node) || ( id->link->value.node.kind != kind_proc ) )
					{
					phoebe_scripter_output ("identifier '%s' is not a macro.\n", in->value.node.args->elem->value.variable);
					out.type = type_void; return out;
					}

				body   = id->link->value.node.args->next->elem;
				params = id->link->value.node.args->next->next;

				/* Now we have to evaluate macro arguments:                           */
				for (list = in->value.node.args->next; list != 0; list = list->next)
					{
					scripter_ast_value val = scripter_ast_evaluate (list->elem);
					switch (val.type)
						{
						case (type_int):
							arguments = scripter_ast_construct_list (scripter_ast_add_int (val.value.i), arguments);
						break;
						case (type_bool):
							arguments = scripter_ast_construct_list (scripter_ast_add_bool (val.value.b), arguments);
						break;
						case (type_double):
							arguments = scripter_ast_construct_list (scripter_ast_add_double (val.value.d), arguments);
						break;
						case (type_string):
							arguments = scripter_ast_construct_list (scripter_ast_add_string (val.value.str), arguments);
						break;
						default:
							phoebe_scripter_output ("*** IMPLEMENT ME!\n");
							out.type = type_void; return out;
						break;
						}
					}
				arguments = scripter_ast_reverse_list (arguments, NULL);

				if (scripter_ast_list_length (params) != scripter_ast_list_length (arguments))
					{
					if (scripter_ast_list_length (params) == 1) phoebe_scripter_output ("macro %s expects %d parameter, but %d are passed.\n", in->value.node.args->elem->value.variable, scripter_ast_list_length (params), scripter_ast_list_length (arguments));
					else if (scripter_ast_list_length (arguments) != 1) phoebe_scripter_output ("macro %s expects %d parameters, but %d are passed.\n", in->value.node.args->elem->value.variable, scripter_ast_list_length (params), scripter_ast_list_length (arguments));
					else phoebe_scripter_output ("macro %s expects %d parameters, but %d is passed.\n", in->value.node.args->elem->value.variable, scripter_ast_list_length (params), scripter_ast_list_length (arguments));
					out.type = type_void; return out;
					}

				/* Start a new environment where the function will be evaluated:      */
				scripter_ast_flow_begin_scope ();

				for (list = params; list != 0; list = list->next, arguments = arguments->next)
					scripter_st_identifier_commit (list->elem->value.variable, arguments->elem, scripter_flow_scope_level);

				out.type = type_void;
				scripter_ast_evaluate (body);
				scripter_ast_flow_end_scope ();

				return out;
				}

			if (in->value.node.kind == kind_execute)
				{
				/* This directive executes an external script. The passed argument is */
				/* either a literal or a string identifier.                           */

				scripter_ast_value value = scripter_ast_evaluate (in->value.node.args->elem);
				out.type = type_void;

				switch (value.type)
					{
					case type_string:
						{
						/* This is the only valid value - a literal or a string ident.    */
						int errcode = scripter_directive_execute (value.value.str);
						switch (errcode)
							{
							case -1:
								phoebe_scripter_output ("filename '%s' cannot be found, aborting.\n", value.value.str);
							break;
							case -2:
								phoebe_scripter_output ("filename '%s' cannot be read, aborting.\n", value.value.str);
							break;
							case -3:
								phoebe_scripter_output ("filename '%s' is a directory, aborting.\n", value.value.str);
							break;
							case -4:
								phoebe_scripter_output ("filename '%s' cannot be open, aborting.\n", value.value.str);
							break;
							}
						}
					break;
					default:
						/* This happens when the passed identifier isn't a string.        */
						phoebe_scripter_output ("'%s' is not a string literal, aborting\n", in->value.node.args->elem->value.variable);
					break;
					}
				return out;
				}

			if (in->value.node.kind == kind_calc)
				{
				/* This is the calculation directive, where a single argument - an    */
				/* expression - is passed. This directive returns void in all cases.  */

				scripter_ast_value value = scripter_ast_evaluate (in->value.node.args->elem);
				out.type = type_void;

				if (value.type == type_void)
					{
					if (in->value.node.args->elem->value.node.kind != kind_func)
						phoebe_scripter_output ("calculation failed.\n");
					return out;
					}

				switch (value.type)
					{
					case (type_int):
						fprintf (PHOEBE_output, "\t%d\n", value.value.i);
					break;
					case (type_bool):
						if (value.value.b == TRUE)
							fprintf (PHOEBE_output, "\tTRUE\n");
						else
							fprintf (PHOEBE_output, "\tFALSE\n");
					break;
					case (type_double):
						fprintf (PHOEBE_output, "\t%lf\n", value.value.d);
					break;
					case (type_string):
						fprintf (PHOEBE_output, "\t\"%s\"\n", value.value.str);
					break;
					case (type_vector):
						fprintf (PHOEBE_output, "\t");
						phoebe_vector_print (value.value.vec);
						fprintf (PHOEBE_output, "\n");
					break;
					case (type_array):
						if (value.value.array->type == TYPE_INT_ARRAY || value.value.array->type == TYPE_DOUBLE_ARRAY)
							{
							fprintf (PHOEBE_output, "\t");
							phoebe_array_print (value.value.array);
							fprintf (PHOEBE_output, "\n");
							}
						else
							phoebe_scripter_output ("calc directive cannot be used with non-numerical arrays.\n");
					break;
					default:
						phoebe_scripter_output ("calc directive cannot be used with that type.\n");
					break;
					}

				return value;
				}

			if (in->value.node.kind == kind_print)
				{
				/* The 'print' directive prints out an arbitrary number of arguments, */
				/* which may be any of PHOEBE's supported types: int, bool, double or */
				/* string. The return value is always void.                           */

				int i;
				scripter_ast_list *s = in->value.node.args;
				out.type = type_void;

				fprintf (PHOEBE_output, "\t");
				while (s != 0)
					{
					scripter_ast_value in = scripter_ast_evaluate (s->elem);
					switch (in.type)
						{
						case (type_int):
							fprintf (PHOEBE_output, "%d", in.value.i);
						break;
						case (type_double):
							fprintf (PHOEBE_output, "%lf", in.value.d);
						break;
						case (type_bool):
							if (in.value.b == TRUE)
								fprintf (PHOEBE_output, "TRUE");
							else fprintf (PHOEBE_output, "FALSE");
						break;
						case (type_string):
							fprintf (PHOEBE_output, "%s", in.value.str);
						break;
						case (type_array):
							phoebe_array_print (in.value.array);
						break;
						case (type_vector):
							phoebe_vector_print (in.value.vec);
						break;
						default:
							phoebe_scripter_output ("cannot print variables of that type.\n");
						break;
						}
					s = s->next;
					}

				fprintf (PHOEBE_output, "\n");
				return out;
				}

			if (in->value.node.kind == kind_if)
				{
				/* This means that we have an IF conditional without the ELSE part.   */
				/* The return value is always void.                                   */

				scripter_ast_value condition  = scripter_ast_evaluate (in->value.node.args->elem);
				scripter_ast *block = in->value.node.args->next->elem;
				out.type = type_void;

				if (condition.type != type_bool)
					{
					phoebe_scripter_output ("the condition is not boolean, so it cannot be evaluated.\n");
					return out;
					}

				if (condition.value.b == TRUE)
					scripter_ast_evaluate (block);

				return out;
				}

			if (in->value.node.kind == kind_ifelse)
				{
				/* This means that we have an IF conditional with the ELSE part. The  */
				/* return value is always void.                                       */

				scripter_ast_value condition  = scripter_ast_evaluate (in->value.node.args->elem);
				scripter_ast *block1 = in->value.node.args->next->elem;
				scripter_ast *block2 = in->value.node.args->next->next->elem;
				out.type = type_void;

				if (condition.type != type_bool)
					{
					phoebe_scripter_output ("the condition is not boolean, so it cannot be evaluated.\n");
					return out;
					}

				if (condition.value.b == TRUE)
					scripter_ast_evaluate (block1);
				else
					scripter_ast_evaluate (block2);

				return out;
				}

			if (in->value.node.kind == kind_stdump)
				{
				scripter_st_print ();
				out.type = type_void;
				return out;
				}

			if (in->value.node.kind == kind_clear)
				{
				if (PHOEBE_output == stdout) printf ("\e[2J");
				}

			if (in->value.node.kind == kind_column)
				{
				scripter_ast_value file = scripter_ast_evaluate (in->value.node.args->elem);
				scripter_ast_value col  = scripter_ast_evaluate (in->value.node.args->next->elem);

				if (file.type != type_string)
					{
					phoebe_scripter_output ("argument 1 to column() is not a string, aborting.\n");
					out.type = type_void;
					return out;
					}
				if (col.type != type_int)
					{
					phoebe_scripter_output ("argument 2 to column() is not an int, aborting.\n");
					out.type = type_void;
					return out;
					}
				out.type = type_vector;
				out.value.vec = phoebe_vector_new_from_column (file.value.str, col.value.i);
				return out;
				}

			if (scripter_ast_list_length (in->value.node.args) == 2)
				{
				/* This means we have a binary arithmetic or a boolean operation:     */

				scripter_ast_value val1 = scripter_ast_evaluate (in->value.node.args->elem);
				scripter_ast_value val2 = scripter_ast_evaluate (in->value.node.args->next->elem);

				/* Now we have to do all binary operations depending on identifier    */
				/* types. The first idea would be to do a double-switch conditional   */
				/* structure, but it is not efficient. Rather, we do case-by-case     */
				/* testing with if sentences according to PHOEBE API documentation,   */
				/* section Arithmetics, Table 6.                                      */

				switch (in->value.node.kind)
					{
					case kind_add:
						if (val1.type == type_int && val2.type == type_int)
							{
							out.type = type_int;
							out.value.i = val1.value.i + val2.value.i;
							return out;
							}
						if (val1.type == type_int && val2.type == type_double)
							{
							out.type = type_double;
							out.value.d = (double) val1.value.i + val2.value.d;
							return out;
							}
						if (val1.type == type_int && val2.type == type_array)
							{
							PHOEBE_array *array;
							if (val2.value.array->type != TYPE_INT_ARRAY && val2.value.array->type != TYPE_DOUBLE_ARRAY)
								{
								out.type = type_void;
								return out;
								}
							}
						if (val1.type == type_int && val2.type == type_vector)
							{
							PHOEBE_vector *vec = phoebe_vector_new ();
							phoebe_vector_alloc (vec, val2.value.vec->dim);
							phoebe_vector_pad (vec, val1.value.i);
							out.type = type_vector;
							out.value.vec = phoebe_vector_new ();
							if (phoebe_vector_add (out.value.vec, vec, val2.value.vec) != 0) out.type = type_void;
							phoebe_vector_free (vec);
							return out;
							}
						if (val1.type == type_int && val2.type == type_string)
							{
							char intstr[20];    /* Max int on PC platforms is +/-2147483647 */
							sprintf (intstr, "%d", val1.value.i);
							out.type = type_string;
							out.value.str = concatenate_strings (intstr, val2.value.str, NULL);
							return out;
							}
						if (val1.type == type_double && val2.type == type_int)
							{
							out.type = type_double;
							out.value.d = val1.value.d + (double) val2.value.i;
							return out;
							}
						if (val1.type == type_double && val2.type == type_double)
							{
							out.type = type_double;
							out.value.d = val1.value.d + val2.value.d;
							return out;
							}
						if (val1.type == type_double && val2.type == type_vector)
							{
							PHOEBE_vector *vec = phoebe_vector_new ();
							phoebe_vector_alloc (vec, val2.value.vec->dim);
							phoebe_vector_pad (vec, val1.value.d);
							out.type = type_vector;
							out.value.vec = phoebe_vector_new ();
							if (phoebe_vector_add (out.value.vec, vec, val2.value.vec) != 0) out.type = type_void;
							phoebe_vector_free (vec);
							return out;
							}
						if (val1.type == type_double && val2.type == type_string)
							{
							char dblstr[100];
							sprintf (dblstr, "%g", val1.value.d);
							out.type = type_string;
							out.value.str = concatenate_strings (dblstr, val2.value.str, NULL);
							return out;
							}
						if (val1.type == type_vector && val2.type == type_int)
							{
							PHOEBE_vector *vec = phoebe_vector_new ();
							phoebe_vector_alloc (vec, val1.value.vec->dim);
							phoebe_vector_pad (vec, val2.value.i);
							out.type = type_vector;
							out.value.vec = phoebe_vector_new ();
							if (phoebe_vector_add (out.value.vec, val1.value.vec, vec) != 0) out.type = type_void;
							phoebe_vector_free (vec);
							return out;
							}
						if (val1.type == type_vector && val2.type == type_double)
							{
							PHOEBE_vector *vec = phoebe_vector_new ();
							phoebe_vector_alloc (vec, val1.value.vec->dim);
							phoebe_vector_pad (vec, val2.value.d);
							out.type = type_vector;
							out.value.vec = phoebe_vector_new ();
							if (phoebe_vector_add (out.value.vec, val1.value.vec, vec) != 0) out.type = type_void;
							phoebe_vector_free (vec);
							return out;
							}
						if (val1.type == type_vector && val2.type == type_vector)
							{
							out.type = type_vector;
							out.value.vec = phoebe_vector_new ();
							if (phoebe_vector_add (out.value.vec, val1.value.vec, val2.value.vec) != 0) out.type = type_void;
							return out;
							}
						if (val1.type == type_string && val2.type == type_int)
							{
							char intstr[20];    /* Max int on PC platforms is +/-2147483647 */
							sprintf (intstr, "%d", val2.value.i);
							out.type = type_string;
							out.value.str = concatenate_strings (val1.value.str, intstr, NULL);
							return out;
							}
						if (val1.type == type_string && val2.type == type_double)
							{
							char dblstr[100];
							sprintf (dblstr, "%g", val2.value.d);
							out.type = type_string;
							out.value.str = concatenate_strings (val1.value.str, dblstr, NULL);
							return out;
							}
						if (val1.type == type_string && val2.type == type_string)
							{
							out.type = type_string;
							out.value.str = concatenate_strings (val1.value.str, val2.value.str, NULL);
							return out;
							}
						if (val1.type == type_bool || val2.type == type_bool)
							phoebe_scripter_output ("operator '+' doesn't act on booleans.\n");

						out.type = type_void;
						return out;
					break;
					case kind_sub:
						if (val1.type == type_int && val2.type == type_int)
							{
							out.type = type_int;
							out.value.i = val1.value.i - val2.value.i;
							return out;
							}
						if (val1.type == type_int && val2.type == type_double)
							{
							out.type = type_double;
							out.value.d = (double) val1.value.i - val2.value.d;
							return out;
							}
						if (val1.type == type_int && val2.type == type_vector)
							{
							PHOEBE_vector *vec = phoebe_vector_new ();
							phoebe_vector_alloc (vec, val2.value.vec->dim);
							phoebe_vector_pad (vec, val1.value.i);
							out.type = type_vector;
							out.value.vec = phoebe_vector_new ();
							if (phoebe_vector_subtract (out.value.vec, vec, val2.value.vec) == -1) out.type = type_void;
							phoebe_vector_free (vec);
							return out;
							}
						if (val1.type == type_double && val2.type == type_int)
							{
							out.type = type_double;
							out.value.d = val1.value.d - (double) val2.value.i;
							return out;
							}
						if (val1.type == type_double && val2.type == type_double)
							{
							out.type = type_double;
							out.value.d = val1.value.d - val2.value.d;
							return out;
							}
						if (val1.type == type_double && val2.type == type_vector)
							{
							PHOEBE_vector *vec = phoebe_vector_new ();
							phoebe_vector_alloc (vec, val2.value.vec->dim);
							phoebe_vector_pad (vec, val1.value.d);
							out.type = type_vector;
							out.value.vec = phoebe_vector_new ();
							if (phoebe_vector_subtract (out.value.vec, vec, val2.value.vec) == -1) out.type = type_void;
							phoebe_vector_free (vec);
							return out;
							}
						if (val1.type == type_vector && val2.type == type_int)
							{
							PHOEBE_vector *vec = phoebe_vector_new ();
							phoebe_vector_alloc (vec, val1.value.vec->dim);
							phoebe_vector_pad (vec, val2.value.i);
							out.type = type_vector;
							out.value.vec = phoebe_vector_new ();
							if (phoebe_vector_subtract (out.value.vec, val1.value.vec, vec) == -1) out.type = type_void;
							phoebe_vector_free (vec);
							return out;
							}
						if (val1.type == type_vector && val2.type == type_double)
							{
							PHOEBE_vector *vec = phoebe_vector_new ();
							phoebe_vector_alloc (vec, val1.value.vec->dim);
							phoebe_vector_pad (vec, val2.value.d);
							out.type = type_vector;
							out.value.vec = phoebe_vector_new ();
							if (phoebe_vector_subtract (out.value.vec, val1.value.vec, vec) == -1) out.type = type_void;
							phoebe_vector_free (vec);
							return out;
							}
						if (val1.type == type_vector && val2.type == type_vector)
							{
							out.type = type_vector;
							out.value.vec = phoebe_vector_new ();
							if (phoebe_vector_subtract (out.value.vec, val1.value.vec, val2.value.vec) == -1) out.type = type_void;
							return out;
							}
						out.type = type_void;
						return out;
					break;
					case kind_mul:
						if (val1.type == type_int && val2.type == type_int)
							{
							out.type = type_int;
							out.value.i = val1.value.i * val2.value.i;
							return out;
							}
						if (val1.type == type_int && val2.type == type_double)
							{
							out.type = type_double;
							out.value.d = (double) val1.value.i * val2.value.d;
							return out;
							}
						if (val1.type == type_int && val2.type == type_vector)
							{
							PHOEBE_vector *vec = phoebe_vector_new ();
							phoebe_vector_alloc (vec, val2.value.vec->dim);
							phoebe_vector_pad (vec, val1.value.i);
							out.type = type_vector;
							out.value.vec = phoebe_vector_new ();
							if (phoebe_vector_multiply (out.value.vec, vec, val2.value.vec) == -1) out.type = type_void;
							phoebe_vector_free (vec);
							return out;
							}
						if (val1.type == type_double && val2.type == type_int)
							{
							out.type = type_double;
							out.value.d = val1.value.d * (double) val2.value.i;
							return out;
							}
						if (val1.type == type_double && val2.type == type_double)
							{
							out.type = type_double;
							out.value.d = val1.value.d * val2.value.d;
							return out;
							}
						if (val1.type == type_double && val2.type == type_vector)
							{
							PHOEBE_vector *vec = phoebe_vector_new ();
							phoebe_vector_alloc (vec, val2.value.vec->dim);
							phoebe_vector_pad (vec, val1.value.d);
							out.type = type_vector;
							out.value.vec = phoebe_vector_new ();
							if (phoebe_vector_multiply (out.value.vec, vec, val2.value.vec) == -1) out.type = type_void;
							phoebe_vector_free (vec);
							return out;
							}
						if (val1.type == type_vector && val2.type == type_int)
							{
							PHOEBE_vector *vec = phoebe_vector_new ();
							phoebe_vector_alloc (vec, val1.value.vec->dim);
							phoebe_vector_pad (vec, val2.value.i);
							out.type = type_vector;
							out.value.vec = phoebe_vector_new ();
							if (phoebe_vector_multiply (out.value.vec, val1.value.vec, vec) == -1) out.type = type_void;
							phoebe_vector_free (vec);
							return out;
							}
						if (val1.type == type_vector && val2.type == type_double)
							{
							PHOEBE_vector *vec = phoebe_vector_new ();
							phoebe_vector_alloc (vec, val1.value.vec->dim);
							phoebe_vector_pad (vec, val2.value.d);
							out.type = type_vector;
							out.value.vec = phoebe_vector_new ();
							if (phoebe_vector_multiply (out.value.vec, val1.value.vec, vec) == -1) out.type = type_void;
							phoebe_vector_free (vec);
							return out;
							}
						if (val1.type == type_vector && val2.type == type_vector)
							{
							out.type = type_vector;
							out.value.vec = phoebe_vector_new ();
							if (phoebe_vector_multiply (out.value.vec, val1.value.vec, val2.value.vec) == -1) out.type = type_void;
							return out;
							}
						out.type = type_void;
						return out;
					break;
					case kind_div:
						if (val1.type == type_int && val2.type == type_int)
							{
							out.type = type_double;
							out.value.d = (double) val1.value.i / (double) val2.value.i;
							return out;
							}
						if (val1.type == type_int && val2.type == type_double)
							{
							out.type = type_double;
							out.value.d = (double) val1.value.i / val2.value.d;
							return out;
							}
						if (val1.type == type_int && val2.type == type_vector)
							{
							PHOEBE_vector *vec = phoebe_vector_new ();
							phoebe_vector_alloc (vec, val2.value.vec->dim);
							phoebe_vector_pad (vec, val1.value.i);
							out.type = type_vector;
							out.value.vec = phoebe_vector_new ();
							if (phoebe_vector_divide (out.value.vec, vec, val2.value.vec) == -1) out.type = type_void;
							phoebe_vector_free (vec);
							return out;
							}
						if (val1.type == type_double && val2.type == type_int)
							{
							out.type = type_double;
							out.value.d = val1.value.d / (double) val2.value.i;
							return out;
							}
						if (val1.type == type_double && val2.type == type_double)
							{
							out.type = type_double;
							out.value.d = val1.value.d / val2.value.d;
							return out;
							}
						if (val1.type == type_double && val2.type == type_vector)
							{
							PHOEBE_vector *vec = phoebe_vector_new ();
							phoebe_vector_alloc (vec, val2.value.vec->dim);
							phoebe_vector_pad (vec, val1.value.d);
							out.type = type_vector;
							out.value.vec = phoebe_vector_new ();
							if (phoebe_vector_divide (out.value.vec, vec, val2.value.vec) == -1) out.type = type_void;
							phoebe_vector_free (vec);
							return out;
							}
						if (val1.type == type_vector && val2.type == type_int)
							{
							PHOEBE_vector *vec = phoebe_vector_new ();
							phoebe_vector_alloc (vec, val1.value.vec->dim);
							phoebe_vector_pad (vec, val2.value.i);
							out.type = type_vector;
							out.value.vec = phoebe_vector_new ();
							if (phoebe_vector_divide (out.value.vec, val1.value.vec, vec) == -1) out.type = type_void;
							phoebe_vector_free (vec);
							return out;
							}
						if (val1.type == type_vector && val2.type == type_double)
							{
							PHOEBE_vector *vec = phoebe_vector_new ();
							phoebe_vector_alloc (vec, val1.value.vec->dim);
							phoebe_vector_pad (vec, val2.value.d);
							out.type = type_vector;
							out.value.vec = phoebe_vector_new ();
							if (phoebe_vector_divide (out.value.vec, val1.value.vec, vec) == -1) out.type = type_void;
							phoebe_vector_free (vec);
							return out;
							}
						if (val1.type == type_vector && val2.type == type_vector)
							{
							out.type = type_vector;
							out.value.vec = phoebe_vector_new ();
							if (phoebe_vector_divide (out.value.vec, val1.value.vec, val2.value.vec) == -1) out.type = type_void;
							return out;
							}
						out.type = type_void;
						return out;
					break;
					case kind_idiv:
						if (val1.type == type_int && val2.type == type_int)
							{
							out.type = type_int;
							out.value.i = val1.value.i % val2.value.i;
							return out;
							}
						out.type = type_void;
						return out;
					break;
					case kind_pot:
						if (val1.type == type_int && val2.type == type_int)
							{
							out.type = type_int;
							out.value.i = (int) pow ((double) val1.value.i, (double) val2.value.i);
							return out;
							}
						if (val1.type == type_int && val2.type == type_double)
							{
							out.type = type_double;
							out.value.d = pow ((double) val1.value.i, val2.value.d);
							return out;
							}
						if (val1.type == type_double && val2.type == type_int)
							{
							out.type = type_double;
							out.value.d = pow (val1.value.d, (double) val2.value.i);
							return out;
							}
						if (val1.type == type_double && val2.type == type_double)
							{
							out.type = type_double;
							out.value.d = pow (val1.value.d, val2.value.d);
							return out;
							}
						if (val1.type == type_vector && val2.type == type_int)
							{
							PHOEBE_vector *vec = phoebe_vector_new ();
							phoebe_vector_alloc (vec, val1.value.vec->dim);
							phoebe_vector_pad (vec, val2.value.i);
							out.type = type_vector;
							out.value.vec = phoebe_vector_new ();
							if (phoebe_vector_raise (out.value.vec, val1.value.vec, vec) == -1) out.type = type_void;
							phoebe_vector_free (vec);
							return out;
							}
						if (val1.type == type_vector && val2.type == type_double)
							{
							PHOEBE_vector *vec = phoebe_vector_new ();
							phoebe_vector_alloc (vec, val1.value.vec->dim);
							phoebe_vector_pad (vec, val2.value.d);
							out.type = type_vector;
							out.value.vec = phoebe_vector_new ();
							if (phoebe_vector_raise (out.value.vec, val1.value.vec, vec) == -1) out.type = type_void;
							phoebe_vector_free (vec);
							return out;
							}
						if (val1.type == type_vector && val2.type == type_vector)
							{
							out.type = type_vector;
							out.value.vec = phoebe_vector_new ();
							if (phoebe_vector_raise (out.value.vec, val1.value.vec, val2.value.vec) == -1) out.type = type_void;
							return out;
							}
						out.type = type_void;
						return out;
					break;
					case kind_equal:
						if (val1.type == type_int && val2.type == type_int)
							{
							out.type = type_bool;
							out.value.b = (val1.value.i == val2.value.i);
							return out;
							}
						if (val1.type == type_int && val2.type == type_double)
							{
							out.type = type_bool;
							out.value.b = (val1.value.i == val2.value.d);
							return out;
							}
						if (val1.type == type_bool && val2.type == type_bool)
							{
							out.type = type_bool;
							out.value.b = (val1.value.b == val2.value.b);
							return out;
							}
						if (val1.type == type_double && val2.type == type_int)
							{
							out.type = type_bool;
							out.value.b = (val1.value.d == val2.value.i);
							return out;
							}
						if (val1.type == type_double && val2.type == type_double)
							{
							out.type = type_bool;
							out.value.b = (val1.value.d == val2.value.d);
							return out;
							}
						if (val1.type == type_vector && val2.type == type_vector)
							{
							out.type = type_bool;
							out.value.b = phoebe_vector_compare (val1.value.vec, val2.value.vec);
							return out;
							}
						if (val1.type == type_string && val2.type == type_string)
							{
							out.type = type_bool;
							if (strcmp (val1.value.str, val2.value.str) == 0) out.value.b = TRUE;
							else out.value.b = FALSE;
							return out;
							}
						out.type = type_void;
						return out;
					break;
					case kind_nequal:
						if (val1.type == type_int && val2.type == type_int)
							{
							out.type = type_bool;
							out.value.b = (val1.value.i != val2.value.i);
							return out;
							}
						if (val1.type == type_int && val2.type == type_double)
							{
							out.type = type_bool;
							out.value.b = (val1.value.i != val2.value.d);
							return out;
							}
						if (val1.type == type_bool && val2.type == type_bool)
							{
							out.type = type_bool;
							out.value.b = (val1.value.b != val2.value.b);
							return out;
							}
						if (val1.type == type_double && val2.type == type_int)
							{
							out.type = type_bool;
							out.value.b = (val1.value.d != val2.value.i);
							return out;
							}
						if (val1.type == type_double && val2.type == type_double)
							{
							out.type = type_bool;
							out.value.b = (val1.value.d != val2.value.d);
							return out;
							}
						if (val1.type == type_vector && val2.type == type_vector)
							{
							out.type = type_bool;
							out.value.b = !phoebe_vector_compare (val1.value.vec, val2.value.vec);
							return out;
							}
						if (val1.type == type_string && val2.type == type_string)
							{
							out.type = type_bool;
							if (strcmp (val1.value.str, val2.value.str) != 0) out.value.b = TRUE;
							else out.value.b = FALSE;
							return out;
							}
						out.type = type_void;
						return out;
					break;
					case kind_lequal:
						if (val1.type == type_int && val2.type == type_int)
							{
							out.type = type_bool;
							out.value.b = (val1.value.i <= val2.value.i);
							return out;
							}
						if (val1.type == type_int && val2.type == type_double)
							{
							out.type = type_bool;
							out.value.b = (val1.value.i <= val2.value.d);
							return out;
							}
						if (val1.type == type_double && val2.type == type_int)
							{
							out.type = type_bool;
							out.value.b = (val1.value.d <= val2.value.i);
							return out;
							}
						if (val1.type == type_double && val2.type == type_double)
							{
							out.type = type_bool;
							out.value.b = (val1.value.d <= val2.value.d);
							return out;
							}
						out.type = type_void;
						return out;
					break;
					case kind_gequal:
						if (val1.type == type_int && val2.type == type_int)
							{
							out.type = type_bool;
							out.value.b = (val1.value.i >= val2.value.i);
							return out;
							}
						if (val1.type == type_int && val2.type == type_double)
							{
							out.type = type_bool;
							out.value.b = (val1.value.i >= val2.value.d);
							return out;
							}
						if (val1.type == type_double && val2.type == type_int)
							{
							out.type = type_bool;
							out.value.b = (val1.value.d >= val2.value.i);
							return out;
							}
						if (val1.type == type_double && val2.type == type_double)
							{
							out.type = type_bool;
							out.value.b = (val1.value.d >= val2.value.d);
							return out;
							}
						out.type = type_void;
						return out;
					break;
					case kind_less:
						if (val1.type == type_int && val2.type == type_int)
							{
							out.type = type_bool;
							out.value.b = (val1.value.i < val2.value.i);
							return out;
							}
						if (val1.type == type_int && val2.type == type_double)
							{
							out.type = type_bool;
							out.value.b = (val1.value.i < val2.value.d);
							return out;
							}
						if (val1.type == type_double && val2.type == type_int)
							{
							out.type = type_bool;
							out.value.b = (val1.value.d < val2.value.i);
							return out;
							}
						if (val1.type == type_double && val2.type == type_double)
							{
							out.type = type_bool;
							out.value.b = (val1.value.d < val2.value.d);
							return out;
							}
						out.type = type_void;
						return out;
					break;
					case kind_greater:
						if (val1.type == type_int && val2.type == type_int)
							{
							out.type = type_bool;
							out.value.b = (val1.value.i > val2.value.i);
							return out;
							}
						if (val1.type == type_int && val2.type == type_double)
							{
							out.type = type_bool;
							out.value.b = (val1.value.i > val2.value.d);
							return out;
							}
						if (val1.type == type_double && val2.type == type_int)
							{
							out.type = type_bool;
							out.value.b = (val1.value.d > val2.value.i);
							return out;
							}
						if (val1.type == type_double && val2.type == type_double)
							{
							out.type = type_bool;
							out.value.b = (val1.value.d > val2.value.d);
							return out;
							}
						out.type = type_void;
						return out;
					break;
					case kind_and:
						if (val1.type == type_bool && val2.type == type_bool)
							{
							out.type = type_bool;
							out.value.b = (val1.value.b && val2.value.b);
							return out;
							}
						out.type = type_void;
						return out;
					break;
					case kind_or:
						if (val1.type == type_bool && val2.type == type_bool)
							{
							out.type = type_bool;
							out.value.b = (val1.value.b || val2.value.b);
							return out;
							}
						out.type = type_void;
						return out;
					break;
					}
				}
 		}
	}

void scripter_ast_flow_begin_scope ()
	{
	/* This function increases the flow depth by one.                           */

	scripter_flow_scope_level++;
	};

void scripter_ast_flow_end_scope ()
	{
	/* This function frees all local variables and reduces the flow depth by 1. */

	int i;
	scripter_symbol *s;

	for (i = 0; i < scripter_st_size; i++)
		if (scripter_st[i] != NULL)
			{
			while (scripter_st[i]->depth == scripter_flow_scope_level)
				{
				scripter_st[i] = scripter_st[i]->next;
				if (scripter_st[i] == NULL) break;
				}
			if (scripter_st[i] == NULL) continue;
			s = scripter_st[i];
			while (s->next != NULL)
				if (s->next->depth == scripter_flow_scope_level) s->next = s->next->next;
				else s = s->next;
			}

	scripter_flow_scope_level--;
	};

void scripter_ast_flow_export_to_main ()
	{
	}

/* ************************************************************************** */
/*             This part of the code describes the Symbol Table:              */
/* ************************************************************************** */

unsigned int scripter_st_identifier_hash (const char *id)
	{
	/* This is a new hash function that behaves much better than the old one,   */
	/* which is still kept for reference below. For formalism details please    */
	/* refer to PHOEBE API.                                                     */

	int MULTIPLIER = 31;
	
	unsigned int h = 0;
	unsigned char *p;

	for (p = (unsigned char *) id; *p != '\0'; p++)
		h = MULTIPLIER * h + *p;

	return h % scripter_st_size;
	}

int scripter_st_identifier_hash_BROKEN (const char *id)
	{
	/* This hash function was suggested by the AST implementation manual (un-   */
	/* fortunately, I don't have it nearby so I can't give the reference), but  */
	/* it fails miserably for our purpose. For example, the function returns a  */
	/* NEGATIVE hash value for 'qwertzuiop' identifier! That's why it has been  */
	/* replaced by the hash function proposed by Kernighan & Pike (1991), TPOP. */

	long out = 0;
	int i = strlen (id) - 1;
	for ( ; i >= 0; i--)
		out = (out << 3) ^ (id[i]);
  return out % scripter_st_size;
	}

void scripter_st_initialize ()
	{
	int i;
	for (i = 0; i < scripter_st_size; i++) scripter_st[i] = 0;
	}

void internal_scripter_st_identifier_insert (const char *id, scripter_ast *ast, int depth)
	{
	/* This is an internal symbol table function to be called exclusively by    */
	/* the 'scripter_st_identifier_commit () function.                          */

	int pos = scripter_st_identifier_hash (id);
	scripter_symbol *s = malloc (sizeof (*s));

	s->name  = strdup (id);
	s->link  = ast;
	s->depth = depth;
	s->next  = scripter_st[pos];
	scripter_st[pos] = s;
	}

int scripter_st_identifier_commit (const char *id, scripter_ast *ast, int depth)
	{
	/* This is the symbol table wrapper function. It checks whether the identi- */
	/* fier has already been initialized and if it has, it updates its value.   */
	/* If it hasn't, it inserts a new identifier. It also checks whether the    */
	/* identifier is local or global (depending on the scope) and acts accor-   */
	/* dingly.                                                                  */

	int hash = scripter_st_identifier_hash (id);
	scripter_symbol *s = scripter_st[hash];

	while (s != NULL)
		{
		if ( (strcmp (s->name, id) == 0) && (s->depth == depth) ) break;
		s = s->next;
		}

	if (s == NULL)
		internal_scripter_st_identifier_insert (id, ast, depth);
	else
		s->link = ast;

	return 0;
	}

int scripter_st_identifier_export (const char *id, scripter_ast *ast)
	{
	int hash = scripter_st_identifier_hash (id);
	scripter_symbol *s = scripter_st[hash];
	while (s != NULL)
		{
		if (strcmp (s->name, id) == 0 && s->depth == scripter_flow_scope_level - 1) break;
		s = s->next;
		}

	if (s == NULL)
		internal_scripter_st_identifier_insert (id, ast, scripter_flow_scope_level - 1);
	else
		s->link = ast;

	return 0;
	}

scripter_symbol *scripter_st_identifier_lookup (const char *id)
	{
	int pos = scripter_st_identifier_hash (id);
	scripter_symbol *s = scripter_st[pos];
	for ( ; s != 0; s = s->next)
		if (strcmp (s->name, id) == 0) return s;
	return 0;
	}

void scripter_st_print ()
	{
	/* This function is used for informational and debugging purposes only. It  */
	/* prints the symbol table on screen.                                       */

	int i;
	scripter_symbol *s;

	fprintf (PHOEBE_output, "------------------------------------------------------------------------\n");
	fprintf (PHOEBE_output, "| Hash: | Variable:    | Dpt | Variable:    | Dpt | Variable:    | Dpt |\n");
	fprintf (PHOEBE_output, "------------------------------------------------------------------------\n");
	for (i = 0; i < scripter_st_size; i++)
		if (scripter_st[i] != 0)
			{
			s = scripter_st[i];
			fprintf (PHOEBE_output, "| %3d   | ", i);
			while (s != NULL)
				{
				char *name = strdup (s->name);
				if (strlen(s->name) >= 12)
					{
					name[11] = '*'; name[12] = '\0';
					}
				fprintf (PHOEBE_output, "%-12s | (%d) | ", name, s->depth);
				s = s->next;
				free (name);
				}
			fprintf (PHOEBE_output, "\n");
			}
	fprintf (PHOEBE_output, "------------------------------------------------------------------------\n");
	}
