#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "phoebe_build_config.h"

#include "phoebe_accessories.h"
#include "phoebe_allocations.h"
#include "phoebe_calculations.h"
#include "phoebe_error_handling.h"
#include "phoebe_fitting.h"
#include "phoebe_fortran_interface.h"
#include "phoebe_keywords.h"
#include "phoebe_type_handling.h"

#include "phoebe_scripter_ast.h"
#include "phoebe_global.h"

int scripter_compute_lc (scripter_ast_list *args)
	{
	/* This command computes a synthetic light curve. The synopsis is:          */
	/*                                                                          */
	/*   compute_lc ("filename", curve [, start, end, vertices] )               */
	/*                                                                          */
	/* Return values:                                                           */
	/*                                                                          */
	/*   -9  ..  independent model variable phoebe_indep readout failure.       */
	/*   -1  ..  arguments' formatting error.                                   */
	/*    0  ..  everything ok, file created.                                   */

	scripter_ast_value file, curve, start, end, vert;
	const char *indep_str = phoebe_get_value_string ("phoebe_indep");
	int indep = 0;
	int    i, errno;

	PHOEBE_data *data;
	FILE *fout;

	phoebe_debug ("entering command compute_lc ().\n");

	if (strcmp (indep_str, "Time (HJD)") == 0) indep = 1;
	if (strcmp (indep_str,      "Phase") == 0) indep = 2;
	if (indep == 0)
		{
		phoebe_scripter_output ("model parameter phoebe_indep is invalid, aborting.\n");
		return -9;
		}

	if (indep == 1) phoebe_scripter_output ("using time (HJD) as independent variable.\n");
	if (indep == 2) phoebe_scripter_output ("  using phase as independent variable.\n");

	file = scripter_ast_evaluate (args->next->elem);
	if (file.type != type_string)
		{
		phoebe_scripter_output ("identifier '%s' is not a string, aborting.\n", args->next->elem->value.variable);
		return -1;
		}

	curve = scripter_ast_evaluate (args->next->next->elem);
	if (curve.type != type_int)
		{
		phoebe_scripter_output ("curve index called with non-integer value, aborting.\n");
		return -1;
		}
	/* Fix the OB1 error:                                                       */
	curve.value.i -= 1;

	/* Now that the type-checking is complete, we initialize the memory:        */
	data = phoebe_data_new ();

	if (args->next->next->next == NULL)
		{
		/* This means that there are no optional range arguments given.           */

		phoebe_scripter_output ("  optional computation range and number of vertices missing;\n");

		if (indep == 1)
			{
			double hjd0 = phoebe_get_value_double ("phoebe_hjd0");
			double period = phoebe_get_value_double ("phoebe_period");

			phoebe_scripter_output ("  assuming range [%3.3lf, %3.3lf] and 300 vertices.\n", hjd0 - period/10.0, hjd0 + 11.0/10.0*period);
			phoebe_data_uniform (data, hjd0 - period/10.0, hjd0 + 11.0/10.0*period, 300);
			}
		else
			{
			phoebe_scripter_output ("  assuming range [-0.6, 0.6] and 300 vertices.\n");
			phoebe_data_uniform (data, -0.6, 0.6, 300);
			}
		}
	else
		{
		/* This means that there are optional arguments to compute_lc (): the     */
		/* range and the number of vertices:                                      */

		start = scripter_ast_evaluate (args->next->next->next->elem);
		if (start.type != type_double)
			{
			phoebe_scripter_output ("phase start called with non-real value, aborting.\n");
			phoebe_data_free (data);
			return -1;
			}
		end = scripter_ast_evaluate (args->next->next->next->next->elem);
		if (end.type != type_double)
			{
			phoebe_scripter_output ("phase end called with non-real value, aborting.\n");
			phoebe_data_free (data);
			return -1;
			}
		vert = scripter_ast_evaluate (args->next->next->next->next->next->elem);
		if (vert.type != type_int)
			{
			phoebe_scripter_output ("vertices passed by non-integer value, aborting.\n");
			phoebe_data_free (data);
			return -1;
			}

		phoebe_scripter_output ("  optional phase range and number of vertices present;\n");
		phoebe_scripter_output ("  using range [%3.3lf, %3.3lf] and %d vertices.\n", start.value.d, end.value.d, vert.value.i);

		phoebe_data_uniform (data, start.value.d, end.value.d, vert.value.i);
		}

	phoebe_debug ("  computing synthetic LC data for curve %d.\n", curve.value.i + 1);
	errno = read_in_synthetic_data (data, curve.value.i, OUTPUT_PHASE, OUTPUT_TOTAL_FLUX, OUTPUT_UNAVAILABLE);

	switch (errno)
		{
		case -2:
			phoebe_scripter_output ("curve %d is not initialized, aborting.\n", curve.value.i + 1);
			phoebe_data_free (data);
			return -2;
		break;
		case -1:
			phoebe_scripter_output ("synthetic data calculation failed.\n");
			phoebe_data_free (data);
			return -1;
		break;
		default:
		break;
		}
	phoebe_debug ("  read-in succesful. Writing LC data to file \"%s\".\n", file.value.str);

	fout = fopen (file.value.str, "w");
	if (fout == NULL)
		{
		phoebe_scripter_output ("file creation failed.\n");
		return -3;
		}
	fprintf (fout, "#        PHASE            FLUX\n");
	for (i = 0; i < data->ptsno; i++)
		fprintf (fout, "%15.6lf\t%15.6lf\n", data->indep[i], data->dep[i]);
	fclose (fout);

	phoebe_data_free (data);

	phoebe_debug ("leaving command compute_lc ().\n");

	phoebe_scripter_output ("file '%s' created.\n", file.value.str);
	return 0;
	}

int scripter_compute_rv (scripter_ast_list *args)
	{
	/* This command computes a synthetic RV curve. The synopsis is:             */
	/*                                                                          */
	/*   compute_rv ("filename", curve [, start, end, vertices] )               */
	/*                                                                          */
	/* Return values:                                                           */
	/*                                                                          */
	/*   -9  ..  independent model variable phoebe_indep readout failure.       */
	/*   -1  ..  arguments' formatting error.                                   */
	/*    0  ..  everything ok, file created.                                   */

	scripter_ast_value file, curve, start, end, vert;
	const char *indep_str = phoebe_get_value_string ("phoebe_indep");
	int indep = 0;
	int    i, errno;

	PHOEBE_data *rv;
	FILE *fout;

	phoebe_debug ("entering command compute_lc ().\n");

	if (strcmp (indep_str, "Time (HJD)") == 0) indep = 1;
	if (strcmp (indep_str,      "Phase") == 0) indep = 2;
	if (indep == 0)
		{
		phoebe_scripter_output ("model parameter phoebe_indep is invalid, aborting.\n");
		return -9;
		}

	if (indep == 1) phoebe_scripter_output ("using time (HJD) as independent variable.\n");
	if (indep == 2) phoebe_scripter_output ("  using phase as independent variable.\n");

	file = scripter_ast_evaluate (args->next->elem);
	if (file.type != type_string)
		{
		phoebe_scripter_output ("identifier '%s' is not a string, aborting.\n", args->next->elem->value.variable);
		return -1;
		}

	curve = scripter_ast_evaluate (args->next->next->elem);
	if (curve.type != type_int)
		{
		phoebe_scripter_output ("curve index called with non-integer value, aborting.\n");
		return -1;
		}
	/* Fix the OB1 error:                                                       */
	curve.value.i -= 1;

	/* Now that the type-checking is complete, we initialize the memory:        */
	rv = phoebe_data_new ();

	if (args->next->next->next == NULL)
		{
		/* This means that there are no optional range arguments given.           */

		phoebe_scripter_output ("  optional computation range and number of vertices missing;\n");

		if (indep == 1)
			{
			double hjd0 = phoebe_get_value_double ("phoebe_hjd0");
			double period = phoebe_get_value_double ("phoebe_period");

			phoebe_scripter_output ("  assuming range [%3.3lf, %3.3lf] and 300 vertices.\n", hjd0 - period/10.0, hjd0 + 11.0/10.0*period);
			phoebe_data_uniform (rv, hjd0 - period/10.0, hjd0 + 11.0/10.0*period, 300);
			}
		else
			{
			phoebe_scripter_output ("  assuming range [-0.6, 0.6] and 300 vertices.\n");
			phoebe_data_uniform (rv, -0.6, 0.6, 300);
			}
		}
	else
		{
		/* This means that there are optional arguments to compute_rv (): the     */
		/* range and the number of vertices:                                      */

		start = scripter_ast_evaluate (args->next->next->next->elem);
		if (start.type != type_double)
			{
			phoebe_scripter_output ("phase start called with non-real value, aborting.\n");
			phoebe_data_free (rv);
			return -1;
			}
		end = scripter_ast_evaluate (args->next->next->next->next->elem);
		if (end.type != type_double)
			{
			phoebe_scripter_output ("phase end called with non-real value, aborting.\n");
			phoebe_data_free (rv);
			return -1;
			}
		vert = scripter_ast_evaluate (args->next->next->next->next->next->elem);
		if (vert.type != type_int)
			{
			phoebe_scripter_output ("vertices passed by non-integer value, aborting.\n");
			phoebe_data_free (rv);
			return -1;
			}

		phoebe_debug ("  optional phase range and number of vertices present;\n");
		phoebe_debug ("  using range [%3.3lf, %3.3lf] and %d vertices.\n", start.value.d, end.value.d, vert.value.i);

		phoebe_data_uniform (rv, start.value.d, end.value.d, vert.value.i);
		}

	phoebe_debug ("  computing synthetic RV data for curve %d.\n", curve.value.i + 1);

	if (curve.value.i == 0)
		errno = read_in_synthetic_data (rv, curve.value.i, OUTPUT_PHASE, OUTPUT_PRIMARY_RV, OUTPUT_UNAVAILABLE);
	if (curve.value.i == 1)
		errno = read_in_synthetic_data (rv, curve.value.i, OUTPUT_PHASE, OUTPUT_SECONDARY_RV, OUTPUT_UNAVAILABLE);

	switch (errno)
		{
		case -2:
			phoebe_scripter_output ("curve %d is not initialized, aborting.\n", curve.value.i + 1);
			phoebe_data_free (rv);
			return -2;
		break;
		case -1:
			phoebe_scripter_output ("synthetic data calculation failed.\n");
			phoebe_data_free (rv);
			return -1;
		break;
		}
	phoebe_debug ("  read-in succesful. Writing RV data to file \"%s\".\n", file.value.str);

	fout = fopen (file.value.str, "w");
	if (fout == NULL)
		{
		phoebe_scripter_output ("file creation failed.\n");
		return -3;
		}
	fprintf (fout, "#        PHASE            RV%d\n", curve.value.i+1);
	for (i = 0; i < rv->ptsno; i++)
		fprintf (fout, "%15.6lf\t%15.6lf\n", rv->indep[i], rv->dep[i]);
	fclose (fout);

	phoebe_data_free (rv);

	phoebe_debug ("leaving command compute_rv ().\n");

	phoebe_scripter_output ("file '%s' created.\n", file.value.str);
	return 0;
	}

int scripter_open_keyword_file (scripter_ast_list *args)
	{
	scripter_ast_value file = scripter_ast_evaluate (args->next->elem);
	int errcode;

	if (file.type != type_string)
		{
		phoebe_scripter_output ("please pass a literal name or a string identifier.\n");
		return -1;
		}
	errcode = open_keyword_file (file.value.str);

	if (errcode == -3) phoebe_scripter_output ("parameter file '%s' not a regular file.\n", file.value.str);
	if (errcode == -2) phoebe_scripter_output ("parameter file '%s' has no read permissions.\n", file.value.str);
	if (errcode == -1) phoebe_scripter_output ("parameter file '%s' doesn't exist.\n", file.value.str);
	if (errcode ==  0) phoebe_scripter_output ("parameter file '%s' opened.\n", file.value.str);
	if (errcode ==  1) phoebe_scripter_output ("'%s' is a legacy parameter file.\n", file.value.str);
	return 0;
	}

int scripter_save_keyword_file (scripter_ast_list *args)
	{
	scripter_ast_value file = scripter_ast_evaluate (args->next->elem);
	int errcode;

	if (file.type != type_string)
		{
		phoebe_scripter_output ("please pass a literal name or a string identifier.\n");
		return -1;
		}

	errcode = save_keyword_file (file.value.str);
	if (errcode == -3) phoebe_scripter_output ("parameter file '%s' not a regular file.\n", file.value.str);
	if (errcode == -2) phoebe_scripter_output ("parameter file '%s' has no write permissions.\n", file.value.str);
	if (errcode ==  0) phoebe_scripter_output ("parameter file '%s' saved.\n", file.value.str);
	return 0;
	}

int scripter_mark_for_adjustment (scripter_ast_list *args)
	{
	char              *qualifier = args->next->elem->value.variable;
	scripter_ast_value state     = scripter_ast_evaluate (args->next->next->elem);

	if (phoebe_kind_from_qualifier (qualifier) != KIND_ADJUSTABLE)
		{
		phoebe_scripter_output ("qualifier %s is not adjustable.\n", qualifier);
		return -1;
		}
	if (state.type == type_bool)
		PHOEBE_parameters[phoebe_index_from_qualifier (qualifier)].tba = state.value.b;
	else if (state.type == type_int)
		if (state.value.i == 0 || state.value.i == 1)
			PHOEBE_parameters[phoebe_index_from_qualifier (qualifier)].tba = state.value.i;
		else
			{
			phoebe_scripter_output ("second argument is not boolean.\n");
			return -2;
			}
	else
		{
		phoebe_scripter_output ("second argument is not boolean.\n");
		return -2;
		}

	return 0;
	}

int scripter_set_parameter_value (scripter_ast_list *args)
	{
	char       *qualifier    = args->next->elem->value.variable;
	PHOEBE_type type         = phoebe_type_from_qualifier (qualifier);
	scripter_ast_value value;
	int         index        = -1;
	int         errno;

	if (type == -1)
		{
		phoebe_scripter_output ("qualifier '%s' doesn't exist.\n", qualifier);
		return -1;
		}

	/* Do the parameter type vs. input type check:                              */

	value = scripter_ast_evaluate (args->next->next->elem);
	if ( (type == TYPE_INT || type == TYPE_INT_ARRAY) && value.type != type_int )
		{
		phoebe_scripter_output ("qualifier '%s' expects an integer value.\n", qualifier);
		return -1;
		}
	if ( (type == TYPE_DOUBLE || type == TYPE_DOUBLE_ARRAY) && (value.type != type_int && value.type != type_double) )
		{
		phoebe_scripter_output ("qualifier '%s' expects a real value.\n", qualifier);
		return -1;
		}
	if ( (type == TYPE_BOOL || type == TYPE_BOOL_ARRAY) && (value.type != type_int && value.type != type_bool) )
		{
		/* It is common to pass 0 or 1 instead of FALSE and TRUE, so we shall     */
		/* propagate these values to boolean type.                                */
		phoebe_scripter_output ("qualifier '%s' expects a boolean value.\n", qualifier);
		return -1;
		}
	if ( (type == TYPE_STRING || type == TYPE_STRING_ARRAY) && value.type != type_string )
		{
		phoebe_scripter_output ("qualifier '%s' expects a string value.\n", qualifier);
		return -1;
		}

	/* Do the type propagation if necessary:                                    */
	if ( (type == TYPE_DOUBLE || type == TYPE_DOUBLE_ARRAY) && value.type == type_int)
		{
		double val;
		
		val = (double) value.value.i; value.type = type_double; value.value.d = val;
		}
	if ( (type == TYPE_BOOL || type == TYPE_BOOL_ARRAY) && value.type == type_int)
		{
		bool val;
		if (value.value.i == 1) val = TRUE;
		else if (value.value.i == 0) val = FALSE;
		else
			{
			phoebe_scripter_output ("qualifier '%s' expects a boolean value.\n", qualifier);
			return -1;
			}
		value.type = type_bool; value.value.b = val;
		}

	/* Do the array check:                                                      */
	if (args->next->next->next != 0)
		{
		/* This means that the there is a passed index parameter to the set_para- */
		/* meter_value () command. We subtract 1 from the index because of OB1    */
		/* user<->machine conflict.                                               */

		scripter_ast_value idx = scripter_ast_evaluate (args->next->next->next->elem);
		if (idx.type != type_int)
			{
			phoebe_scripter_output ("curve index must be an integer value.\n");
			return -1;
			}
		index = idx.value.i;
		index--;
		}

	switch (type)
		{
		case TYPE_INT:
			if (index != -1)
				{
				phoebe_scripter_output ("qualifier '%s' is a scalar value, aborting.\n", qualifier);
				return -1;
				}
			errno = phoebe_set_parameter_value (qualifier, (int) value.value.i);
		break;
		case TYPE_INT_ARRAY:
			if (index < 0)
				{
				phoebe_scripter_output ("qualifier '%s' is a vector value, aborting.\n", qualifier);
				return -1;
				}
			errno = phoebe_set_parameter_value (qualifier, index, value.value.i);
			if (errno == -2) phoebe_scripter_output ("array index [%d] out of bounds, aborting.\n", index+1);
		break;
		case TYPE_BOOL:
			if (index != -1)
				{
				phoebe_scripter_output ("qualifier '%s' is a scalar value, aborting.\n", qualifier);
				return -1;
				}
			errno = phoebe_set_parameter_value (qualifier, value.value.b);
		break;
		case TYPE_BOOL_ARRAY:
			if (index < 0)
				{
				phoebe_scripter_output ("qualifier '%s' is a vector value, aborting.\n", qualifier);
				return -1;
				}
			errno = phoebe_set_parameter_value (qualifier, index, value.value.b);
			if (errno == -2) phoebe_scripter_output ("array index [%d] out of bounds, aborting.\n", index+1);
		break;
		case TYPE_DOUBLE:
			if (index != -1)
				{
				phoebe_scripter_output ("qualifier '%s' is a scalar value, aborting.\n", qualifier);
				return -1;
				}
			errno = phoebe_set_parameter_value (qualifier, value.value.d);
		break;
		case TYPE_DOUBLE_ARRAY:
			if (index < 0)
				{
				phoebe_scripter_output ("qualifier '%s' is a vector value, aborting.\n", qualifier);
				return -1;
				}
			errno = phoebe_set_parameter_value (qualifier, index, value.value.d);
			if (errno == -2) phoebe_scripter_output ("array index [%d] out of bounds, aborting.\n", index+1);
		break;
		case TYPE_STRING:
			if (index != -1)
				{
				phoebe_scripter_output ("qualifier '%s' is not an array, aborting.\n", qualifier);
				return -1;
				}
			errno = phoebe_set_parameter_value (qualifier, value.value.str);
		break;
		case TYPE_STRING_ARRAY:
			if (index < 0)
				{
				phoebe_scripter_output ("qualifier '%s' is a string array, aborting.\n", qualifier);
				return -1;
				}
			errno = phoebe_set_parameter_value (qualifier, index, value.value.str);
			if (errno == -2) phoebe_scripter_output ("array index [%d] out of bounds, aborting.\n", index+1);
		break;
		}

	return 0;
	}

int scripter_set_parameter_limits (scripter_ast_list *args)
	{
	char              *qualifier = args->next->elem->value.variable;
	scripter_ast_value valmin    = scripter_ast_evaluate (args->next->next->elem);
	scripter_ast_value valmax    = scripter_ast_evaluate (args->next->next->next->elem);
	PHOEBE_type        type      = phoebe_type_from_qualifier (qualifier);
	int                index     = phoebe_index_from_qualifier (qualifier);

	if (type == -1)
		{
		phoebe_scripter_output ("qualifier '%s' doesn't exist.\n", qualifier);
		return -1;
		}
	if (phoebe_kind_from_qualifier (qualifier) != KIND_ADJUSTABLE)
		{
		phoebe_scripter_output ("qualifier %s is not adjustable.\n", qualifier);
		return -1;
		}
	if (valmin.type != type_int && valmin.type != type_double)
		{
		phoebe_scripter_output ("the passed lower limit value is not a real value, aborting.\n");
		return -1;
		}
	if (valmax.type != type_int && valmax.type != type_double)
		{
		phoebe_scripter_output ("the passed upper limit value is not a real value, aborting.\n");
		return -1;
		}

	/* Propagate int to double if necessary:                                    */
	if (valmin.type == type_int)
		{
		double val = (double) valmin.value.i; valmin.type = type_double; valmin.value.d = val;
		}
	if (valmax.type == type_int)
		{
		double val = (double) valmax.value.i; valmax.type = type_double; valmax.value.d = val;
		}

	/* Check whether valmax is larger than valmin:                              */
	if (valmin.value.d >= valmax.value.d)
		{
		phoebe_scripter_output ("lower limit is larger than the upper limit, aborting.\n");
		return -2;
		}

	phoebe_set_parameter_limits (qualifier, valmin.value.d, valmax.value.d);
	phoebe_debug ("limits for %s set to (%lf, %lf)\n", qualifier, PHOEBE_parameters[index].min, PHOEBE_parameters[index].max);

	return 0;
	}

int scripter_set_parameter_step (scripter_ast_list *args)
	{
	char              *qualifier = args->next->elem->value.variable;
	scripter_ast_value value     = scripter_ast_evaluate (args->next->next->elem);
	PHOEBE_type        type      = phoebe_type_from_qualifier (qualifier);
	int                index     = phoebe_index_from_qualifier (qualifier);

	if (type == -1)
		{
		phoebe_scripter_output ("qualifier '%s' doesn't exist.\n", qualifier);
		return -1;
		}
	if (phoebe_kind_from_qualifier (qualifier) != KIND_ADJUSTABLE)
		{
		phoebe_scripter_output ("qualifier %s is not adjustable.\n", qualifier);
		return -1;
		}
	if (value.type != type_int && value.type != type_double)
		{
		phoebe_scripter_output ("the passed step value is not a real value, aborting.\n");
		return -1;
		}

	/* Propagate int to double if necessary:                                    */
	if (value.type == type_int)
		{
		double val = (double) value.value.i; value.type = type_double; value.value.d = val;
		}

	phoebe_set_parameter_step (qualifier, value.value.d);
	phoebe_debug ("step for %s set to %lf.\n", qualifier, PHOEBE_parameters[index].step);

	return 0;
	}

int scripter_create_wd_lci_file (scripter_ast_list *args)
	{
	/* This part allows the user to create WD2003 lci input file to use it di-  */
	/* rectly with WD.                                                          */
	/*                                                                          */
	/* Return values:                                                           */
	/*   -3  ..  third argument mismatch, aborting                              */
	/*   -2  ..  second argument mismatch, aborting                             */
	/*   -1  ..  first argument mismatch, aborting                              */
	/*    0  ..  everything ok, LCI file created                                */
	
	scripter_ast_value file  = scripter_ast_evaluate (args->next->elem);
	scripter_ast_value mpage = scripter_ast_evaluate (args->next->next->elem);
	scripter_ast_value curve = scripter_ast_evaluate (args->next->next->next->elem);

	int  errno;
	WD_LCI_parameters params;

	if (file.type != type_string)
		{
		phoebe_scripter_output ("identifier '%s' is not a string, aborting.\n", args->next->elem->value.variable);
		return -1;
		}
	if (mpage.type != type_int)
		{
		phoebe_scripter_output ("2nd argument type mismatch (int expected), aborting.\n");
		return -2;
		}
	if (curve.type != type_int)
		{
		phoebe_scripter_output ("3rd argument type mismatch (int expected), aborting.\n");
		return -3;
		}

	errno = read_in_wd_lci_parameters (&params, mpage.value.i, curve.value.i-1);
	if (errno == -2)
		phoebe_scripter_output ("curve %d is not defined, aborting.\n", curve.value.i);
	if (errno == 1)
		phoebe_scripter_output ("the passed MPAGE=%d isn't (yet) supported.\n", mpage.value.i);
	if (errno == 0)
		{
		create_lci_file (file.value.str, params);
		phoebe_scripter_output ("WD lci file '%s' created.\n", file.value.str);
		}
	return 0;
	}

scripter_ast_value scripter_get_parameter_value (scripter_ast_list *args)
	{
	/* This function returns the value of the passed qualifier. It takes one or */
	/* two arguments; the first one is always the qualifier and the second one  */
	/* is the array index, if the qualifier is wavelength-dependent.            */

	/* The following assignment cannot fail because of grammar rules.           */
	char *qualifier = args->elem->value.variable;

	PHOEBE_type type = phoebe_type_from_qualifier (qualifier);
	int index = -1;

	scripter_ast_value out;

	if (type == -1)
		{
		phoebe_scripter_output ("qualifier '%s' doesn't exist, aborting.\n", qualifier);
		out.type = type_void;
		return out;
		}

	switch (type)
		{
		case (TYPE_INT):
			out.type = type_int;
			out.value.i = phoebe_get_value_int (qualifier);
			return out;
		break;
		case (TYPE_BOOL):
			out.type = type_bool;
			out.value.b = phoebe_get_value_bool (qualifier);
			return out;
		break;
		case (TYPE_DOUBLE):
			out.type = type_double;
			out.value.d = phoebe_get_value_double (qualifier);
			return out;
		break;
		case (TYPE_STRING):
			out.type = type_string;
			out.value.str = (char *) phoebe_get_value_string (qualifier);
			return out;
		break;
		case (TYPE_INT_ARRAY):
			out.type = type_array;
			out.value.array = phoebe_array_new_from_qualifier (qualifier);
			return out;
		break;
		case (TYPE_BOOL_ARRAY):
			out.type = type_array;
			out.value.array = phoebe_array_new_from_qualifier (qualifier);
			return out;
		break;
		case (TYPE_DOUBLE_ARRAY):
			out.type = type_vector;
			out.value.vec = phoebe_vector_new_from_qualifier (qualifier);
			return out;
		break;
		case (TYPE_STRING_ARRAY):
			out.type = type_array;
			out.value.array = phoebe_array_new_from_qualifier (qualifier);
			return out;
		break;
		default:
			phoebe_scripter_output ("how the heck did I get here?? Please report this!\n");
			out.type = type_void;
			return out;
		break;
		}
	}

int scripter_kick_parameters (scripter_ast_list *args)
	{
	/* This function calls the 'kicker'. Its purpose is to kick the solution    */
	/* obtained by e.g. NMS stochastically using the Gaussian PDF. This helps   */
	/* the overall minimization process to escape from local minima.            */
	/*                                                                          */
	/* Return values:                                                           */
	/*   -1  ..  the passed argument is not numeric                             */
	/*    0  ..  everything ok                                                  */

#ifdef HAVE_LIBGSL
#ifndef PHOEBE_GSL_DISABLED

	scripter_ast_value sigma = scripter_ast_evaluate (args->next->elem);

	if (sigma.type == type_int)
		{
		/* This is custom when main-sequence constraint is applied, since the     */
		/* temperatures are fitted to +/- 1K or so. We propagate int to double:   */
		int value = sigma.value.i; sigma.type = type_double; sigma.value.d = (double) value;
		}
	if (sigma.type != type_double)
		{
		phoebe_scripter_output ("the passed sigma is not a numeric value, aborting.\n");
		return -1;
		}

	kick_parameters (sigma.value.d);

	return 0;

#endif
#endif

	phoebe_scripter_output ("kick_parameters () needs GSL support, aborting.\n");
	return 0;
	}

scripter_ast_value scripter_minimize_using_dc (scripter_ast_list *args)
	{
	/* This function calls WD's Differential Corrections (DC) minimizer. Unlike */
	/* NMS, DC doesn't depend on any external library.                          */

	scripter_ast_value out;
	int status;

	phoebe_scripter_output ("DC: initiating the minimization.\n");

	out.type = type_double;
	status = find_minimum_with_dc (&(out.value.d));

	/* Test for exceptions:                                                     */
	if (status == -9)
		phoebe_scripter_output ("independent variable enumeration readout failed.\n");
	if (status == -8)
		phoebe_scripter_output ("dependent variable enumeration readout failed.\n");
	if (status == -7)
		phoebe_scripter_output ("weighting variable enumeration readout failed.\n");
	if (status == -2)
		phoebe_scripter_output ("observational data failure - an error occured during readout.\n");
	if (status ==  1)
		phoebe_scripter_output ("no curves selected for adjustment, nothing to be done.\n");
	if (status ==  2)
		phoebe_scripter_output ("no parameters selected for adjustment, nothing to be done.\n");

	if (status != 0) out.type = type_void;

	/* Say goodbye:                                                             */
	phoebe_scripter_output ("DC minimization done.\n");

	return out;
	}

scripter_ast_value scripter_minimize_using_nms (scripter_ast_list *args)
	{
	/* This function calls the Nelder & Mead simplex minimizer. It only works   */
	/* if GSL is present.                                                       */

	scripter_ast_value out;

#ifdef HAVE_LIBGSL
#ifndef PHOEBE_GSL_DISABLED

	scripter_ast_value tol      = scripter_ast_evaluate (args->elem);
	scripter_ast_value max_iter = scripter_ast_evaluate (args->next->elem);
	int    result;
	double chi2;

	if (tol.type == type_int)
		{
		/* This is custom when main-sequence constraint is applied, since the     */
		/* temperatures are fitted to +/- 1K or so. We propagate int to double:   */
		int value = tol.value.i; tol.type = type_double; tol.value.d = (double) value;
		}
	if (tol.type != type_double)
		{
		phoebe_scripter_output ("the passed tolerance is not a real value, aborting.\n");
		out.type = type_void;
		return out;
		}
	if (max_iter.type != type_int)
		{
		phoebe_scripter_output ("the passed number of iterations is not an integer value, aborting.\n");
		out.type = type_void;
		return out;
		}

	if (max_iter.value.i < 0)
		{
		phoebe_scripter_output ("negative maximum number of iterations, aborting.\n");
		out.type = type_void;
		return out;
		}
	if (max_iter.value.i == 0)
		{
		phoebe_scripter_output ("NMS: tol = %2.2e, unlimited iterations:\n", tol.value.d);
		max_iter.value.i = 1e8;
		}
	else
		phoebe_scripter_output ("NMS: tol = %2.2e, max %d iterations:\n", tol.value.d, max_iter.value.i);

	/* Call the downhill simplex minimizer:                                     */
	result = find_minimum_with_nms (tol.value.d, max_iter.value.i, &chi2);

	/* Test for exceptions:                                                     */
	if (result == -9)
		phoebe_scripter_output ("independent variable enumeration readout failed.\n");
	if (result == -8)
		phoebe_scripter_output ("dependent variable enumeration readout failed.\n");
	if (result == -7)
		phoebe_scripter_output ("weighting variable enumeration readout failed.\n");
	if (result == -2)
		phoebe_scripter_output ("experimental data failure - an error occured during the readout.\n");
	if (result ==  1)
		phoebe_scripter_output ("no curves selected for adjustment, nothing to be done.\n");
	if (result ==  2)
		phoebe_scripter_output ("no parameters selected for adjustment, nothing to be done.\n");

	/* Return the chi2 value if everything was ok:                              */

	if (result == 0)
		{
		out.type = type_double;
		out.value.d = chi2;
		}
	else
		out.type = type_void;

	/* Say goodbye:                                                             */
	phoebe_scripter_output ("NMS minimization done.\n");
	return out;

#endif
#endif

	phoebe_scripter_output ("cannot run NMS minimization without GSL support.\n");
	out.type = type_void;
	return out;
	}

int scripter_phoebe_data_list (scripter_ast_list *args)
	{
	/* This function lists PHOEBE_data contents to screen.                      */

	scripter_ast_value file = scripter_ast_evaluate (args->next->elem);
	PHOEBE_data *data;
	int ptsno;
	int i;

	if (file.type != type_string && file.type != type_data)
		{
		phoebe_scripter_output ("identifier is not a string or a data variable.\n");
		return -1;
		}

	if (file.type == type_string)
		{
		data = phoebe_data_new ();
		ptsno = read_in_experimental_data (file.value.str, data, -1, -1, -1, -1, -1, -1);
		if (ptsno <= 0)
			{
			phoebe_data_free (data);
			phoebe_scripter_output ("filename '%s' doesn't contain experimental data.\n", file.value.str);
			return -1;
			}
		}
	else
		{
		ptsno = file.value.data->ptsno;
		data = file.value.data;
		}

	for (i = 0; i < ptsno; i++)
		printf ("  %lf\t%lf\t%lf\n", data->indep[i], data->dep[i], data->weight[i]);

	return 0;
	}

int scripter_phoebe_data_stats (scripter_ast_list *args)
	{
	/* This function prints experimental data statistics to screen. There are 4 */
	/* ways to call this function:                                              */
	/*                                                                          */
	/*   1) phoebe_data_stats ("literal_path_to_file")                          */
	/*   2) phoebe_data_stats (literal_identifier)                              */
	/*   3) phoebe_data_stats (data_identifier)                                 */
	/*   4) phoebe_data_stats (qualifier, curve_index)                          */

	PHOEBE_data *data;
	scripter_ast_value file = scripter_ast_evaluate (args->next->elem);
	int ptsno;

	if (file.type != type_string && file.type != type_data)
		{
		phoebe_scripter_output ("argument isn't a literal, data or string identifier.\n");
		return -1;
		}

	/* If it is data, we don't have to do anything; if not, we have to read in  */
	/* the data from the file.                                                  */
	if (file.type == type_data)
		{
		ptsno = file.value.data->ptsno;
		data = file.value.data;
		}
	else
		{
		if ( (!filename_is_regular_file      (file.value.str)) ||
		     (!filename_has_read_permissions (file.value.str)) ||
		     (!filename_exists               (file.value.str)) )
			{
			phoebe_scripter_output ("filename '%s' cannot be read.\n", file.value.str);
			return -1;
			}

		data = phoebe_data_new ();
		ptsno = read_in_experimental_data (file.value.str, data, -1, -1, -1, -1, -1, -1);
		if (ptsno <= 0)
			{
			phoebe_data_free (data);
			phoebe_scripter_output ("filename '%s' doesn't contain experimental data.\n", file.value.str);
			return -1;
			}
		}

	/* Now comes the printing part. The following block is a fancy way of crea- */
	/* ting an underlined header that includes the filename:                    */
		{
		char *header = malloc (255 * sizeof (char));
		int i;
		if (file.type != type_data)
			sprintf (header, "\nData statistics for '%s':", file.value.str);
		else
			sprintf (header, "\nData statistics:");
		printf ("%s\n", header);
		for (i = 0; i < strlen (header)-1; i++) printf ("-"); printf ("\n");
		free (header);
		}

	printf ("  Number of points:      %d\n",     ptsno);

	/* The following block calculates the minimum and the maximum value of the  */
	/* data set:                                                                */
		{
		double min, max;
		calculate_data_min_max (*data, &min, &max);
		printf ("  Minimum value:         %6.6lf\n", min);
		printf ("  Maximum value:         %6.6lf\n", max);
		}

	printf ("  Unweighted average:    %6.6lf\n", calculate_average (data));
	printf ("  Weighted average:      %6.6lf\n", calculate_weighted_average (data));
	printf ("  Unweighted sigma:      %6.6lf\n", calculate_sigma (data));
	printf ("  Weighted sigma:        %6.6lf\n", calculate_weighted_sigma (data));
	printf ("  Data median:           %6.6lf\n", calculate_median (*data));
	printf ("  Unweighted sum:        %6.6lf\n", calculate_sum (data));
	printf ("  Weighted sum:          %6.6lf\n", calculate_weighted_sum (data));
	printf ("\n");

	phoebe_data_free (data);
	return 0;
	}

int scripter_transform_hjd_to_phase (scripter_ast_list *args)
	{
	/* This function transforms independent HJD variable to phase. The synopsis */
	/* is as follows:                                                           */
	/*                                                                          */
	/*  set new = transform_hjd_to_phase (orig [, hjd0, period, dpdt, pshift])  */
	/*                                                                          */
	/* Ephemeris parameters are optional. If they're missing, the value is ta-  */
	/* ken from the keyword table.                                              */

	PHOEBE_data *data;
	char *ident = args->next->elem->value.variable;
	scripter_ast_value file  = scripter_ast_evaluate (args->next->next->elem);
	double hjd0, period, dpdt, pshift;
	int ptsno;

	if (file.type != type_data)
		{
		phoebe_scripter_output ("argument isn't a data identifier, aborting.\n");
		return -1;
		}

	ptsno = file.value.data->ptsno;
	data = phoebe_data_duplicate (file.value.data);
	read_in_ephemeris_parameters (&hjd0, &period, &dpdt, &pshift);

	/* If we have any optional arguments, use them to overwrite original values */
	if (scripter_ast_list_length (args) > 3)
		hjd0 = (scripter_ast_evaluate (args->next->next->next->elem)).value.d;
	if (scripter_ast_list_length (args) > 4)
		period = (scripter_ast_evaluate (args->next->next->next->next->elem)).value.d;
	if (scripter_ast_list_length (args) > 5)
		dpdt = (scripter_ast_evaluate (args->next->next->next->next->next->elem)).value.d;
	if (scripter_ast_list_length (args) > 6)
		pshift = (scripter_ast_evaluate (args->next->next->next->next->next->next->elem)).value.d;

	/* Now we have data stored in the 'data' variable. We now transform it:     */
	transform_hjd_to_phase (data, hjd0, period, dpdt, pshift);

	/* Finally, let's define the new variable to hold transformed values:       */
	scripter_st_identifier_commit (ident, scripter_ast_add_data (data), scripter_flow_scope_level);
	phoebe_data_free (data);
	}

int scripter_transform_flux_to_magnitude (scripter_ast_list *args)
	{
	/* This function transforms dependent flux variable to magnitude. The       */
	/* synopsis is as follows:                                                  */
	/*                                                                          */
	/*  set new = transform_flux_to_magnitude (orig [, mnorm])                  */
	/*                                                                          */
	/* Norm magnitude argument is optional. If it's missing, the value is taken */
	/* from the keyword table.                                                  */

	PHOEBE_data *data;
	char *ident = args->next->elem->value.variable;
	scripter_ast_value file  = scripter_ast_evaluate (args->next->next->elem);
	double mnorm;
	int ptsno;

	phoebe_warning ("scripter_transform_flux_to_magnitude() is leaking memory!\n");

	if (file.type != type_data)
		{
		phoebe_scripter_output ("argument isn't a data identifier, aborting.\n");
		return -1;
		}

	ptsno = file.value.data->ptsno;
	data = phoebe_data_duplicate (file.value.data);
	mnorm = phoebe_get_value_double ("phoebe_mnorm");

	/* If we have an optional argument, use it to overwrite original mnorm:     */
	if (scripter_ast_list_length (args) > 3)
		mnorm = (scripter_ast_evaluate (args->next->next->next->elem)).value.d;

	/* Now we have data stored in the 'data' variable. We now transform it:     */
	transform_flux_to_magnitude (data, mnorm);

	/* Finally, let's define the new variable to hold transformed values:       */
	scripter_st_identifier_commit (ident, scripter_ast_add_data (data), scripter_flow_scope_level);
	phoebe_data_free (data);
	}
