#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "phoebe_global.h"

#include "phoebe_allocations.h"
#include "phoebe_error_handling.h"
#include "phoebe_keywords.h"
#include "phoebe_type_handling.h"

PHOEBE_vector *phoebe_vector_new ()
	{
	/* This function initializes a vector for allocation and/or writing.        */

	PHOEBE_vector *vec = phoebe_malloc (sizeof (*vec));

	vec->dim = 0;
	vec->val = NULL;

	return vec;
	}

PHOEBE_vector *phoebe_vector_new_from_qualifier (char *qualifier)
	{
	/* This function returns a newly allocated vector the values of which are   */
	/* taken from the array parameter represented by its qualifier.             */
	/*                                                                          */
	/* If an error occured, NULL is returned.                                   */

	PHOEBE_vector *vec;

	int i, j, range;

	if (phoebe_type_from_qualifier (qualifier) != TYPE_DOUBLE_ARRAY)
		return NULL;

	vec = phoebe_vector_new ();
	i = phoebe_index_from_qualifier (qualifier);
	range = phoebe_get_value_int (PHOEBE_parameters[i].dependency);

	if (range == 0) return NULL;

	phoebe_vector_alloc (vec, range);
	for (j = 0; j < range; j++)
		vec->val[j] = phoebe_get_value_list_double (qualifier, j);

	return vec;
	}

PHOEBE_vector *phoebe_vector_new_from_list (scripter_ast_list *list)
	{
	/* This function returns a vector which is defined by a list, e.g.          */
	/* set a={1,2,3}. If the vector is not valid, NULL is returned.             */

	PHOEBE_vector *vec = phoebe_vector_new ();
	int dim = scripter_ast_list_length (list);
	scripter_ast_value val;
	int i;

	phoebe_vector_alloc (vec, dim);

	for (i = 0; i < dim; i++)
		{
		val = scripter_ast_evaluate (list->elem);
		if (val.type == type_int)
			{
			int value = val.value.i; val.type = type_double; val.value.d = (double) value;
			}
		if (val.type != type_double) return NULL;
		vec->val[i] = val.value.d;
		list = list->next;
		}

	return vec;
	}

PHOEBE_vector *phoebe_vector_new_from_column (char *filename, int col)
	{
	FILE *input;
	PHOEBE_vector *vec = phoebe_vector_new ();
	int i, linecount = 1;

	input = fopen (filename, "r");
	if (input == NULL) return vec;

	while (!feof (input))
		{
		double val;
		char line[255];
		char *delimeter = line;

		fgets (line, 254, input);
		if (feof (input)) break;

		/* Remove the trailing newline:                                           */
		line[strlen(line)-1] = '\0';

		/* Remove comments (if any):                                              */
		if (strchr (line, '#') != NULL) (strchr (line, '#'))[0] = '\0';

		/* Remove any leading whitespaces and empty lines:                        */
		while ( (delimeter[0] == ' ' || delimeter[0] == '\t') && delimeter[0] != '\0') delimeter++;
		if (delimeter[0] == '\0')
			{
			linecount++;
			continue;
			}

		for (i = 1; i < col; i++)
			{
			while (delimeter[0] != ' ' && delimeter[0] != '\t' && delimeter[0] != '\0') delimeter++;
			while ( (delimeter[0] == ' ' || delimeter[0] == '\t') && delimeter[0] != '\0') delimeter++;
			if (delimeter[0] == '\0')
				{
				phoebe_warning ("column %d in line %d cannot be read, skipping.\n", col, linecount);
				break;
				}
			}

		if (delimeter[0] != '\0')
			if (sscanf (delimeter, "%lf", &val) == 1)
				{
				phoebe_vector_realloc (vec, vec->dim + 1);
				vec->val[vec->dim-1] = val;
				}

		linecount++;
		}

	return vec;
	}

PHOEBE_vector *phoebe_vector_duplicate (PHOEBE_vector *vec)
	{
	/* This function makes a duplicate copy of vector 'vec'.                    */

	int i;
	PHOEBE_vector *new;

	if (vec == NULL) return NULL;

	new = phoebe_vector_new ();
	phoebe_vector_alloc (new, vec->dim);
	for (i = 0; i < vec->dim; i++)
		new->val[i] = vec->val[i];

	return new;
	}

int phoebe_vector_alloc (PHOEBE_vector *vec, int dimension)
	{
	/* This function allocates storage memory for a vector of 'dimension'.      */
	/*                                                                          */
	/* Return values:                                                           */
	/*                                                                          */
	/*   -2  ..  vector isn't initialized or it is already allocated.           */
	/*   -1  ..  the passed dimension is invalid (less than 1).                 */
	/*    0  ..  allocation successful.                                         */

	if (vec->dim != 0)
		return -2;

	if (dimension < 1)
		return -1;

	vec->dim = dimension;
	vec->val = phoebe_malloc (sizeof (*(vec->val)) * dimension);
	return 0;
	}

int phoebe_vector_realloc (PHOEBE_vector *vec, int dimension)
	{
	/* This function reallocates storage memory for a vector of 'dimension'.    */
	/*                                                                          */
	/* Return values:                                                           */
	/*                                                                          */
	/*   -1  ..  the passed dimension is invalid (less than 1).                 */
	/*    0  ..  allocation successful.                                         */

	if (dimension < 1)
		return -1;

	vec->dim = dimension;
	vec->val = phoebe_realloc (vec->val, sizeof (*(vec->val)) * dimension);
	return 0;
	}

int phoebe_vector_pad (PHOEBE_vector *vec, double value)
	{
	/* This function pads all vector components with 'value'.                   */

	int i;

	for (i = 0; i < vec->dim; i++)
		vec->val[i] = value;

	return 0;
	}

int phoebe_vector_print (PHOEBE_vector *vec)
	{
	/* This function prints the contents of vector 'vec' to screen.             */
	/*                                                                          */
	/* Return values:                                                           */
	/*   -1   ..   the passed vector is invalid (null)                          */

	int i;

	if (vec == NULL)
		{
		fprintf (PHOEBE_output, "<empty array>");
		return -1;
		}

	fprintf (PHOEBE_output, "{");
	for (i = 0; i < vec->dim - 1; i++)
		fprintf (PHOEBE_output, "%f, ", vec->val[i]);
	fprintf (PHOEBE_output, "%f}", vec->val[i]);

	return 0;
	}

int phoebe_vector_free (PHOEBE_vector *vec)
	{
	/* This function frees the storage memory allocated for vector 'vec'.       */

	if (vec->val != NULL) free (vec->val);
	free (vec);
	return 0;
	}

int phoebe_vector_add (PHOEBE_vector *result, PHOEBE_vector *fac1, PHOEBE_vector *fac2)
	{
	/* This function adds vectors fac1 and fac2 and returns the sum vector.     */
	/*                                                                          */
	/* Return values:                                                           */
	/*   -1  ..  vectors have different dimensions.                             */
	/*    0  ..  calculation successful.                                        */

	int i;

	if (fac1->dim != fac2->dim) return -1;

	phoebe_vector_alloc (result, fac1->dim);
	for (i = 0; i < fac1->dim; i++)
		result->val[i] = fac1->val[i] + fac2->val[i];

	return 0;
	}

int phoebe_vector_subtract (PHOEBE_vector *result, PHOEBE_vector *fac1, PHOEBE_vector *fac2)
	{
	/* This function subtracts vectors fac1 and fac2 and returns the difference */
	/* vector.                                                                  */
	/*                                                                          */
	/* Return values:                                                           */
	/*   -1  ..  vectors have different dimensions.                             */
	/*    0  ..  calculation successful.                                        */

	int i;

	if (fac1->dim != fac2->dim)
		return -1;

	phoebe_vector_alloc (result, fac1->dim);
	for (i = 0; i < fac1->dim; i++)
		result->val[i] = fac1->val[i] - fac2->val[i];

	return 0;
	}

int phoebe_vector_multiply (PHOEBE_vector *result, PHOEBE_vector *fac1, PHOEBE_vector *fac2)
	{
	/* This function multiplies vectors fac1 and fac2 and returns the product   */
	/* vector.                                                                  */
	/*                                                                          */
	/* Return values:                                                           */
	/*   -1  ..  vectors have different dimensions.                             */
	/*    0  ..  calculation successful.                                        */

	int i;

	if (fac1->dim != fac2->dim)
		return -1;

	phoebe_vector_alloc (result, fac1->dim);
	for (i = 0; i < fac1->dim; i++)
		result->val[i] = fac1->val[i] * fac2->val[i];

	return 0;
	}

int phoebe_vector_divide (PHOEBE_vector *result, PHOEBE_vector *fac1, PHOEBE_vector *fac2)
	{
	/* This function divides vectors fac1 and fac2 and returns the quotient     */
	/* vector.                                                                  */
	/*                                                                          */
	/* Return values:                                                           */
	/*   -1  ..  vectors have different dimensions.                             */
	/*    0  ..  calculation successful.                                        */

	int i;

	if (fac1->dim != fac2->dim)
		return -1;

	phoebe_vector_alloc (result, fac1->dim);
	for (i = 0; i < fac1->dim; i++)
		result->val[i] = fac1->val[i] / fac2->val[i];

	return 0;
	}

int phoebe_vector_raise (PHOEBE_vector *result, PHOEBE_vector *fac1, PHOEBE_vector *fac2)
	{
	/* This function raises all elements of vector fac1 to the factor of ele-   */
	/* ment of the vector fac2, basically new[i] = fac1[i]^fac2[i].             */
	/*                                                                          */
	/* Return values:                                                           */
	/*   -1  ..  vectors have different dimensions.                             */
	/*    0  ..  calculation successful.                                        */

	int i;

	if (fac1->dim != fac2->dim)
		return -1;

	phoebe_vector_alloc (result, fac1->dim);
	for (i = 0; i < fac1->dim; i++)
		result->val[i] = pow (fac1->val[i], fac2->val[i]);

	return 0;
	}

int phoebe_vector_multiply_by (PHOEBE_vector *fac1, double factor)
	{
	/* This function multiplies all elements of the vector fac1 with the scalar */
	/* value 'factor'.                                                          */
	/*                                                                          */
	/* Return values:                                                           */
	/*    0  ..  calculation successful.                                        */

	int i;

	for (i = 0; i < fac1->dim; i++)
		fac1->val[i] = fac1->val[i] * factor;

	return 0;
	}

int phoebe_vector_dot_product (double *result, PHOEBE_vector *fac1, PHOEBE_vector *fac2)
	{
	/* This function returns the scalar (dot) product of the two vectors.       */
	/*                                                                          */
	/* Return values:                                                           */
	/*   -1  ..  vectors have different dimensions.                             */
	/*    0  ..  calculation successful.                                        */

	int i;
	*result = 0.0;

	if (fac1->dim != fac2->dim)
		return -1;

	for (i = 0; i < fac1->dim; i++)
		*result += fac1->val[i] * fac2->val[i];
	*result = sqrt (*result);

	return 0;
	}

int phoebe_vector_vec_product (PHOEBE_vector *result, PHOEBE_vector *fac1, PHOEBE_vector *fac2)
	{
	/* This function returns the vector product of the two vectors.             */
	/*                                                                          */
	/* Return values:                                                           */
	/*   -2  ..  the dimension is different than 3.                             */
	/*   -1  ..  vectors have different dimensions.                             */
	/*    0  ..  calculation successful.                                        */

	int i;

	if (fac1->dim != fac2->dim)
		return -1;

	if (fac1->dim != 3)
		return -2;

	phoebe_vector_alloc (result, fac1->dim);
	result->val[0] = fac1->val[1] * fac2->val[2] - fac1->val[2] * fac2->val[1];
	result->val[1] = fac1->val[2] * fac2->val[0] - fac1->val[0] * fac2->val[2];
	result->val[2] = fac1->val[0] * fac2->val[1] - fac1->val[1] * fac2->val[0];

	return 0;
	}

int phoebe_vector_submit (PHOEBE_vector *result, PHOEBE_vector *vec, double func ())
	{
	/* This function calculates the functional value of func() for each element */
	/* of the vector individually.                                              */

	int i;

	phoebe_vector_alloc (result, vec->dim);
	for (i = 0; i < vec->dim; i++)
		result->val[i] = func (vec->val[i]);

	return 0;
	}

int phoebe_vector_norm (double *result, PHOEBE_vector *vec)
	{
	/* This function returns the norm of the vector.                            */
	/*                                                                          */
	/* Return values:                                                           */
	/*    0  ..  calculation successful.                                        */

	int i;

	*result = 0.0;
	for (i = 0; i < vec->dim; i++)
		*result += vec->val[i] * vec->val[i];
	*result = sqrt (*result);

	return 0;
	}

int phoebe_vector_dim (int *result, PHOEBE_vector *vec)
	{
	/* This function returns the dimension of the vector.                       */
	/*                                                                          */
	/* Return values:                                                           */
	/*    0  ..  calculation successful.                                        */

	*result = vec->dim;
	return 0;
	}

int phoebe_vector_randomize (PHOEBE_vector *result, double limit)
	{
	/* This function fills all vector elements with random numbers from the in- */
	/* terval [0, limit]. 'limit' may also be negative, then it's [limit, 0].   */
	/* The vector must be allocated prior to calling this function.             */
	/*                                                                          */
	/* Return values:                                                           */
	/*   -1  ..  the vector is initialized, but the memory isn't allocated      */
	/*    0  ..  calculation successful.                                        */

	int i;

	if (result->dim == 0) return -1;

	for (i = 0; i < result->dim; i++)
		result->val[i] = limit * rand() / RAND_MAX;

	return 0;
	}

bool phoebe_vector_compare (PHOEBE_vector *vec1, PHOEBE_vector *vec2)
	{
	/* This function compares two passed vectors. It returns TRUE if all vector */
	/* elements are the same; it returns FALSE otherwise. The comparison is do- */
	/* ne by comparing the difference of both elements to EPSILON.              */

	int i;
	if (vec1->dim != vec2->dim) return FALSE;

	for (i = 0; i < vec1->dim; i++)
		if (fabs (vec1->val[i] - vec2->val[i]) > PHOEBE_NUMERICAL_ACCURACY) return FALSE;

	return TRUE;
	}

PHOEBE_data *phoebe_data_new ()
	{
	/* This function initializes memory for allocation of PHOEBE_data struct.   */

	PHOEBE_data *data = phoebe_malloc (sizeof (*data));

	data->ptsno  = 0;
	data->indep  = NULL;
	data->dep    = NULL;
	data->weight = NULL;

	return data;
	}

int phoebe_data_alloc (PHOEBE_data *data, int dim)
	{
	/* This function allocates storage memory for data of dimension 'dim'.      */
	/*                                                                          */
	/* Return values:                                                           */
	/*                                                                          */
	/*   -2  ..  data isn't initialized or it is already allocated.             */
	/*   -1  ..  the passed dimension is invalid (less than 1).                 */
	/*    0  ..  allocation successful.                                         */

	if (data->ptsno != 0)
		return -2;

	if (dim < 1)
		return -1;

	data->ptsno  = dim;
	data->indep  = phoebe_malloc (sizeof (*(data->indep))  * dim);
	data->dep    = phoebe_malloc (sizeof (*(data->dep))    * dim);
	data->weight = phoebe_malloc (sizeof (*(data->weight)) * dim);

	return 0;
	}

int phoebe_data_realloc (PHOEBE_data *data, int dim)
	{
	/* This function reallocates storage memory for data of dimension 'dim'.    */
	/*                                                                          */
	/* Return values:                                                           */
	/*                                                                          */
	/*   -1  ..  the passed dimension is invalid (less than 1).                 */
	/*    0  ..  allocation successful.                                         */

	if (dim < 1) return -1;

	data->ptsno  = dim;
	data->indep  = phoebe_realloc (data->indep,  sizeof (*(data->indep))  * dim);
	data->dep    = phoebe_realloc (data->dep,    sizeof (*(data->dep))    * dim);
	data->weight = phoebe_realloc (data->weight, sizeof (*(data->weight)) * dim);

	return 0;
	}

PHOEBE_data *phoebe_data_duplicate (PHOEBE_data *data)
	{
	/* This function makes a duplicate copy of data 'data'. Double-think if you */
	/* really need to use this function, since all data structs are actually    */
	/* pointers and you may simply relate pointers themselves. The duplication  */
	/* is meant for proper AST handling and nothing else. If you use it for any */
	/* temporary storage, you introduce a large overhead to the code.           */

	int i;
	PHOEBE_data *new = phoebe_data_new ();

	phoebe_data_alloc (new, data->ptsno);
	for (i = 0; i < data->ptsno; i++)
		{
		new->indep[i]  = data->indep[i];
		new->dep[i]    = data->dep[i];
		new->weight[i] = data->weight[i];
		}

	return new;
	}

int phoebe_data_uniform (PHOEBE_data *data, double start, double end, int vertices)
	{
	/* This function pads the 'data' variable with uniform range of phases. The */
	/* 'data' must be initialized prior to calling this function.               */
	/*                                                                          */
	/* Return values:                                                           */
	/*   -1  ..  data isn't initialized or is already allocated                 */
	/*    0  ..  everything ok                                                  */

	int i;

	if (data->ptsno != 0) return -1;

	phoebe_data_alloc (data, vertices+1);

	for (i = 0; i <= vertices; i++)
		{
		data->indep[i] = start + i * (end - start) / vertices;
		data->weight[i] = 1.0;
		}

	return 0;
	}

int phoebe_data_free (PHOEBE_data *data)
	{
	/* This function frees the storage memory allocated for data 'data'.        */

	if (data->indep != NULL)  free (data->indep);
	if (data->dep != NULL)    free (data->dep);
	if (data->weight != NULL) free (data->weight);

	free (data);
	return 0;
	}

int phoebe_data_multiply_by (PHOEBE_data *result, PHOEBE_data *fac1, double factor)
	{
	/* This function multiplies the data 'fac1' by 'factor' and output the re-  */
	/* sult in 'result'.                                                        */
	/*                                                                          */
	/* Return values:                                                           */
	/*    0  ..  calculation successful.                                        */

	int i;

	phoebe_data_alloc (result, fac1->ptsno);
	for (i = 0; i < fac1->ptsno; i++)
		result->dep[i] = fac1->dep[i] * factor;

	return 0;
	}

int phoebe_lc_params_update (WD_LCI_parameters params, int curve)
	{
	phoebe_set_parameter_value ("phoebe_hjd0",    params.HJD0);
	phoebe_set_parameter_value ("phoebe_period",  params.PERIOD);
	phoebe_set_parameter_value ("phoebe_dpdt",    params.DPDT);
	phoebe_set_parameter_value ("phoebe_pshift",  params.PSHIFT);
	phoebe_set_parameter_value ("phoebe_ecc",     params.E);
	phoebe_set_parameter_value ("phoebe_sma",     params.SMA);
	phoebe_set_parameter_value ("phoebe_f1",      params.F1);
	phoebe_set_parameter_value ("phoebe_f2",      params.F2);
	phoebe_set_parameter_value ("phoebe_vga",     params.VGA);
	phoebe_set_parameter_value ("phoebe_incl",    params.INCL);
	phoebe_set_parameter_value ("phoebe_grb1",    params.GR1);
	phoebe_set_parameter_value ("phoebe_grb2",    params.GR2);
	phoebe_set_parameter_value ("phoebe_teff1",   params.TAVH);
	phoebe_set_parameter_value ("phoebe_teff2",   params.TAVC);
	phoebe_set_parameter_value ("phoebe_alb1",    params.ALB1);
	phoebe_set_parameter_value ("phoebe_alb2",    params.ALB2);
	phoebe_set_parameter_value ("phoebe_pot1",    params.PHSV);
	phoebe_set_parameter_value ("phoebe_pot2",    params.PCSV);
	phoebe_set_parameter_value ("phoebe_rm",      params.RM);
	phoebe_set_parameter_value ("phoebe_hla",     curve, params.HLA);
	phoebe_set_parameter_value ("phoebe_cla",     curve, params.CLA);
	phoebe_set_parameter_value ("phoebe_ld_lcx1", curve, params.X1A);
	phoebe_set_parameter_value ("phoebe_ld_lcx2", curve, params.X2A);
	phoebe_set_parameter_value ("phoebe_ld_lcy1", curve, params.Y1A);
	phoebe_set_parameter_value ("phoebe_ld_lcy2", curve, params.Y2A);
	}

int phoebe_rv_params_update (WD_LCI_parameters params, int curve)
	{
	phoebe_set_parameter_value ("phoebe_hjd0",    params.HJD0);
	phoebe_set_parameter_value ("phoebe_period",  params.PERIOD);
	phoebe_set_parameter_value ("phoebe_dpdt",    params.DPDT);
	phoebe_set_parameter_value ("phoebe_pshift",  params.PSHIFT);
	phoebe_set_parameter_value ("phoebe_ecc",     params.E);
	phoebe_set_parameter_value ("phoebe_sma",     params.SMA);
	phoebe_set_parameter_value ("phoebe_f1",      params.F1);
	phoebe_set_parameter_value ("phoebe_f2",      params.F2);
	phoebe_set_parameter_value ("phoebe_vga",     params.VGA);
	phoebe_set_parameter_value ("phoebe_incl",    params.INCL);
	phoebe_set_parameter_value ("phoebe_grb1",    params.GR1);
	phoebe_set_parameter_value ("phoebe_grb2",    params.GR2);
	phoebe_set_parameter_value ("phoebe_teff1",   params.TAVH);
	phoebe_set_parameter_value ("phoebe_teff2",   params.TAVC);
	phoebe_set_parameter_value ("phoebe_alb1",    params.ALB1);
	phoebe_set_parameter_value ("phoebe_alb2",    params.ALB2);
	phoebe_set_parameter_value ("phoebe_pot1",    params.PHSV);
	phoebe_set_parameter_value ("phoebe_pot2",    params.PCSV);
	phoebe_set_parameter_value ("phoebe_rm",      params.RM);
	phoebe_set_parameter_value ("phoebe_ld_rvx1", curve, params.X1A);
	phoebe_set_parameter_value ("phoebe_ld_rvx2", curve, params.X2A);
	phoebe_set_parameter_value ("phoebe_ld_rvy1", curve, params.Y1A);
	phoebe_set_parameter_value ("phoebe_ld_rvy2", curve, params.Y2A);
	}

PHOEBE_array *phoebe_array_new (PHOEBE_type type)
	{
	/* This function initializes a new array for allocation and/or writing.     */

	PHOEBE_array *array = phoebe_malloc (sizeof (*array));

	array->dim  = 0;
	array->type = type;
	switch (type)
		{
		case TYPE_INT_ARRAY:    array->  val.iarray = NULL; break;
		case TYPE_DOUBLE_ARRAY: array->  val.darray = NULL; break;
		case TYPE_BOOL_ARRAY:   array->  val.barray = NULL; break;
		case TYPE_STRING_ARRAY: array->val.strarray = NULL; break;
		}

	return array;
	}

int phoebe_array_alloc (PHOEBE_array *array, int dimension)
	{
	/* This function allocates storage memory for a array of 'dimension'.       */
	/*                                                                          */
	/* Return values:                                                           */
	/*                                                                          */
	/*   -2  ..  array isn't initialized or it is already allocated.            */
	/*   -1  ..  the passed dimension is invalid (less than 1).                 */
	/*    0  ..  allocation successful.                                         */

	if (array->dim != 0)
		return -2;

	if (dimension < 1)
		return -1;

	array->dim = dimension;
	switch (array->type)
		{
		case TYPE_INT_ARRAY:    array->  val.iarray = phoebe_malloc (sizeof (*(array->  val.iarray)) * dimension); break;
		case TYPE_DOUBLE_ARRAY: array->  val.darray = phoebe_malloc (sizeof (*(array->  val.darray)) * dimension); break;
		case TYPE_BOOL_ARRAY:   array->  val.barray = phoebe_malloc (sizeof (*(array->  val.barray)) * dimension); break;
		case TYPE_STRING_ARRAY: array->val.strarray = phoebe_malloc (sizeof (*(array->val.strarray)) * dimension); break;
		}
	return 0;
	}

PHOEBE_array *phoebe_array_new_from_qualifier (char *qualifier)
	{
	/* This function returns a newly allocated array the values of which are    */
	/* taken from the array parameter represented by its qualifier.             */
	/*                                                                          */
	/* If an error occured, NULL is returned.                                   */

	PHOEBE_array *array;
	PHOEBE_type type = phoebe_type_from_qualifier (qualifier);

	int i, j, range;

	if ( (type != TYPE_INT_ARRAY) &&
       (type != TYPE_DOUBLE_ARRAY) &&
       (type != TYPE_BOOL_ARRAY) &&
       (type != TYPE_STRING_ARRAY) )
		return NULL;

	array = phoebe_array_new (type);
	i = phoebe_index_from_qualifier (qualifier);
	range = phoebe_get_value_int (PHOEBE_parameters[i].dependency);

	if (range == 0) return NULL;

	phoebe_array_alloc (array, range);
	for (j = 0; j < range; j++)
		{
		switch (type)
			{
			case TYPE_INT_ARRAY:    array->  val.iarray[j] = phoebe_get_value_list_int    (qualifier, j); break;
			case TYPE_DOUBLE_ARRAY: array->  val.darray[j] = phoebe_get_value_list_double (qualifier, j); break;
			case TYPE_BOOL_ARRAY:   array->  val.barray[j] = phoebe_get_value_list_bool   (qualifier, j); break;
			case TYPE_STRING_ARRAY: array->val.strarray[j] = phoebe_get_value_list_string (qualifier, j); break;
			}
		}

	return array;
	}

PHOEBE_array *phoebe_array_duplicate (PHOEBE_array *array)
	{
	/* This function makes a duplicate copy of array 'array'.                   */

	int i;
	PHOEBE_array *new = phoebe_array_new (array->type);

	phoebe_array_alloc (new, array->dim);
	new->type = array->type;

	switch (new->type)
		{
		case (TYPE_INT_ARRAY):
			for (i = 0; i < new->dim; i++) new->val.iarray[i] = array->val.iarray[i];
		break;
		case (TYPE_BOOL_ARRAY):
			for (i = 0; i < new->dim; i++) new->val.barray[i] = array->val.barray[i];
		break;
		case (TYPE_DOUBLE_ARRAY):
			for (i = 0; i < new->dim; i++) new->val.darray[i] = array->val.darray[i];
		break;
		case (TYPE_STRING_ARRAY):
			for (i = 0; i < new->dim; i++) new->val.strarray[i] = array->val.strarray[i];
		break;
		}

	return new;
	}

int phoebe_array_print (PHOEBE_array *array)
	{
	/* This function prints the contents of array 'array' to screen.            */
	/*                                                                          */
	/* Return values:                                                           */
	/*   -1   ..   the passed array is invalid (null)                           */

	int i;

	if (array == NULL) return -1;

	fprintf (PHOEBE_output, "{");

	for (i = 0; i < array->dim; i++)
		{
		switch (array->type)
			{
			case TYPE_INT_ARRAY:    fprintf (PHOEBE_output, "%d",     array->  val.iarray[i]); break;
			case TYPE_DOUBLE_ARRAY: fprintf (PHOEBE_output, "%f",     array->  val.darray[i]); break;
			case TYPE_BOOL_ARRAY:   fprintf (PHOEBE_output, "%d",     array->  val.barray[i]); break;
			case TYPE_STRING_ARRAY: fprintf (PHOEBE_output, "\"%s\"", array->val.strarray[i]); break;
			}
		if (i == array->dim - 1) fprintf (PHOEBE_output, "}");
		else fprintf (PHOEBE_output, ", ");
		}

	return 0;
	}

int phoebe_array_free (PHOEBE_array *array)
	{
	/* This function frees the storage memory allocated for array 'array'.       */

	switch (array->type)
		{
		case TYPE_INT_ARRAY:    if (array->  val.iarray != NULL) free (array->  val.iarray); break;
		case TYPE_DOUBLE_ARRAY: if (array->  val.darray != NULL) free (array->  val.darray); break;
		case TYPE_BOOL_ARRAY:   if (array->  val.barray != NULL) free (array->  val.barray); break;
		case TYPE_STRING_ARRAY: if (array->val.strarray != NULL) free (array->val.strarray); break;
		}
	free (array);
	return 0;
	}
